#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <iomanip>
#include <gdkmm.h>
#include <gtkmm.h>

#include "cfmuvalidate.hh"
#include "icaofpl.h"

CFMUValidate::MainWindow::MainWindow(BaseObjectType* cobject, const Glib::RefPtr<Builder>& builder)
	: Gtk::Window(cobject), m_engine(0), m_builder(builder), m_aboutwindow(0),
	  m_textviewresults(0), m_scrolledwindowbrowser(0), m_buttonabout(0), m_buttonclear(0),
	  m_buttonvalidate(0), m_buttonvalidatelocal(0), m_buttonvalidateadr(0),
	  m_buttonquit(0), m_portal_webview(0), m_validate_webview(0),
	  m_portal_webview_close_sighandler(0), m_portal_webview_create_sighandler(0),
	  m_portal_webview_ready_sighandler(0), m_portal_webview_loadstatus_sighandler(0),
	  m_portal_webview_navpolicy_sighandler(0), m_portal_webview_newwndpolicy_sighandler(0),
	  m_portal_webview_rsrcload_sighandler(0), m_portal_webview_consolemsg_sighandler(0),
	  m_validate_webview_close_sighandler(0), m_validate_webview_create_sighandler(0),
	  m_validate_webview_ready_sighandler(0), m_validate_webview_loadstatus_sighandler(0),
	  m_validate_webview_navpolicy_sighandler(0), m_validate_webview_newwndpolicy_sighandler(0),
	  m_validate_webview_rsrcload_sighandler(0), m_validate_webview_consolemsg_sighandler(0),
	  m_app(0), m_wvstate(wvstate_failed), m_tfrloaded(false), m_honourprofilerules(true)
{
        m_builder->get_widget("cfmuaboutdialog", m_aboutwindow);
	m_builder->get_widget("textviewfplan", m_textviewfplan);
	m_builder->get_widget("textviewresults", m_textviewresults);
	m_builder->get_widget("scrolledwindowbrowser", m_scrolledwindowbrowser);
	m_builder->get_widget("buttonabout", m_buttonabout);
	m_builder->get_widget("buttonclear", m_buttonclear);
	m_builder->get_widget("buttonvalidate", m_buttonvalidate);
	m_builder->get_widget("buttonvalidatelocal", m_buttonvalidatelocal);
	m_builder->get_widget("buttonvalidateadr", m_buttonvalidateadr);
	m_builder->get_widget("buttonquit", m_buttonquit);

	if (m_buttonvalidate)
		m_buttonvalidate->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::validate_clicked));
	if (m_buttonvalidatelocal)
		m_buttonvalidatelocal->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::validate_local_clicked));
	if (m_buttonvalidateadr)
		m_buttonvalidateadr->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::validate_adr_clicked));
	if (m_buttonquit)
		m_buttonquit->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::quit_clicked));
	if (m_buttonabout)
		m_buttonabout->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::about_clicked));
	if (m_buttonclear)
		m_buttonclear->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::clear_clicked));
	if (m_aboutwindow)
		m_aboutwindow->signal_response().connect(sigc::mem_fun(*this, &MainWindow::about_done));
	if (m_textviewfplan) {
		Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewfplan->get_buffer());
		if (buf) {
			buf->set_text("-(FPL-HBPBX-IG -1P28R/L -SDFGLO/S -LSZH1235 -N0135F070 ZUE -EDNY0026 LSZR -)");
			buf->signal_changed().connect(sigc::mem_fun(*this, &MainWindow::fpl_changed));
		}
	}

	if (m_textviewresults) {
		Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewresults->get_buffer());
		if (buf) {
			m_txttagmsginfo = buf->create_tag();
			m_txttagmsginfo->property_weight() = 400;
			m_txttagmsginfo->property_weight_set() = true;
			m_txttagmsginfo->property_style() = Pango::STYLE_NORMAL;
			m_txttagmsginfo->property_style_set() = true;
			m_txttagmsgwarn = buf->create_tag();
			m_txttagmsgwarn->property_weight() = 400;
			m_txttagmsgwarn->property_weight_set() = true;
			m_txttagmsgwarn->property_style() = Pango::STYLE_ITALIC;
			m_txttagmsgwarn->property_style_set() = true;
			m_txttagrule = buf->create_tag();
			m_txttagrule->property_weight() = 600;
			m_txttagrule->property_weight_set() = true;
			m_txttagrule->property_style() = Pango::STYLE_NORMAL;
			m_txttagrule->property_style_set() = true;
			m_txttagdesc = buf->create_tag();
			m_txttagdesc->property_weight() = 400;
			m_txttagdesc->property_weight_set() = true;
			m_txttagdesc->property_style() = Pango::STYLE_NORMAL;
			m_txttagdesc->property_style_set() = true;
			m_txttagoprgoal = buf->create_tag();
			m_txttagoprgoal->property_weight() = 400;
			m_txttagoprgoal->property_weight_set() = true;
			m_txttagoprgoal->property_style() = Pango::STYLE_ITALIC;
			m_txttagoprgoal->property_style_set() = true;
			m_txttagcond = buf->create_tag();
			m_txttagcond->property_weight() = 300;
			m_txttagcond->property_weight_set() = true;
			m_txttagcond->property_style() = Pango::STYLE_NORMAL;
			m_txttagcond->property_style_set() = true;
			m_txttagrestrict = buf->create_tag();
			m_txttagrestrict->property_weight() = 400;
			m_txttagrestrict->property_weight_set() = true;
			m_txttagrestrict->property_style() = Pango::STYLE_NORMAL;
			m_txttagrestrict->property_style_set() = true;
			m_txttagfpl = buf->create_tag();
			m_txttagfpl->property_weight() = 400;
			m_txttagfpl->property_weight_set() = true;
			m_txttagfpl->property_style() = Pango::STYLE_NORMAL;
			m_txttagfpl->property_style_set() = true;
		}
	}

	prepare_webview();
	add_resulttext("Note: the first local validation will take some time due to loading the eRAD restrictions; "
		       "after the first local validation, failed CFMU validation rules will display a rule explanation\n", m_txttagcond);
}

CFMUValidate::MainWindow::~MainWindow()
{
	remove_portal_sighandlers();
	remove_validate_sighandlers();
	delete m_aboutwindow;
	delete m_portal_webview;
	delete m_validate_webview;
	delete m_engine;
}

void CFMUValidate::MainWindow::add_portal_sighandlers(void)
{
	remove_portal_sighandlers();
	m_portal_webview_close_sighandler = g_signal_connect(m_portal_webview->gobj(), "close-web-view", G_CALLBACK(&MainWindow::close_webview_1), this);
	m_portal_webview_create_sighandler = g_signal_connect(m_portal_webview->gobj(), "create-web-view", G_CALLBACK(&MainWindow::create_webview_1), this);
	m_portal_webview_ready_sighandler = g_signal_connect(m_portal_webview->gobj(), "web-view-ready", G_CALLBACK(&MainWindow::webview_ready_1), this);
	m_portal_webview_loadstatus_sighandler = g_signal_connect(m_portal_webview->gobj(), "notify::load-status", G_CALLBACK(&MainWindow::loadstatus_webview_1), this);
	m_portal_webview_navpolicy_sighandler = g_signal_connect(m_portal_webview->gobj(), "navigation-policy-decision-requested", G_CALLBACK(&MainWindow::policy_decision_required_1), this);
	m_portal_webview_newwndpolicy_sighandler = g_signal_connect(m_portal_webview->gobj(), "new-window-policy-decision-requested", G_CALLBACK(&MainWindow::policy_decision_required_1), this);
	m_portal_webview_rsrcload_sighandler = g_signal_connect(m_portal_webview->gobj(), "resource-load-finished", G_CALLBACK(&MainWindow::resourceload_finished_1), this);
	m_portal_webview_consolemsg_sighandler = g_signal_connect(m_portal_webview->gobj(), "console-message", G_CALLBACK(&MainWindow::console_message_1), this);
}

void CFMUValidate::MainWindow::remove_portal_sighandlers(void)
{
        if (m_portal_webview_close_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_close_sighandler);
		m_portal_webview_close_sighandler = 0;
        }
        if (m_portal_webview_create_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_create_sighandler);
		m_portal_webview_create_sighandler = 0;
        }
	if (m_portal_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_ready_sighandler);
		m_portal_webview_ready_sighandler = 0;
	}
        if (m_portal_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_loadstatus_sighandler);
		m_portal_webview_loadstatus_sighandler = 0;
        }
        if (m_portal_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_navpolicy_sighandler);
		m_portal_webview_navpolicy_sighandler = 0;
        }
        if (m_portal_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_newwndpolicy_sighandler);
		m_portal_webview_newwndpolicy_sighandler = 0;
        }
        if (m_portal_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_rsrcload_sighandler);
		m_portal_webview_rsrcload_sighandler = 0;
        }
        if (m_portal_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_portal_webview->gobj(), m_portal_webview_consolemsg_sighandler);
		m_portal_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidate::MainWindow::add_validate_sighandlers(void)
{
	remove_validate_sighandlers();
	m_validate_webview_close_sighandler = g_signal_connect(m_validate_webview->gobj(), "close-web-view", G_CALLBACK(&MainWindow::close_webview_1), this);
	m_validate_webview_create_sighandler = g_signal_connect(m_validate_webview->gobj(), "create-web-view", G_CALLBACK(&MainWindow::create_webview_1), this);
	m_validate_webview_ready_sighandler = g_signal_connect(m_validate_webview->gobj(), "web-view-ready", G_CALLBACK(&MainWindow::webview_ready_1), this);
	m_validate_webview_loadstatus_sighandler = g_signal_connect(m_validate_webview->gobj(), "notify::load-status", G_CALLBACK(&MainWindow::loadstatus_webview_1), this);
	m_validate_webview_navpolicy_sighandler = g_signal_connect(m_validate_webview->gobj(), "navigation-policy-decision-requested", G_CALLBACK(&MainWindow::policy_decision_required_1), this);
	m_validate_webview_newwndpolicy_sighandler = g_signal_connect(m_validate_webview->gobj(), "new-window-policy-decision-requested", G_CALLBACK(&MainWindow::policy_decision_required_1), this);
	m_validate_webview_rsrcload_sighandler = g_signal_connect(m_validate_webview->gobj(), "resource-load-finished", G_CALLBACK(&MainWindow::resourceload_finished_1), this);
	m_validate_webview_consolemsg_sighandler = g_signal_connect(m_validate_webview->gobj(), "console-message", G_CALLBACK(&MainWindow::console_message_1), this);
}

void CFMUValidate::MainWindow::remove_validate_sighandlers(void)
{
        if (m_validate_webview_close_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_close_sighandler);
		m_validate_webview_close_sighandler = 0;
        }
        if (m_validate_webview_create_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_create_sighandler);
		m_validate_webview_create_sighandler = 0;
        }
	if (m_validate_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_ready_sighandler);
		m_validate_webview_ready_sighandler = 0;
	}
        if (m_validate_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_loadstatus_sighandler);
		m_validate_webview_loadstatus_sighandler = 0;
        }
        if (m_validate_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_navpolicy_sighandler);
		m_validate_webview_navpolicy_sighandler = 0;
        }
        if (m_validate_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_newwndpolicy_sighandler);
		m_validate_webview_newwndpolicy_sighandler = 0;
        }
        if (m_validate_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_rsrcload_sighandler);
		m_validate_webview_rsrcload_sighandler = 0;
        }
        if (m_validate_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_validate_webview->gobj(), m_validate_webview_consolemsg_sighandler);
		m_validate_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidate::MainWindow::set_app(CFMUValidate *app)
{
	m_app = app;
}

void CFMUValidate::MainWindow::prepare_webview(void)
{
	remove_portal_sighandlers();
	delete m_portal_webview;
	// add cookie jar
	std::string cookiefile(Glib::build_filename(FPlan::get_userdbpath(), "cookies.txt"));
	SoupSession *session(webkit_get_default_session());
	SoupCookieJar *cookiejar(soup_cookie_jar_text_new(cookiefile.c_str(), false));
	soup_session_add_feature(session, SOUP_SESSION_FEATURE(cookiejar));
	SoupCookie *bwarn(soup_cookie_new("eurocontrol_nm_portal_hideBroswerWarning", "hide", "www.public.cfmu.eurocontrol.int", "/", -1));
	soup_cookie_jar_add_cookie(cookiejar, bwarn);
	bwarn = soup_cookie_new("eurocontrol_nm_portal_hideCookieWarning", "hide", "www.public.cfmu.eurocontrol.int", "/", -1);
	soup_cookie_jar_add_cookie(cookiejar, bwarn);
	g_object_unref(cookiejar);
	// open webview
	m_portal_webview = Gtk::manage(Glib::wrap(webkit_web_view_new()));
	m_portal_webview->show();
	add_portal_sighandlers();
	if (m_scrolledwindowbrowser) {
		m_scrolledwindowbrowser->remove();
		m_scrolledwindowbrowser->add(*m_portal_webview);
	}
	WebKitWebView *webview = WEBKIT_WEB_VIEW(m_portal_webview->gobj());
	webkit_web_view_load_uri(webview, "https://www.public.cfmu.eurocontrol.int/PUBPORTAL/gateway/spec/index.html");
	//webkit_web_view_load_uri(webview, "http://www.baycom.org/~tom");
	set_wvstate(wvstate_loadmainpage);
}

void CFMUValidate::MainWindow::set_wvstate(wvstate_t newstate)
{
	if (newstate == m_wvstate)
		return;
	m_wvstate = newstate;
	if (m_buttonvalidate)
		m_buttonvalidate->set_sensitive(m_wvstate == wvstate_ready);
}

void CFMUValidate::MainWindow::validate_clicked(void)
{
	if (false) {
		WebKitWebView *webview(0);
		if (m_validate_webview)
			webview = WEBKIT_WEB_VIEW(m_validate_webview->gobj());
		else if (m_portal_webview)
			webview = WEBKIT_WEB_VIEW(m_portal_webview->gobj());
		else
			return;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMNodeList *elements(webkit_dom_document_get_elements_by_tag_name(document, "*"));
		gulong element_count(webkit_dom_node_list_get_length(elements));
		if (true)
			std::cerr << "Elements: " << element_count << std::endl;
		for (gulong i = 0; i < element_count; i++) {
			WebKitDOMNode *element(webkit_dom_node_list_item(elements, i));
			{
				const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
				if (typ)
					std::cerr << "element " << i << " type " << typ << std::endl;
			}
			{
				gchar *txt(webkit_dom_node_get_text_content(element));
				if (txt) {
					if (*txt)
						std::cerr << "element " << i << " text \"" << txt << "\"" << std::endl;
					g_free(txt);
				}
			}
			if (WEBKIT_DOM_IS_HTML_INPUT_ELEMENT(element)) {
				char *element_type;
				g_object_get(element, "type", &element_type, NULL);
				std::cerr << "element " << i << " type " << element_type << std::endl;
				g_free(element_type);
			}
			if (WEBKIT_DOM_IS_HTML_ELEMENT(element)) {
				char *class_name = webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(element));
				if (class_name) {
					if (*class_name)
						std::cerr << "element " << i << " html class name " << class_name << std::endl;
					g_free(class_name);
				}
				char *innertxt = webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(element));
				if (innertxt) {
					if (*innertxt)
						std::cerr << "element " << i << " html inner text " << innertxt << std::endl;
					g_free(innertxt);
				}
			}
		}
		g_object_unref(elements);
	}
	if (false) {
		WebKitWebView *webview(0);
		if (m_validate_webview)
			webview = WEBKIT_WEB_VIEW(m_validate_webview->gobj());
		else if (m_portal_webview)
			webview = WEBKIT_WEB_VIEW(m_portal_webview->gobj());
		else
			return;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMNodeList *elements(webkit_dom_document_get_elements_by_tag_name(document, "*"));
		gulong element_count(webkit_dom_node_list_get_length(elements));
		for (gulong i = 0; i < element_count; i++) {
			WebKitDOMNode *element(webkit_dom_node_list_item(elements, i));
			if (!WEBKIT_DOM_IS_HTML_ELEMENT(element))
				continue;
			gchar *txt(webkit_dom_node_get_text_content(element));
			if (!txt)
				continue;
			{
				bool ised(!strcmp(txt, "Free Text Editor"));
				g_free(txt);
				if (!ised)
					continue;
			}
#if 0
			if (!webkit_dom_node_has_attributes(element))
				continue;
			{
				WebKitDOMNamedNodeMap *map(webkit_dom_node_get_attributes(element));
				std::cerr << "Attr:";
				gulong nrels(webkit_dom_named_node_map_get_length(map));
				for (gulong i = 0; i < nrels; ++i) {
					WebKitDOMNode *n(webkit_dom_named_node_map_item(map, i));
					if (!n)
						continue;
					gchar *lname(webkit_dom_node_get_local_name(n));
					if (lname) {
						if (*lname)
							std::cerr << ' ' << lname << ':';
						g_free(lname);
					}
					gchar *txt(webkit_dom_node_get_node_value(n));
					if (!txt)
						continue;
					if (*txt)
						std::cerr << ' ' << txt;
					g_free(txt);
				}
				std::cerr << std::endl;

				//WebKitDOMNode *id(webkit_dom_named_node_map_get_named_item(map, "textContent"));
				//std::cerr << "onclick: " << webkit_dom_node_get_node_value(id) << std::endl;
			}
#endif
			std::cerr << "Free Text Editor Element found!" << std::endl;
			{
				gchar *uri(webkit_dom_node_get_base_uri(element));
				if (uri) {
					if (*uri)
						std::cerr << "  Base URI: " << uri << std::endl;
					g_free(uri);
				}
			}
		}
	}
	if (!m_validate_webview)
		return;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview->gobj()));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		if (!document)
			return;
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD"));
		if (!element || !WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element))
			return;
		if (!m_textviewfplan)
			return;
		Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewfplan->get_buffer());
		if (!buf)
			return;
		const Glib::ustring& txt(buf->get_text());
		webkit_dom_html_text_area_element_set_value(WEBKIT_DOM_HTML_TEXT_AREA_ELEMENT(element), txt.c_str());
		if (txt.empty())
			return;
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL");
		if (!element)
			return;
		zap_errors();
		webkit_web_view_execute_script(webview, "document.getElementById(\"FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL\").click();");
		set_wvstate(wvstate_query);
	}
}
//<button type="button" class="portal_buttonToolsButton_enabled" id="FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL">Validate</button>
//<textarea class="gwt-TextArea ifpuv_freeTextPanel_fplTextArea" id="FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD"></textarea>
//<table cellspacing="0" cellpadding="0" class="cfmu_DisclosurePanel cfmu_DisclosurePanel-closed" id="FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"><tbody><tr><td align="left" style="vertical-align: top; "><a href="javascript:void(0);" style="display: block; " class="header"><table cellspacing="0" cellpadding="0" class="portal_fullWidthStyle"><tbody><tr><td align="left" style="vertical-align: top; " width="" height="" rowspan="1"><div class="cfmu_DisclosurePanel_Label cfmu_DisclosurePanel_Enabled">Validation&nbsp;Results</div></td><td align="left" style="vertical-align: top; width: 100%; " width="" height=""><div class="gwt-HTML">&nbsp;</div></td><td align="left" style="vertical-align: top; " width="" height="" rowspan="1"><div class="portal_disclosurePanelRightTextStyle_0"></div></td></tr></tbody></table></a></td></tr><tr><td align="left" style="vertical-align: top; "><div style="overflow-x: hidden; overflow-y: hidden; display: none; padding-top: 0px; padding-right: 0px; padding-bottom: 0px; padding-left: 12px; "><table cellspacing="0" cellpadding="0" class="portal_cfmuDisclosurePanel_contentContainer content"><tbody><tr><td align="left" style="vertical-align: top; "><table cellspacing="0" cellpadding="0" class="portal_fullSizeStyle portal_fullWidthStyle" style="display: none; "><tbody><tr><td align="left" style="vertical-align: top; "><div class="portal_warningLabel" style="display: none; "></div></td></tr><tr><td align="left" style="vertical-align: top; "><div class="portal_dataValue" style="text-align: left; display: none; ">NO ERRORS</div></td></tr><tr><td align="left" style="vertical-align: top; "><div style="padding-top: 0px; padding-right: 0px; padding-bottom: 0px; padding-left: 0px; position: relative; display: none; " id="FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA.ERRORS_DATA_TABLE" class="portal_fullWidthStyle"><div tabindex="0" style="display: none; position: relative; " class="portal_sliderBar"><input type="text" tabindex="-1" style="opacity: 0; height: 1px; width: 1px; z-index: -1; overflow-x: hidden; overflow-y: hidden; position: absolute; "><div style="position: absolute; left: 0px; " class="portal_sliderBarLine"></div><div class="portal_sliderBarKnob" style="position: absolute; left: -1px; "></div><div style="position: absolute; left: -1px; visibility: visible; " class="portal_sliderBarTick"></div><div style="position: absolute; left: -1px; visibility: visible; " class="portal_sliderBarLabel">1</div><div style="position: absolute; left: 0px; visibility: visible; " class="portal_sliderBarTick"></div><div style="position: absolute; left: 0px; visibility: visible; " class="portal_sliderBarLabel">1</div></div><div></div><table class="portal_noScreen portal_autopagetablecontainer_headerbackground portal_noPrint" cellpadding="0" cellspacing="0"><colgroup><col></colgroup><tbody><tr class="portal_autoPageTableContainer_headerRowStyle"><td class="portal_autoPageTableContainer_headerCellStyle"><div class="gwt-HTML">Errors</div></td></tr></tbody></table><div style="width: 100%; padding-top: 0px; padding-right: 0px; padding-bottom: 0px; padding-left: 0px; margin-top: 0px; margin-right: 0px; margin-bottom: 0px; margin-left: 0px; border-top-style: none; border-right-style: none; border-bottom-style: none; border-left-style: none; border-width: initial; border-color: initial; border-image: initial; position: relative; "><table cellpadding="0" cellspacing="0" class="portal_decotable"><colgroup><col></colgroup><tbody></tbody></table></div></div></td></tr><tr><td align="left" style="vertical-align: top; "><table cellspacing="0" cellpadding="0" class="portal_fullWidthStyle"><tbody><tr><td align="left" style="vertical-align: top; "><div class="portal_dataValue" style="display: none; ">No routes found.</div></td></tr><tr><td align="left" style="vertical-align: top; "><table cellspacing="0" cellpadding="0" id="FREE_TEXT_EDITOR.PROPOSE_ROUTES_RESULTS_AREA.ORIGINAL_ROUTE_DATA_TABLE"><tbody><tr><td align="left" style="vertical-align: top; "><table class="portal_headerTableDecorator_headerbackground" cellpadding="0" cellspacing="0"><colgroup><col></colgroup><tbody><tr class="portal_headerTableDecorator_headerRowStyle"></tr></tbody></table></td></tr><tr><td align="left" style="vertical-align: top; "><table cellpadding="0" cellspacing="0" class="portal_decotable"><colgroup><col></colgroup><tbody></tbody></table></td></tr></tbody></table></td></tr></tbody></table></td></tr></tbody></table></td></tr></tbody></table></div></td></tr></tbody></table>
//<div class="portal_dataValue" style="text-align: left; ">NO ERRORS</div>
//<td class="portal_decotableCell ">ROUTE126: FLIGHT RULES Y WITH NO VFR PART.</td>
//<td class="portal_decotableCell ">ROUTE130: UNKNOWN DESIGNATOR HOC3Z</td>

void CFMUValidate::MainWindow::clear_clicked(void)
{
	if (!m_textviewfplan)
		return;
	Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewfplan->get_buffer());
	if (!buf)
		return;
	buf->set_text("");
}

void CFMUValidate::MainWindow::quit_clicked(void)
{
	if (m_app)
		m_app->quit();
}

void CFMUValidate::MainWindow::about_clicked(void)
{
	if (m_aboutwindow)
		m_aboutwindow->show();
}

void CFMUValidate::MainWindow::about_done(int resp)
{
	if (m_aboutwindow)
		m_aboutwindow->hide();
}

void CFMUValidate::MainWindow::fpl_changed(void)
{
	if (!m_textviewresults)
		return;
	Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewresults->get_buffer());
	if (!buf)
		return;
	buf->set_text("");
}

void CFMUValidate::MainWindow::parse_results(void)
{
	if (!m_validate_webview)
		return;
	typedef std::pair<guint,bool> lvlvisible_t;
	typedef std::pair<WebKitDOMNode *,lvlvisible_t> nodelvlvisible_t;
	std::stack<nodelvlvisible_t> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview->gobj()));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(std::make_pair(WEBKIT_DOM_NODE(element), std::make_pair(0U, true)));
	}
	typedef std::vector<Glib::ustring> results_t;
	results_t results;
	while (!subtree.empty()) {
		WebKitDOMNode *node;
		guint level;
		bool visible;
		{
			nodelvlvisible_t nl(subtree.top());
			node = nl.first;
			level = nl.second.first;
			visible = nl.second.second;
		}
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					visible = !!strcmp(disp, "none");
					g_free(disp);
				}
			}
		}
		if (visible) {
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(std::make_pair(ptr, std::make_pair(level + 1U, visible)));
			}
		}
		if (!WEBKIT_DOM_IS_HTML_ELEMENT(node))
			continue;
		char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
		if (false) {
			std::cerr << std::string(2U * level, ' ') << "element:";
			gchar *lname(webkit_dom_node_get_local_name(WEBKIT_DOM_NODE(node)));
			if (lname) {
				if (*lname)
					std::cerr << " name \"" << lname << "\"";
				g_free(lname);
			}
			lname = webkit_dom_html_element_get_id(WEBKIT_DOM_HTML_ELEMENT(node));
			if (lname) {
				if (*lname)
					std::cerr << " id \"" << lname << "\"";
				g_free(lname);
			}
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(node)));
			if (typ && *typ)
				std::cerr << " type \"" << typ << "\"";
			if (class_name && *class_name)
				std::cerr << " class \"" << class_name << "\"";
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					if (*disp)
						std::cerr << " display \"" << disp << "\"" << std::endl;
					g_free(disp);
				}
			}
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt)
					std::cerr << " text \"" << txt << "\"";
				g_free(txt);
			}
			std::cerr << std::endl;
		}
		if (!class_name)
			continue;
		if (!visible) {
			g_free(class_name);
			continue;
		}
		if (!strcmp(class_name, "portal_dataValue") && WEBKIT_DOM_IS_HTML_DIV_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results.push_back(Glib::ustring(txt));
					if (true)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (false && !strcmp(class_name, "portal_decotableCell ") && WEBKIT_DOM_IS_HTML_TABLE_CELL_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results.push_back(Glib::ustring(txt));
					if (true)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (true && (!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
		    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results.push_back(Glib::ustring(txt));
					if (true)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		g_free(class_name);
	}
	if (!m_textviewresults)
		return;
	Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewresults->get_buffer());
	if (!buf)
		return;
	buf->set_text("");
	// locally look up failing rules and get a description
	for (results_t::const_iterator ri(results.begin()), re(results.end()); ri != re; ++ri) {
		add_resulttext(*ri + "\n", m_txttagmsginfo);
		if (!m_tfrloaded)
			continue;
		Glib::ustring::size_type n(ri->rfind('['));
		if (n == Glib::ustring::npos)
			continue;
		Glib::ustring::size_type ne(ri->find(']', n + 1));
		if (ne == Glib::ustring::npos || ne - n < 2 || ne - n > 11)
			continue;
		std::string ruleid(ri->substr(n + 1, ne - n - 1));
		{
			uint64_t tm(time(0));
			ADR::RestrictionEval::rules_t::const_iterator ri(m_reval.get_rules().begin()), re(m_reval.get_rules().end());
			for (; ri != re; ++ri) {
				const ADR::Object::ptr_t& p(*ri);
				ADR::FlightRestrictionTimeSlice& ts(p->operator()(tm).as_flightrestriction());
				if (!ts.is_valid())
					continue;
				if (ts.get_ident() != ruleid)
					continue;
				add_resulttext(p, tm);
				break;
			}
			if (ri != re)
				continue;
		}
		std::vector<TrafficFlowRestrictions::RuleResult> rr(m_tfr.find_rules(ruleid, true));
		if (rr.empty())
			continue;
		add_resulttext(rr[0]);
	}
	if (!results.empty())
		set_wvstate(wvstate_ready);
}

void CFMUValidate::MainWindow::zap_errors(void)
{
	if (!m_validate_webview)
		return;
	std::stack<WebKitDOMNode *> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview->gobj()));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(WEBKIT_DOM_NODE(element));
	}
	std::ostringstream results;
	while (!subtree.empty()) {
		WebKitDOMNode *node(subtree.top());
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (class_name) {
		     		if ((!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
				    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
					g_free(class_name);
					while (webkit_dom_node_has_child_nodes(node)) {
						GError *error(0);
						webkit_dom_node_remove_child(node, webkit_dom_node_get_first_child(node), &error);
						if (error) {
							std::cerr << "Cannot remove child: " << error->message << std::endl;
							g_error_free(error);
						}
					}
					continue;
				}
				g_free(class_name);
			}
		}
		{
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(ptr);
			}
		}
	}
}

gboolean CFMUValidate::MainWindow::close_webview_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->close_webview(webview);
}

gboolean CFMUValidate::MainWindow::close_webview(WebKitWebView* webview)
{
        set_wvstate(wvstate_failed);
	hide();
	return true;
}

WebKitWebView *CFMUValidate::MainWindow::create_webview_1(WebKitWebView *webview, WebKitWebFrame *frame, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->create_webview(webview, frame);
}

WebKitWebView *CFMUValidate::MainWindow::create_webview(WebKitWebView *webview, WebKitWebFrame *frame)
{
	std::cerr << "new webview!" << std::endl;
	if (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj()))
		return 0;
        g_assert(m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj()));
	if (m_wvstate != wvstate_loadmainpage)
		return 0;
	remove_validate_sighandlers();
	delete m_validate_webview;
	m_validate_webview = Gtk::manage(Glib::wrap(webkit_web_view_new()));
	webkit_web_view_set_settings(WEBKIT_WEB_VIEW(m_validate_webview->gobj()), webkit_web_view_get_settings(WEBKIT_WEB_VIEW(m_portal_webview->gobj())));
	add_validate_sighandlers();
	set_wvstate(wvstate_loadquerypage);
	WebKitWebView *wv(WEBKIT_WEB_VIEW(m_validate_webview->gobj()));
	return wv;
}

gboolean CFMUValidate::MainWindow::webview_ready_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->webview_ready(webview);
}

gboolean CFMUValidate::MainWindow::webview_ready(WebKitWebView* webview)
{
	if (true)
		std::cerr << "WebView: ready" << std::endl;
	if (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj())) {
		if (m_wvstate == wvstate_loadquerypage && m_scrolledwindowbrowser) {
			m_scrolledwindowbrowser->remove();
			m_scrolledwindowbrowser->add(*m_validate_webview);
		}
		m_validate_webview->show_all();
		return false;
	}
	g_assert(m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj()));
	m_portal_webview->show_all();
	return false;
}

void CFMUValidate::MainWindow::loadstatus_webview_1(WebKitWebView *webview, GParamSpec *pspec, gpointer userdata)
{
	static_cast<MainWindow *>(userdata)->loadstatus_webview(webview, pspec);
}

void CFMUValidate::MainWindow::loadstatus_webview(WebKitWebView *webview, GParamSpec *pspec)
{
        WebKitLoadStatus status(webkit_web_view_get_load_status(webview));
        GObject *object(G_OBJECT(webview));
        g_assert((m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj())) ||
		 (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj())));
        g_object_freeze_notify(object);
	switch (status) {
        case WEBKIT_LOAD_FINISHED:
		if (true)
			std::cerr << "WebView: load finished" << std::endl;
		if (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj()))
			m_validate_webview->show_all();
		else if (m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj()))
			m_portal_webview->show_all();
		break;

        case WEBKIT_LOAD_PROVISIONAL:
		if (true)
			std::cerr << "WebView: load provisional" << std::endl;
                break;

        case WEBKIT_LOAD_COMMITTED:
		if (true)
			std::cerr << "WebView: load committed" << std::endl;
                break;

        case WEBKIT_LOAD_FIRST_VISUALLY_NON_EMPTY_LAYOUT:
		if (true)
			std::cerr << "WebView: load first visually nonempty layout" << std::endl;
                break;

        case WEBKIT_LOAD_FAILED:
		if (true)
			std::cerr << "WebView: load failed" << std::endl;
		set_wvstate(wvstate_failed);
                break;

        default:
                break;
        }
        g_object_thaw_notify(object);
}

gboolean CFMUValidate::MainWindow::policy_decision_required_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							      WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->policy_decision_required(webview, webframe, request, action, decision);
}

gboolean CFMUValidate::MainWindow::policy_decision_required(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							    WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision)
{
        g_assert((m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj())) ||
		 (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj())));
	WebKitWebNavigationReason reason(webkit_web_navigation_action_get_reason(action));
        gint button(webkit_web_navigation_action_get_button(action));
        gint state(webkit_web_navigation_action_get_modifier_state(action));
        const char *uri(webkit_network_request_get_uri(request));
	if (true)
		std::cerr << "policy_decision_required: reason " << reason << " button " << button << " state " << state
			  << " uri " << uri << std::endl;
	return false;
}

void CFMUValidate::MainWindow::resourceload_finished_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource, gpointer userdata)
{
	static_cast<MainWindow *>(userdata)->resourceload_finished(webview, webframe, webresource);
}

void CFMUValidate::MainWindow::resourceload_finished(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource)
{
        g_assert((m_portal_webview && webview == WEBKIT_WEB_VIEW(m_portal_webview->gobj())) ||
		 (m_validate_webview && webview == WEBKIT_WEB_VIEW(m_validate_webview->gobj())));
 	if (true) {
		time_t t(time(0));
		char buf[64];
		std::cerr << "resource load finished! " << ctime_r(&t, buf) << " state " << (int)m_wvstate << std::endl;
	}
	switch (m_wvstate) {
	case wvstate_loadmainpage:
	{
		if (!m_portal_webview || webview != WEBKIT_WEB_VIEW(m_portal_webview->gobj()))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL"));
		if (!element)
			break;
		if (false)
			std::cerr << "Button found! Injecting script" << std::endl;
		webkit_web_view_execute_script(webview, "document.getElementById(\"IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL\").click();");
		set_wvstate(wvstate_loadquerypage);
		break;
	}

	case wvstate_loadquerypage:
	{
		if (!m_validate_webview || webview != WEBKIT_WEB_VIEW(m_validate_webview->gobj()))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL"));
		if (!element)
			break;
		if (true) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Validate element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD");
		if (!element)
			break;
		if (true) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "FPL entry element type " << typ << std::endl;
		}
		if (!WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element))
			break;
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA.ERRORS_DATA_TABLE");
		if (!element)
			break;
		if (true) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Result Errors element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA");
		if (!element)
			break;
		webkit_dom_event_target_add_event_listener(WEBKIT_DOM_EVENT_TARGET(element), "DOMSubtreeModified", G_CALLBACK(&MainWindow::result_change_1), true, this);
		if (true) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Results element type " << typ << std::endl;
		}
		set_wvstate(wvstate_ready);
		break;
	}

	case wvstate_query:
	{
		if (!m_validate_webview || webview != WEBKIT_WEB_VIEW(m_validate_webview->gobj()))
			break;
		set_wvstate(wvstate_waitresult);
		parse_results();
		break;
	}

	default:
		break;
	}
}

gboolean CFMUValidate::MainWindow::result_change_1(WebKitDOMElement *target, WebKitDOMEvent *e, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->result_change(target, e);
}

gboolean CFMUValidate::MainWindow::result_change(WebKitDOMElement *target, WebKitDOMEvent *e)
{
	std::cerr << "result change! state " << (int)m_wvstate << std::endl;
	switch (m_wvstate) {
	case wvstate_waitresult:
		parse_results();
		break;

	default:
		break;
	}
	return true;
}

gboolean CFMUValidate::MainWindow::console_message_1(WebKitWebView *webview, gchar *message, gint line, gchar *source_id, gpointer userdata)
{
	return static_cast<MainWindow *>(userdata)->console_message(webview, message, line, source_id);
}

gboolean CFMUValidate::MainWindow::console_message(WebKitWebView *webview, gchar *message, gint line, gchar *source_id)
{
	if (true) {
		std::cerr << message << " (" << (source_id ? "*null*" : source_id) << ':' << line << ')' << std::endl;
	}
	return true;
}

void CFMUValidate::MainWindow::set_dbdirs(const std::string& dir_main, Engine::auxdb_mode_t auxdbmode, const std::string& dir_aux)
{
	m_dir_main = dir_main;
	m_dir_aux = dir_aux;
	m_auxdbmode = auxdbmode;
}

void CFMUValidate::MainWindow::add_resulttext(const Glib::ustring& txt, const Glib::RefPtr<Gtk::TextTag>& tag)
{
	if (!m_textviewresults)
		return;
	Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewresults->get_buffer());
	if (!buf)
		return;
	buf->insert_with_tag(buf->end(), txt, tag);
}

void CFMUValidate::MainWindow::add_resulttext(const TrafficFlowRestrictions::Message& m)
{
	std::ostringstream oss;
	oss << m.get_type_char() << ':';
	if (!m.get_rule().empty())
		oss << " R:" << m.get_rule();
	if (!m.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(m.get_vertexset().begin()), e(m.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!m.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(m.get_edgeset().begin()), e(m.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	oss << ' ' << m.get_text();
	add_resulttext(oss.str() + "\n", (m.get_type() <= TrafficFlowRestrictions::Message::type_warning) ? m_txttagmsgwarn : m_txttagmsginfo);
}

void CFMUValidate::MainWindow::add_resulttext(const TrafficFlowRestrictions::Result& res)
{
	// print Messages
	for (unsigned int i = 0; i < res.messages_size(); ++i)
		add_resulttext(res.get_message(i));
	add_resulttext(res ? "OK\n" : "FAIL\n", m_txttagrule);
	for (unsigned int i = 0; i < res.size(); ++i)
		add_resulttext(res[i]);
}

void CFMUValidate::MainWindow::add_resulttext(const TrafficFlowRestrictions::RuleResult& rule)
{
	std::ostringstream oss;
	oss << "Rule: " << rule.get_rule() << ' ';
	switch (rule.get_codetype()) {
	case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
		oss << "Mandatory";
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
		oss << "Forbidden";
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_closed:
		oss << "ClosedForCruising";
		break;

	default:
		oss << '?';
		break;
	}
	if (rule.is_dct() || rule.is_unconditional() || rule.is_routestatic() || rule.is_mandatoryinbound() || rule.is_disabled()) {
		oss << " (";
		if(rule.is_disabled())
			oss << '-';
		if(rule.is_dct())
			oss << 'D';
		if (rule.is_unconditional())
			oss << 'U';
		if (rule.is_routestatic())
			oss << 'S';
		if (rule.is_mandatoryinbound())
			oss << 'I';
		oss << ')';
	}
	if (!rule.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_vertexset().begin()), e(rule.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!rule.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!rule.get_reflocset().empty()) {
		oss << " L:";
		bool subseq(false);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_reflocset().begin()), e(rule.get_reflocset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	add_resulttext(oss.str() + "\n", m_txttagrule);
	add_resulttext("Desc: " + rule.get_desc() + "\n", m_txttagdesc);
	add_resulttext("OprGoal: " + rule.get_oprgoal() + "\n", m_txttagoprgoal);
	add_resulttext("Cond: " + rule.get_cond() + "\n", m_txttagcond);
	for (unsigned int i = 0; i < rule.size(); ++i) {
		const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[i]);
		{
			std::ostringstream oss;
			oss << "  Alternative " << i;
			if (!alt.get_vertexset().empty()) {
				oss << " V:";
				bool subseq(false);
				for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_vertexset().begin()), e(alt.get_vertexset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			if (!alt.get_edgeset().empty()) {
				oss << " E:";
				bool subseq(false);
				for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_edgeset().begin()), e(alt.get_edgeset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			add_resulttext(oss.str() + "\n", m_txttagrestrict);
		}
		for (unsigned int j = 0; j < alt.size(); ++j) {
			const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& el(alt[j]);
			std::ostringstream oss;
			oss << "    " << el.get_type_string();
			switch (el.get_type()) {
			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway:
				oss << ' ' << el.get_ident();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_sid:
				oss << ' ' << el.get_ident() << ' ' << el.get_airport();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_star:
				oss << ' ' << el.get_ident() << ' ' << el.get_airport();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airspace:
				oss << ' ' << el.get_ident() << '/'
				    << AirspacesDb::Airspace::get_class_string(el.get_bdryclass(), AirspacesDb::Airspace::typecode_ead);
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct:
			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point:
			default:
				break;
			}
			oss << " F" << std::setw(3) << std::setfill('0') << ((el.get_lower_alt() + 50) / 100);
			if (el.is_lower_alt_valid())
				oss << '*';
			oss << "...F"
			    << std::setw(3) << std::setfill('0') << ((std::min((int32_t)el.get_upper_alt(), (int32_t)99900) + 50) / 100);
			if (el.is_upper_alt_valid())
				oss << '*';
			add_resulttext(oss.str() + "\n", m_txttagrestrict);
			if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
			    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct ||
			    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point) {
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point0());
				std::ostringstream oss;
				oss << "      " << pt.get_ident() << ' ' << pt.get_coord().get_lat_str2() << ' '
				    << pt.get_coord().get_lon_str2() << ' ' << to_str(pt.get_type());
				add_resulttext(oss.str() + "\n", m_txttagrestrict);
			}
			if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
			    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct) {
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point1());
				std::ostringstream oss;
				oss << "      " << pt.get_ident() << ' ' << pt.get_coord().get_lat_str2() << ' '
				    << pt.get_coord().get_lon_str2() << ' ' << to_str(pt.get_type());
				add_resulttext(oss.str() + "\n", m_txttagrestrict);
			}
		}
	}
}

void CFMUValidate::MainWindow::print_tracerules(void)
{
	std::string r;
	bool subseq(false);
	for (TrafficFlowRestrictions::tracerules_const_iterator_t ti(m_tfr.tracerules_begin()), te(m_tfr.tracerules_end()); ti != te; ++ti) {
		if (subseq)
			r += ",";
		subseq = true;
		r += *ti;
	}
	add_resulttext("trace:" + r + "\n", m_txttagmsginfo);
}

void CFMUValidate::MainWindow::print_disabledrules(void)
{
	std::string r;
	bool subseq(false);
	for (TrafficFlowRestrictions::disabledrules_const_iterator_t ti(m_tfr.disabledrules_begin()), te(m_tfr.disabledrules_end()); ti != te; ++ti) {
		if (subseq)
			r += ",";
		subseq = true;
		r += *ti;
	}
	add_resulttext("disabled:" + r + "\n", m_txttagmsginfo);
}

void CFMUValidate::MainWindow::validate_local_clicked(void)
{
	if (m_textviewresults) {
		Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewresults->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (!m_textviewfplan)
		return;
	Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewfplan->get_buffer());
	if (!buf)
		return;
	const Glib::ustring& txt(buf->get_text());
	if (txt.empty())
		return;
	if (!m_tfrloaded) {
		// first open the databases
		delete m_engine;
		m_engine = new Engine(m_dir_main, m_auxdbmode, m_dir_aux, false, false);
		std::string dir_aux1 = m_engine->get_aux_dir(m_auxdbmode, m_dir_aux);
		m_airportdb.open(m_dir_main);
		if (!dir_aux1.empty() && m_airportdb.is_dbfile_exists(dir_aux1))
			m_airportdb.attach(dir_aux1);
		if (m_airportdb.is_aux_attached())
			std::cerr << "Auxillary airports database attached" << std::endl;
		m_navaiddb.open(m_dir_main);
		if (!dir_aux1.empty() && m_navaiddb.is_dbfile_exists(dir_aux1))
			m_navaiddb.attach(dir_aux1);
		if (m_navaiddb.is_aux_attached())
			std::cerr << "Auxillary navaids database attached" << std::endl;
		m_waypointdb.open(m_dir_main);
		if (!dir_aux1.empty() && m_waypointdb.is_dbfile_exists(dir_aux1))
			m_waypointdb.attach(dir_aux1);
		if (m_waypointdb.is_aux_attached())
			std::cerr << "Auxillary waypoints database attached" << std::endl;
		m_airwaydb.open(m_dir_main);
		if (!dir_aux1.empty() && m_airwaydb.is_dbfile_exists(dir_aux1))
			m_airwaydb.attach(dir_aux1);
		if (m_airwaydb.is_aux_attached())
			std::cerr << "Auxillary airways database attached" << std::endl;
		m_airspacedb.open(m_dir_main);
		if (!dir_aux1.empty() && m_airspacedb.is_dbfile_exists(dir_aux1))
			m_airspacedb.attach(dir_aux1);
		if (m_airspacedb.is_aux_attached())
			std::cerr << "Auxillary airspaces database attached" << std::endl;
		if (m_textviewresults) {
			Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewresults->get_buffer());
			if (buf)
				buf->set_text("");
		}
		Glib::TimeVal tv;
		tv.assign_current_time();
		m_tfr.start_add_rules();
		std::vector<std::string> tfrfnames;
		bool ok(false);
		{
			std::string fname(Glib::build_filename(dir_aux1, "eRAD.bin"));
			if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
				add_resulttext("Loading traffic flow restrictions " + fname + "\n", m_txttagmsginfo);
				std::vector<TrafficFlowRestrictions::Message> msg;
				ok = m_tfr.add_binary_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
				for (unsigned int i = 0; i < msg.size(); ++i)
					add_resulttext(msg[i]);
				if (ok)
					tfrfnames.push_back(fname);
			}
		}
		if (!ok) {
			std::string fname(Glib::build_filename(dir_aux1, "eRAD.xml"));
			if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
				add_resulttext("Loading traffic flow restrictions " + fname + "\n", m_txttagmsginfo);
				std::vector<TrafficFlowRestrictions::Message> msg;
				ok = m_tfr.add_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
				for (unsigned int i = 0; i < msg.size(); ++i)
					add_resulttext(msg[i]);
				if (ok)
					tfrfnames.push_back(fname);
			}
		}
		bool okloc(false);
		{
			std::string fname(Glib::build_filename(dir_aux1, "eRAD_local.bin"));
			if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
				add_resulttext("Loading traffic flow restrictions " + fname + "\n", m_txttagmsginfo);
				std::vector<TrafficFlowRestrictions::Message> msg;
				okloc = m_tfr.add_binary_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
				for (unsigned int i = 0; i < msg.size(); ++i)
					add_resulttext(msg[i]);
				if (okloc)
					tfrfnames.push_back(fname);
			}
		}
		if (!okloc) {
			std::string fname(Glib::build_filename(dir_aux1, "eRAD_local.xml"));
			if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
				add_resulttext("Loading traffic flow restrictions " + fname + "\n", m_txttagmsginfo);
				std::vector<TrafficFlowRestrictions::Message> msg;
				okloc = m_tfr.add_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
				for (unsigned int i = 0; i < msg.size(); ++i)
					add_resulttext(msg[i]);
				if (okloc)
					tfrfnames.push_back(fname);
			}
		}
		bool okuser(false);
		{
			std::string fname(Glib::build_filename(FPlan::get_userdbpath(), "eRAD.bin"));
			if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
				add_resulttext("Loading traffic flow restrictions " + fname + "\n", m_txttagmsginfo);
				std::vector<TrafficFlowRestrictions::Message> msg;
				okuser = m_tfr.add_binary_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
				for (unsigned int i = 0; i < msg.size(); ++i)
					add_resulttext(msg[i]);
				if (okuser)
					tfrfnames.push_back(fname);
			}
		}
		if (!okuser) {
			std::string fname(Glib::build_filename(FPlan::get_userdbpath(), "eRAD.xml"));
			if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
				add_resulttext("Loading traffic flow restrictions " + fname + "\n", m_txttagmsginfo);
				std::vector<TrafficFlowRestrictions::Message> msg;
				okuser = m_tfr.add_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
				for (unsigned int i = 0; i < msg.size(); ++i)
					add_resulttext(msg[i]);
				if (okuser)
					tfrfnames.push_back(fname);
			}
		}
		{
			Glib::TimeVal tv1;
			tv1.assign_current_time();
			tv = tv1 - tv;
		}
		if (!ok) {
		        add_resulttext("Cannot load traffic flow restrictions " + Glib::build_filename(dir_aux1, "eRAD.bin") + ", " +
				       Glib::build_filename(dir_aux1, "eRAD.xml") + "\n", m_txttagmsgwarn);
			return;
		}
		m_tfr.end_add_rules();
		if (ok || okloc || okuser) {
			add_resulttext("Origin:    " + m_tfr.get_origin() + "\n", m_txttagmsginfo);
			add_resulttext("Created:   " + m_tfr.get_created() + "\n", m_txttagmsginfo);
			add_resulttext("Effective: " + m_tfr.get_effective() + "\n", m_txttagmsginfo);
			m_tfr.reset_rules();
			{
				std::ostringstream oss;
				oss << m_tfr.count_rules() << " rules loaded traffic flow restrictions ";
				for (std::vector<std::string>::const_iterator i(tfrfnames.begin()), e(tfrfnames.end()); i != e; ) {
					oss << *i;
					++i;
					if (i == e)
						break;
					oss << ", ";
				}
				oss << ", created " << m_tfr.get_created() << ", effective " << m_tfr.get_effective() << ", "
				    << std::fixed << std::setprecision(3) << tv.as_double() << 's' << std::endl;
				add_resulttext(oss.str(), m_txttagmsginfo);
			}
		}
		{
			std::string fname(Glib::build_filename(dir_aux1, "eRAD_rules.xml"));
			if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
				if (m_tfr.load_disabled_trace_rules_file(fname))
					add_resulttext("Loaded global disabled/trace rules from " + fname + "\n", m_txttagmsginfo);
			}
		}
		{
			std::string fname(Glib::build_filename(FPlan::get_userdbpath(), "eRAD_rules.xml"));
			if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
				if (m_tfr.load_disabled_trace_rules_file(fname))
					add_resulttext("Loaded user disabled/trace rules from " + fname + "\n", m_txttagmsginfo);
			}
		}
		if (true) {
			Glib::TimeVal tv;
			tv.assign_current_time();
			m_tfr.simplify_rules_typeofflight('G');
			m_tfr.simplify_rules_mil(false);
			m_tfr.simplify_rules_gat(true);
			{
				Glib::TimeVal tv1;
				tv1.assign_current_time();
				tv = tv1 - tv;
			}
			std::ostringstream oss;
			oss << m_tfr.count_rules() << " rules after fixing type of flight G CIV GAT, "
			    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
		        add_resulttext(oss.str() + "\n", m_txttagmsginfo);
		}
		m_tfrloaded = true;
	}
	if (!txt.compare(0, 6, "trace?")) {
		print_tracerules();
		return;
	}
	if (!txt.compare(0, 6, "trace:") || !txt.compare(0, 6, "trace+")) {
		Glib::ustring txt1(txt);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		m_tfr.tracerules_add(txt1.substr(6));
		print_tracerules();
		return;
	}
	if (!txt.compare(0, 6, "trace-")) {
		Glib::ustring txt1(txt);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		m_tfr.tracerules_erase(txt1.substr(6));
		print_tracerules();
		return;
	}
	if (!txt.compare(0, 6, "trace*")) {
		m_tfr.tracerules_clear();
		print_tracerules();
		return;
	}
	if (!txt.compare(0, 9, "disabled?")) {
		print_disabledrules();
		return;
	}
	if (!txt.compare(0, 9, "disabled:") || !txt.compare(0, 9, "disabled+")) {
		Glib::ustring txt1(txt);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		m_tfr.disabledrules_add(txt1.substr(9));
		print_disabledrules();
		return;
	}
	if (!txt.compare(0, 9, "disabled-")) {
		Glib::ustring txt1(txt);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		m_tfr.disabledrules_erase(txt1.substr(9));
		print_disabledrules();
		return;
	}
	if (!txt.compare(0, 9, "disabled*")) {
		m_tfr.disabledrules_clear();
		print_disabledrules();
		return;
	}

	FPlanRoute fpl(*(FPlan *)0);
	char type_of_flight('G');
	std::string acft_type;
	std::string equipment;
	Aircraft::pbn_t pbn(Aircraft::pbn_none);
	{
		IcaoFlightPlan icaofpl(*m_engine);
	        IcaoFlightPlan::errors_t err(icaofpl.parse(txt, false));
		if (!err.empty()) {
			std::string er("EFPL: ");
			bool subseq(false);
			for (IcaoFlightPlan::errors_t::const_iterator i(err.begin()), e(err.end()); i != e; ++i) {
				if (subseq)
					er += "; ";
				subseq = true;
				er += *i;
			}
			add_resulttext(er + "\n", m_txttagmsgwarn);
			return;
		}
		icaofpl.set_route(fpl);
		type_of_flight = icaofpl.get_flighttype();
		acft_type = icaofpl.get_aircrafttype();
		equipment = icaofpl.get_equipment();
		pbn = icaofpl.get_pbn();
		if (true) {
			std::ostringstream oss;
			oss << "I: FPL Parsed, " << fpl.get_nrwpt() << " waypoints";
			add_resulttext(oss.str() + "\n", m_txttagfpl);
		}
	}
	// print Parsed Flight Plan
	if (false) {
		for (unsigned int i = 0; i < fpl.get_nrwpt(); ++i) {
			const FPlanWaypoint& wpt(fpl[i]);
			std::ostringstream oss;
			oss << "PFPL[" << i << "]: " << wpt.get_icao() << '/' << wpt.get_name() << ' ';
			switch (wpt.get_pathcode()) {
			case FPlanWaypoint::pathcode_sid:
                        case FPlanWaypoint::pathcode_star:
			case FPlanWaypoint::pathcode_airway:
				oss << wpt.get_pathname();
				break;

                        case FPlanWaypoint::pathcode_directto:
				oss << "DCT";
				break;

			default:
				oss << '-';
				break;
			}
			oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
			    << " D" << std::fixed << std::setprecision(1) << wpt.get_dist_nmi()
			    << " T" << std::fixed << std::setprecision(0) << wpt.get_truetrack_deg()
			    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
			    << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
			add_resulttext(oss.str() + "\n", m_txttagfpl);
		}
	}
	TrafficFlowRestrictions::Result res(m_tfr.check_fplan(fpl, type_of_flight, acft_type, equipment, pbn,
							      m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb));
	// print Flight Plan
	{
		{
			std::ostringstream oss;
			oss << "FR: " << res.get_aircrafttype() << ' ' << res.get_aircrafttypeclass() << ' '
			    << res.get_equipment() << " PBN/" << res.get_pbn_string() << ' ' << res.get_typeofflight();
			add_resulttext(oss.str() + "\n", m_txttagfpl);
		}
		const TrafficFlowRestrictions::Fpl& fpl(res.get_fplan());
		for (unsigned int i = 0; i < fpl.size(); ++i) {
			const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
			std::ostringstream oss;
			oss << "F[" << i << '/' << wpt.get_wptnr() << "]: " << wpt.get_ident() << ' ' << wpt.get_type_string() << ' ';
			switch (wpt.get_pathcode()) {
			case FPlanWaypoint::pathcode_sid:
                        case FPlanWaypoint::pathcode_star:
			case FPlanWaypoint::pathcode_airway:
				oss << wpt.get_pathname();
				break;

                        case FPlanWaypoint::pathcode_directto:
				oss << "DCT";
				break;

			default:
				oss << '-';
				break;
			}
			oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
			    << " D" << std::fixed << std::setprecision(1) << wpt.get_dist_nmi()
			    << " T" << std::fixed << std::setprecision(0) << wpt.get_truetrack_deg()
			    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
			    << ' ' << wpt.get_icao() << '/' << wpt.get_name() << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
			{
				time_t t(wpt.get_time_unix());
				struct tm tm;
				if (gmtime_r(&t, &tm)) {
					oss << ' ' << std::setw(2) << std::setfill('0') << tm.tm_hour
					    << ':' << std::setw(2) << std::setfill('0') << tm.tm_min;
					if (t >= 48 * 60 * 60)
						oss << ' ' << std::setw(4) << std::setfill('0') << (tm.tm_year + 1900)
						    << '-' << std::setw(2) << std::setfill('0') << (tm.tm_mon + 1)
						    << '-' << std::setw(2) << std::setfill('0') << tm.tm_mday;
				}
			}
			add_resulttext(oss.str() + "\n", m_txttagfpl);
		}
	}
	add_resulttext(res);
}

void CFMUValidate::MainWindow::print_adr_tracerules(void)
{
	std::string r;
	bool subseq(false);
	uint64_t tm(time(0));
	for (ADR::RestrictionEval::rules_t::const_iterator ri(m_reval.get_rules().begin()), re(m_reval.get_rules().end()); ri != re; ++ri) {
		const ADR::Object::ptr_t& p(*ri);
		const ADR::FlightRestrictionTimeSlice& ts(p->operator()(tm).as_flightrestriction());
		if (!ts.is_valid())
			continue;
		if (!ts.is_trace())
			continue;
		if (subseq)
			r += ",";
		subseq = true;
		r += ts.get_ident();
	}
	add_resulttext("trace:" + r + "\n", m_txttagmsginfo);
}

void CFMUValidate::MainWindow::print_adr_disabledrules(void)
{
	std::string r;
	bool subseq(false);
	uint64_t tm(time(0));
	for (ADR::RestrictionEval::rules_t::const_iterator ri(m_reval.get_rules().begin()), re(m_reval.get_rules().end()); ri != re; ++ri) {
		const ADR::Object::ptr_t& p(*ri);
		const ADR::FlightRestrictionTimeSlice& ts(p->operator()(tm).as_flightrestriction());
		if (!ts.is_valid())
			continue;
		if (ts.is_enabled())
			continue;
		if (subseq)
			r += ",";
		subseq = true;
		r += ts.get_ident();
	}
	add_resulttext("disabled:" + r + "\n", m_txttagmsginfo);
}

void CFMUValidate::MainWindow::add_resulttext(const ADR::Message& m)
{
	std::ostringstream oss;
	oss << m.get_type_char() << ':';
	{
		const std::string& id(m.get_ident());
		if (!id.empty())
			oss << " R:" << id;
	}
	if (!m.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (ADR::Message::set_t::const_iterator i(m.get_vertexset().begin()), e(m.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!m.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (ADR::Message::set_t::const_iterator i(m.get_edgeset().begin()), e(m.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	oss << ' ' << m.get_text();
	add_resulttext(oss.str() + "\n", (m.get_type() <= ADR::Message::type_warning) ? m_txttagmsgwarn : m_txttagmsginfo);
}

void CFMUValidate::MainWindow::add_resulttext(const ADR::RestrictionResult& res)
{
	const ADR::FlightRestrictionTimeSlice& tsrule(res.get_rule()->operator()(m_reval.get_departuretime()).as_flightrestriction());
	if (!tsrule.is_valid()) {
		add_resulttext("Rule: ??\n", m_txttagrule);
		return;
	}
	std::ostringstream oss;
	oss << "Rule: " << tsrule.get_ident() << ' ';
	switch (tsrule.get_type()) {
	case ADR::FlightRestrictionTimeSlice::type_mandatory:
		oss << "Mandatory";
		break;

	case ADR::FlightRestrictionTimeSlice::type_forbidden:
		oss << "Forbidden";
		break;

	case ADR::FlightRestrictionTimeSlice::type_closed:
		oss << "ClosedForCruising";
		break;

	case ADR::FlightRestrictionTimeSlice::type_allowed:
		oss << "Allowed";
		break;

	default:
		oss << '?';
		break;
	}
	if (tsrule.is_dct() || tsrule.is_unconditional() || tsrule.is_routestatic() ||
	    tsrule.is_mandatoryinbound() || tsrule.is_mandatoryoutbound() || !tsrule.is_enabled()) {
		oss << " (";
		if (!tsrule.is_enabled())
			oss << '-';
		if (tsrule.is_dct())
			oss << 'D';
		if (tsrule.is_unconditional())
			oss << 'U';
		if (tsrule.is_routestatic())
			oss << 'S';
		if (tsrule.is_mandatoryinbound())
			oss << 'I';
		if (tsrule.is_mandatoryoutbound())
			oss << 'O';
		oss << ')';
	}
	if (!res.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (ADR::RestrictionResult::set_t::const_iterator i(res.get_vertexset().begin()), e(res.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!res.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (ADR::RestrictionResult::set_t::const_iterator i(res.get_edgeset().begin()), e(res.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!res.has_refloc())
		oss << " L:" << res.get_refloc();
	add_resulttext(oss.str() + "\n", m_txttagrule);
	add_resulttext("Desc: " + tsrule.get_instruction() + "\n", m_txttagdesc);
	//add_resulttext("OprGoal: \n", m_txttagoprgoal);
	if (tsrule.get_condition())
		add_resulttext("Cond: " + tsrule.get_condition()->to_str(m_reval.get_departuretime()) + "\n", m_txttagcond);
	for (unsigned int i = 0, n = tsrule.get_restrictions().size(); i < n; ++i) {
		const ADR::RestrictionSequence& rs(tsrule.get_restrictions()[i]);
		const ADR::RestrictionSequenceResult& rsr(res[i]);
		{
			std::ostringstream oss;
			oss << "  Alternative " << i;
			if (!rsr.get_vertexset().empty()) {
				oss << " V:";
				bool subseq(false);
				for (ADR::RestrictionResult::set_t::const_iterator i(rsr.get_vertexset().begin()), e(rsr.get_vertexset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			if (!rsr.get_edgeset().empty()) {
				oss << " E:";
				bool subseq(false);
				for (ADR::RestrictionResult::set_t::const_iterator i(rsr.get_edgeset().begin()), e(rsr.get_edgeset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			add_resulttext(oss.str() + "\n", m_txttagrestrict);
		}
		add_resulttext(rs.to_str(m_reval.get_departuretime()) + "\n", m_txttagrestrict);
	}
}

void CFMUValidate::MainWindow::add_resulttext(const ADR::Object::ptr_t& p, uint64_t tm)
{
	ADR::FlightRestrictionTimeSlice& tsrule(p->operator()(tm).as_flightrestriction());
	if (!tsrule.is_valid()) {
		add_resulttext("Rule: ??\n", m_txttagrule);
		return;
	}
	std::ostringstream oss;
	oss << "Rule: " << tsrule.get_ident() << ' ';
	switch (tsrule.get_type()) {
	case ADR::FlightRestrictionTimeSlice::type_mandatory:
		oss << "Mandatory";
		break;

	case ADR::FlightRestrictionTimeSlice::type_forbidden:
		oss << "Forbidden";
		break;

	case ADR::FlightRestrictionTimeSlice::type_closed:
		oss << "ClosedForCruising";
		break;

	case ADR::FlightRestrictionTimeSlice::type_allowed:
		oss << "Allowed";
		break;

	default:
		oss << '?';
		break;
	}
	if (tsrule.is_dct() || tsrule.is_unconditional() || tsrule.is_routestatic() || tsrule.is_mandatoryinbound() || !tsrule.is_enabled()) {
		oss << " (";
		if (!tsrule.is_enabled())
			oss << '-';
		if (tsrule.is_dct())
			oss << 'D';
		if (tsrule.is_unconditional())
			oss << 'U';
		if (tsrule.is_routestatic())
			oss << 'S';
		if (tsrule.is_mandatoryinbound())
			oss << 'I';
		oss << ')';
	}
	add_resulttext(oss.str() + "\n", m_txttagrule);
	add_resulttext("Desc: " + tsrule.get_instruction() + "\n", m_txttagdesc);
	//add_resulttext("OprGoal: \n", m_txttagoprgoal);
	if (tsrule.get_condition())
		add_resulttext("Cond: " + tsrule.get_condition()->to_str(m_reval.get_departuretime()) + "\n", m_txttagcond);
	for (unsigned int i = 0, n = tsrule.get_restrictions().size(); i < n; ++i) {
		const ADR::RestrictionSequence& rs(tsrule.get_restrictions()[i]);
		std::ostringstream oss;
		oss << "  Alternative " << i << ' ' << rs.to_str(m_reval.get_departuretime());
		add_resulttext(oss.str() + "\n", m_txttagrestrict);
	}
}

void CFMUValidate::MainWindow::validate_adr_clicked(void)
{
	if (m_textviewresults) {
		Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewresults->get_buffer());
		if (buf)
			buf->set_text("");
	}
	if (!m_textviewfplan)
		return;
	Glib::RefPtr<Gtk::TextBuffer> buf(m_textviewfplan->get_buffer());
	if (!buf)
		return;
	const Glib::ustring& txt(buf->get_text());
	if (txt.empty())
		return;
	if (m_reval.get_rules().empty()) {
		// first open the databases
		m_adrdb.set_path(m_dir_aux);
		m_reval.set_db(&m_adrdb);
		Glib::TimeVal tv;
		tv.assign_current_time();
		m_reval.load_rules();
		{
			Glib::TimeVal tv1;
			tv1.assign_current_time();
			tv = tv1 - tv;
		}
		std::ostringstream oss;
		oss << "Loaded " << m_reval.count_rules() << " Restrictions, "
		    << std::fixed << std::setprecision(3) << tv.as_double() << 's' << std::endl;
		add_resulttext(oss.str(), m_txttagmsginfo);
	}
	if (!txt.compare(0, 6, "trace?")) {
		print_adr_tracerules();
		return;
	}
	if (!txt.compare(0, 6, "trace:") || !txt.compare(0, 6, "trace+")) {
		Glib::ustring txt1(txt);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		txt1 = txt1.substr(6);
		uint64_t tm(time(0));
		for (ADR::RestrictionEval::rules_t::const_iterator ri(m_reval.get_rules().begin()), re(m_reval.get_rules().end()); ri != re; ++ri) {
			const ADR::Object::ptr_t& p(*ri);
			ADR::FlightRestrictionTimeSlice& ts(p->operator()(tm).as_flightrestriction());
			if (!ts.is_valid())
				continue;
			if (ts.get_ident() != txt1)
				continue;
			ts.set_trace(true);
		}
		print_adr_tracerules();
		return;
	}
	if (!txt.compare(0, 6, "trace-")) {
		Glib::ustring txt1(txt);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		txt1 = txt1.substr(6);
		uint64_t tm(time(0));
		for (ADR::RestrictionEval::rules_t::const_iterator ri(m_reval.get_rules().begin()), re(m_reval.get_rules().end()); ri != re; ++ri) {
			const ADR::Object::ptr_t& p(*ri);
			ADR::FlightRestrictionTimeSlice& ts(p->operator()(tm).as_flightrestriction());
			if (!ts.is_valid())
				continue;
			if (ts.get_ident() != txt1)
				continue;
			ts.set_trace(false);
		}
		print_adr_tracerules();
		return;
	}
	if (!txt.compare(0, 6, "trace*")) {
		uint64_t tm(time(0));
		for (ADR::RestrictionEval::rules_t::const_iterator ri(m_reval.get_rules().begin()), re(m_reval.get_rules().end()); ri != re; ++ri) {
			const ADR::Object::ptr_t& p(*ri);
			ADR::FlightRestrictionTimeSlice& ts(p->operator()(tm).as_flightrestriction());
			if (!ts.is_valid())
				continue;
			ts.set_trace(false);
		}
		print_adr_tracerules();
		return;
	}
	if (!txt.compare(0, 9, "disabled?")) {
		print_adr_disabledrules();
		return;
	}
	if (!txt.compare(0, 9, "disabled:") || !txt.compare(0, 9, "disabled+")) {
		Glib::ustring txt1(txt);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		txt1 = txt1.substr(9);
		uint64_t tm(time(0));
		for (ADR::RestrictionEval::rules_t::const_iterator ri(m_reval.get_rules().begin()), re(m_reval.get_rules().end()); ri != re; ++ri) {
			const ADR::Object::ptr_t& p(*ri);
			ADR::FlightRestrictionTimeSlice& ts(p->operator()(tm).as_flightrestriction());
			if (!ts.is_valid())
				continue;
			if (ts.get_ident() != txt1)
				continue;
			ts.set_enabled(false);
		}
		print_adr_disabledrules();
		return;
	}
	if (!txt.compare(0, 9, "disabled-")) {
		Glib::ustring txt1(txt);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		txt1 = txt1.substr(9);
		uint64_t tm(time(0));
		for (ADR::RestrictionEval::rules_t::const_iterator ri(m_reval.get_rules().begin()), re(m_reval.get_rules().end()); ri != re; ++ri) {
			const ADR::Object::ptr_t& p(*ri);
			ADR::FlightRestrictionTimeSlice& ts(p->operator()(tm).as_flightrestriction());
			if (!ts.is_valid())
				continue;
			if (ts.get_ident() != txt1)
				continue;
			ts.set_enabled(true);
		}
		print_adr_disabledrules();
		return;
	}
	if (!txt.compare(0, 9, "disabled*")) {
		m_tfr.disabledrules_clear();
		uint64_t tm(time(0));
		for (ADR::RestrictionEval::rules_t::const_iterator ri(m_reval.get_rules().begin()), re(m_reval.get_rules().end()); ri != re; ++ri) {
			const ADR::Object::ptr_t& p(*ri);
			ADR::FlightRestrictionTimeSlice& ts(p->operator()(tm).as_flightrestriction());
			if (!ts.is_valid())
				continue;
			ts.set_enabled(true);
		}
		print_adr_disabledrules();
		return;
	}

	if (!txt.compare(0, 4, "fpr?")) {
		add_resulttext(m_honourprofilerules ? "fpr+\n" : "fpr-\n", m_txttagmsginfo);
		return;
	}
	if (!txt.compare(0, 4, "fpr+")) {
		m_honourprofilerules = true;
		add_resulttext("fpr+\n", m_txttagmsginfo);
		return;
	}
	if (!txt.compare(0, 4, "fpr-")) {
		m_honourprofilerules = false;
		add_resulttext("fpr-\n", m_txttagmsginfo);
		return;
	}
	{
		ADR::FlightPlan adrfpl;
		ADR::FlightPlan::errors_t err(adrfpl.parse(m_adrdb, txt, true));
		if (!err.empty()) {
			std::string er("EFPL: ");
			bool subseq(false);
			for (ADR::FlightPlan::errors_t::const_iterator i(err.begin()), e(err.end()); i != e; ++i) {
				if (subseq)
					er += "; ";
				subseq = true;
				er += *i;
			}
			add_resulttext(er + "\n", m_txttagmsgwarn);
			return;
		}
		m_reval.set_fplan(adrfpl);
		if (true) {
			std::ostringstream oss;
			oss << "I: FPL Parsed, " << adrfpl.size() << " waypoints";
			add_resulttext(oss.str() + "\n", m_txttagfpl);
		}
		// print Parsed Flight Plan
		if (false) {
			for (unsigned int i(0), n(adrfpl.size()); i < n; ++i) {
				const ADR::FPLWaypoint& wpt(adrfpl[i]);
				std::ostringstream oss;
				oss << "PFPL[" << i << "]: " << wpt.get_icao() << '/' << wpt.get_name() << ' ';
				switch (wpt.get_pathcode()) {
				case FPlanWaypoint::pathcode_sid:
				case FPlanWaypoint::pathcode_star:
				case FPlanWaypoint::pathcode_airway:
					oss << wpt.get_pathname();
					break;

				case FPlanWaypoint::pathcode_directto:
					oss << "DCT";
					break;

				default:
					oss << '-';
					break;
				}
				oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
				    << " D" << std::fixed << std::setprecision(1) << wpt.get_dist_nmi()
				    << " T" << std::fixed << std::setprecision(0) << wpt.get_truetrack_deg()
				    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
				    << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
				if (wpt.get_ptobj())
					oss << " ptobj " << wpt.get_ptobj()->get_uuid();
				if (wpt.get_pathobj())
					oss << " pathobj " << wpt.get_pathobj()->get_uuid();
				if (wpt.is_expanded())
					oss << " (E)";
				add_resulttext(oss.str() + "\n", m_txttagfpl);
			}
		}
	}
	bool res(m_reval.check_fplan(m_honourprofilerules));
	// print Flight Plan
	{
		ADR::FlightPlan f(m_reval.get_fplan());
 		{
			std::ostringstream oss;
			oss << "FR: " << f.get_aircrafttype() << ' ' << Aircraft::get_aircrafttypeclass(f.get_aircrafttype()) << ' '
			    << f.get_equipment() << " PBN/" << f.get_pbn_string() << ' ' << f.get_flighttype();
			add_resulttext(oss.str() + "\n", m_txttagfpl);
		}
		for (unsigned int i(0), n(f.size()); i < n; ++i) {
			const ADR::FPLWaypoint& wpt(f[i]);
			std::ostringstream oss;
			oss << "F[" << i << '/' << wpt.get_wptnr() << "]: " << wpt.get_ident() << ' ' << wpt.get_type_string() << ' ';
			switch (wpt.get_pathcode()) {
			case FPlanWaypoint::pathcode_sid:
                        case FPlanWaypoint::pathcode_star:
			case FPlanWaypoint::pathcode_airway:
				oss << wpt.get_pathname();
				break;

                        case FPlanWaypoint::pathcode_directto:
				oss << "DCT";
				break;

			default:
				oss << '-';
				break;
			}
			oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
			    << " D" << std::fixed << std::setprecision(1) << wpt.get_dist_nmi()
			    << " T" << std::fixed << std::setprecision(0) << wpt.get_truetrack_deg()
			    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
			    << ' ' << wpt.get_icao() << '/' << wpt.get_name() << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
			{
				time_t t(wpt.get_time_unix());
				struct tm tm;
				if (gmtime_r(&t, &tm)) {
					oss << ' ' << std::setw(2) << std::setfill('0') << tm.tm_hour
					    << ':' << std::setw(2) << std::setfill('0') << tm.tm_min;
					if (t >= 48 * 60 * 60)
						oss << ' ' << std::setw(4) << std::setfill('0') << (tm.tm_year + 1900)
						    << '-' << std::setw(2) << std::setfill('0') << (tm.tm_mon + 1)
						    << '-' << std::setw(2) << std::setfill('0') << tm.tm_mday;
				}
			}
			if (wpt.get_ptobj())
				oss << " ptobj " << wpt.get_ptobj()->get_uuid();
			if (wpt.get_pathobj())
				oss << " pathobj " << wpt.get_pathobj()->get_uuid();
			if (wpt.is_expanded())
				oss << " (E)";
			add_resulttext(oss.str() + "\n", m_txttagfpl);
		}
	}
	for (ADR::RestrictionEval::messages_const_iterator mi(m_reval.messages_begin()), me(m_reval.messages_end()); mi != me; ++mi)
		add_resulttext(*mi);
	add_resulttext(res ? "OK\n" : "FAIL\n", m_txttagrule);
	for (ADR::RestrictionEval::results_const_iterator ri(m_reval.results_begin()), re(m_reval.results_end()); ri != re; ++ri)
		add_resulttext(*ri);
}


CFMUValidate::CFMUValidate(int& argc, char**& argv, Glib::OptionContext& option_context)
	: Gtk::Main(argc, argv, option_context), m_mainwindow(0)
{
	m_builder = load_glade_file("cfmuvalidate" UIEXT);
        m_builder->get_widget_derived("cfmuvalidatewindow", m_mainwindow);
	if (m_mainwindow)
		m_mainwindow->set_app(this);
	{
		guint nsigids(0);
		guint *sigids(g_signal_list_ids(WEBKIT_TYPE_DOM_HTML_ELEMENT, &nsigids));
		std::cerr << "DOM HTML Element signal list: " << nsigids << std::endl;
		for (guint i(0); i < nsigids; ++i) {
			GSignalQuery q;
			g_signal_query(sigids[i], &q);
			std::cerr << "  " << q.signal_name << std::endl;
		}
	}
}

CFMUValidate::~CFMUValidate()
{
	delete m_mainwindow;
}

void CFMUValidate::run(void)
{
	if (m_mainwindow)
		Gtk::Main::run(*m_mainwindow);
}

void CFMUValidate::set_dbdirs(const std::string& dir_main, Engine::auxdb_mode_t auxdbmode, const std::string& dir_aux)
{
	if (!m_mainwindow)
		return;
	m_mainwindow->set_dbdirs(dir_main, auxdbmode, dir_aux);
}

int main(int argc, char *argv[])
{
        try {
                std::string dir_main(""), dir_aux("");
                Engine::auxdb_mode_t auxdbmode = Engine::auxdb_prefs;
                bool dis_aux(false);
                Glib::init();
                Glib::OptionGroup optgroup("cfmuvalidate", "CFMU Validate Application Options", "Options controlling the CFMU Validate Application");
                Glib::OptionEntry optmaindir, optauxdir, optdisableaux;
                optmaindir.set_short_name('m');
                optmaindir.set_long_name("maindir");
                optmaindir.set_description("Directory containing the main database");
                optmaindir.set_arg_description("DIR");
                optauxdir.set_short_name('a');
                optauxdir.set_long_name("auxdir");
                optauxdir.set_description("Directory containing the auxilliary database");
                optauxdir.set_arg_description("DIR");
                optdisableaux.set_short_name('A');
                optdisableaux.set_long_name("disableaux");
                optdisableaux.set_description("Disable the auxilliary database");
                optdisableaux.set_arg_description("BOOL");
                optgroup.add_entry_filename(optmaindir, dir_main);
                optgroup.add_entry_filename(optauxdir, dir_aux);
                optgroup.add_entry(optdisableaux, dis_aux);
                Glib::OptionContext optctx("[--maindir=<dir>] [--auxdir=<dir>] [--disableaux]");
                optctx.set_help_enabled(true);
                optctx.set_ignore_unknown_options(false);
                optctx.set_main_group(optgroup);
                CFMUValidate app(argc, argv, optctx);
                if (dis_aux)
                        auxdbmode = Engine::auxdb_none;
                else if (!dir_aux.empty())
                        auxdbmode = Engine::auxdb_override;
                Glib::set_application_name("CFMU Validate");
                Glib::thread_init();
		app.set_dbdirs(dir_main, auxdbmode, dir_aux);
                app.run();
                return 0;
        } catch(const Glib::FileError& ex) {
                std::cerr << "FileError: " << ex.what() << std::endl;
                return 1;
        } catch(const Gtk::BuilderError& ex) {
                std::cerr << "BuilderError: " << ex.what() << std::endl;
                return 1;
        }
        return 1;
}
