//
// C++ Interface: baro
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef BARO_H
#define BARO_H

class IcaoAtmosphere {
  public:
	static const float const_g   = 9.80665;                     /* gravity const. */
	static const float const_m   = 28.9644;                     /* Molecular Weight of air */
	static const float const_r   = 8.31432;                     /* gas const. for air */
	static const float const_l0  = 6.5/1000;                    /* Lapse rate first layer */
	static const float const_gmr;                               /* Hydrostatic constant */
	static const float degc_to_kelvin = 273.15;
	static const float std_sealevel_temperature = 15 + 273.15/*degc_to_kelvin*/;
	static const float std_sealevel_pressure = 1013.25;
	static const float std_sealevel_density;

	IcaoAtmosphere(float qnh = std_sealevel_pressure, float temp = std_sealevel_temperature);
	void set(float qnh = std_sealevel_pressure, float temp = std_sealevel_temperature);
	float get_qnh(void) const { return m_bdry[0].p; }
	float get_temp(void) const { return m_bdry[0].t; }
	void altitude_to_pressure(float *press, float *temp, float alt) const { altitude_to_pressure(press, temp, alt, m_bdry); }
	void pressure_to_altitude(float *alt, float *temp, float press) const { pressure_to_altitude(alt, temp, press, m_bdry); }
	static void std_altitude_to_pressure(float *press, float *temp, float alt) { altitude_to_pressure(press, temp, alt, stdbdry); }
	static void std_pressure_to_altitude(float *alt, float *temp, float press) { pressure_to_altitude(alt, temp, press, stdbdry); }
	static float pressure_to_density(float pressure, float temp);
	static float density_to_pressure(float density, float temp);

  protected:
	struct AtmospericLayer {
		float limit;
		float lapserate;
	};

	struct atmospheric_boundary {
		double t;
		double p;
	};

	//static const unsigned int nr_layers = sizeof(atmosperic_layers)/sizeof(atmosperic_layers[0])-1;
	static const unsigned int nr_layers = 7;
	static const AtmospericLayer atmosperic_layers[nr_layers+1];
	static const atmospheric_boundary stdbdry[nr_layers+1];
	atmospheric_boundary m_bdry[nr_layers+1];

	static void altitude_to_pressure(float *press, float *temp, float alt, const atmospheric_boundary b[nr_layers+1]);
	static void pressure_to_altitude(float *alt, float *temp, float press, const atmospheric_boundary b[nr_layers+1]);
	void compute_boundaries(void);
};

#endif /* BARO_H */
