/*****************************************************************************/

/*
 *      metgraph.cc  --  Meteo Graphs.
 *
 *      Copyright (C) 2014  Thomas Sailer (t.sailer@alumni.ethz.ch)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*****************************************************************************/

#include <iomanip>

#include <librsvg/rsvg.h>

#define __RSVG_RSVG_H_INSIDE__
#include <librsvg/rsvg-cairo.h>
#undef __RSVG_RSVG_H_INSIDE__

#include "metgraph.h"
#include "baro.h"

int MeteoProfile::LinePoint::compare(const LinePoint& x) const
{
	if (get_x() < x.get_x())
		return -1;
	if (x.get_x() < get_x())
		return 1;
	if (get_y() < x.get_y())
		return -1;
	if (x.get_y() < get_y())
		return 1;
	if (is_dir() < x.is_dir())
		return -1;
	if (x.is_dir() < is_dir())
		return 1;
	return 0;
}

MeteoProfile::CloudIcons::CloudIcons(const std::string& basefile, const Cairo::RefPtr<Cairo::Surface> sfc)
	: m_ok(true)
{
	for (unsigned int i = 0; i < 4; ++i) {
		std::string fname;
		{
			std::ostringstream fn;
			fn << basefile << ((i + 1) * 25) << ".svg";
			fname = Glib::build_filename(PACKAGE_DATA_DIR, "gramet", fn.str());
		}
		GError *err(0);
		RsvgHandle *h(rsvg_handle_new_from_file(fname.c_str(), &err));
		if (!h) {
			std::cerr << "Cannot open " << fname;
			if (err)
				std::cerr << ": " << err->message;
			std::cerr << std::endl;
			g_error_free(err);
			err = 0;
			m_ok = false;
			break;
		}
		RsvgDimensionData dim;
		rsvg_handle_get_dimensions(h, &dim);
		if (true)
			std::cerr << "Icon " << fname << " WxH " << dim.width << ' ' << dim.height << std::endl;
		m_icons[i] = Cairo::Surface::create(sfc, Cairo::CONTENT_COLOR_ALPHA,
						    dim.width, dim.height);
		Cairo::RefPtr<Cairo::Context> ctx(Cairo::Context::create(m_icons[i]));
		m_ok = rsvg_handle_render_cairo(h, ctx->cobj());
		g_object_unref(G_OBJECT(h));
		if (!m_ok) {
			std::cerr << "Cannot render " << fname << std::endl;
			break;
		}
	}
	if (m_ok)
		return;
	for (unsigned int i = 0; i < 4; ++i)
		m_icons[i].clear();
}

Cairo::RefPtr<Cairo::Pattern> MeteoProfile::CloudIcons::operator[](unsigned int i) const
{
	Cairo::RefPtr<Cairo::SurfacePattern> p;
	if (!*this || i >= 4 || !m_icons[i])
		return p;
	if (false) {
		double cover((i + 1) * 0.25);
		double gray(1.0 - 0.25 * cover);
		return Cairo::SolidPattern::create_rgba(gray, gray, gray, 0.5 + 2.0 * cover);
	}
	p = Cairo::SurfacePattern::create(m_icons[i]);
	p->set_extend(Cairo::EXTEND_REPEAT);
	p->set_filter(Cairo::FILTER_GOOD);
	return p;
}

MeteoProfile::ConvCloudIcons::Icon::Icon(const Cairo::RefPtr<Cairo::Surface>& sfc, double width, double height)
	: m_surface(sfc), m_width(width), m_height(height), m_area(0)
{
	if (false && sfc) {
		std::cerr << sfc->get_type() << " surface" << std::endl;
		Cairo::RefPtr<Cairo::ImageSurface> imgsfc(Cairo::RefPtr<Cairo::ImageSurface>::cast_dynamic(sfc));
		if (imgsfc) {
			std::cerr << "Image surface format " << imgsfc->get_format() << std::endl;
		}
	}
	if (sfc && width > 1 && height > 1) {
		Cairo::RefPtr<Cairo::ImageSurface> imgsfc(Cairo::ImageSurface::create(Cairo::FORMAT_A8, std::ceil(width), std::ceil(height)));
		{
			Cairo::RefPtr<Cairo::Context> ctx(Cairo::Context::create(imgsfc));
			Cairo::RefPtr<Cairo::SurfacePattern> p(Cairo::SurfacePattern::create(sfc));
			p->set_extend(Cairo::EXTEND_NONE);
			p->set_filter(Cairo::FILTER_GOOD);
			ctx->set_source(p);
			ctx->paint();
		}
		unsigned int width1(imgsfc->get_width()), height1(imgsfc->get_height()), stride1(imgsfc->get_stride());
		const unsigned char *data1(imgsfc->get_data());
		uint64_t area1(0);
		for (unsigned int y = 0; y < height1; ++y) {
			const unsigned char *data2(data1 + y * stride1);
			for (unsigned int x = 0; x < width1; ++x)
				area1 += *data2++;
		}
		m_area = area1 * (1.0 / 255);
		if (false)
			std::cerr << "Tile area " << width * height << " area " << m_area << std::endl;
	}
}

MeteoProfile::ConvCloudIcons::ConvCloudIcons(const std::string& basefile, const Cairo::RefPtr<Cairo::Surface> sfc)
{
	for (unsigned int i = 1; i <= 1024; ++i) {
		std::string fname;
		{
			std::ostringstream fn;
			fn << basefile << i << ".svg";
			fname = Glib::build_filename(PACKAGE_DATA_DIR, "gramet", fn.str());
		}
		if (!Glib::file_test(fname, Glib::FILE_TEST_EXISTS) || !Glib::file_test(fname, Glib::FILE_TEST_IS_REGULAR))
			break;
		GError *err(0);
		RsvgHandle *h(rsvg_handle_new_from_file(fname.c_str(), &err));
		if (!h) {
			std::cerr << "Cannot open " << fname;
			if (err)
				std::cerr << ": " << err->message;
			std::cerr << std::endl;
			g_error_free(err);
			err = 0;
			m_icons.clear();
			break;
		}
		RsvgDimensionData dim;
		rsvg_handle_get_dimensions(h, &dim);
		if (true)
			std::cerr << "Icon " << fname << " WxH " << dim.width << ' ' << dim.height << std::endl;
		Cairo::RefPtr<Cairo::Surface>  sfc1(Cairo::Surface::create(sfc, Cairo::CONTENT_COLOR_ALPHA,
									   dim.width, dim.height));
		Cairo::RefPtr<Cairo::Context> ctx(Cairo::Context::create(sfc1));
		bool ok(rsvg_handle_render_cairo(h, ctx->cobj()));
		g_object_unref(G_OBJECT(h));
		if (!ok) {
			m_icons.clear();
			std::cerr << "Cannot render " << fname << std::endl;
			break;
		}
		m_icons.insert(Icon(sfc1, dim.width, dim.height));
	}
}

Cairo::RefPtr<Cairo::Pattern> MeteoProfile::ConvCloudIcons::get_cloud(double& width, double& height, double& area, int32_t top, int32_t tropopause) const
{
	if (top <= 0 || tropopause <= 0) {
		icons_t::const_iterator i(m_icons.lower_bound(Icon(Cairo::RefPtr<Cairo::Surface>(), 0, height)));
		if (i != m_icons.begin()) {
			icons_t::const_iterator i0(i);
			--i0;
			if (i == m_icons.end() || height - i0->get_height() < i->get_height() - height)
				i = i0;
		}
		if (i == m_icons.end()) {
			width = height = area = std::numeric_limits<double>::quiet_NaN();
			return Cairo::RefPtr<Cairo::Pattern>();
		}
		width = i->get_width();
		height = i->get_height();
		area = i->get_area();
		Cairo::RefPtr<Cairo::SurfacePattern> p(Cairo::SurfacePattern::create(i->get_surface()));
		p->set_extend(Cairo::EXTEND_NONE);
		p->set_filter(Cairo::FILTER_GOOD);
		return p;
	}
	icons_t::size_type n(m_icons.size());
	width = height = area = std::numeric_limits<double>::quiet_NaN();
	if (!n)
		return Cairo::RefPtr<Cairo::Pattern>();
	icons_t::size_type j(std::min((icons_t::size_type)((10 * top) / (8 * tropopause)), n - 1));
	icons_t::const_iterator i(m_icons.begin());
	while (j > 0) {
		icons_t::const_iterator i1(i);
		++i1;
		if (i1 == m_icons.end())
			break;
		i = i1;
		--j;
	}
	width = i->get_width();
	height = i->get_height();
	area = i->get_area();
	Cairo::RefPtr<Cairo::SurfacePattern> p(Cairo::SurfacePattern::create(i->get_surface()));
	p->set_extend(Cairo::EXTEND_NONE);
	p->set_filter(Cairo::FILTER_GOOD);
	return p;
}

MeteoProfile::StratusCloudBlueNoise::Group::Group(const Cairo::RefPtr<Cairo::Surface>& sfc, unsigned int nr, double x, double y)
	: m_surface(sfc), m_x(x), m_y(y), m_nr(nr)
{
	for (unsigned int idx = 0; idx < 16; ++idx)
		m_density[idx] = std::numeric_limits<double>::quiet_NaN();
}

Cairo::RefPtr<Cairo::Pattern> MeteoProfile::StratusCloudBlueNoise::Group::get_pattern(void) const
{
	Cairo::RefPtr<Cairo::SurfacePattern> p(Cairo::SurfacePattern::create(get_surface()));
	p->set_extend(Cairo::EXTEND_NONE);
	p->set_filter(Cairo::FILTER_GOOD);
	return p;
}

int MeteoProfile::StratusCloudBlueNoise::Group::compare(const Group& x) const
{
	if (get_x() < x.get_x())
		return -1;
	if (x.get_x() < get_x())
		return 1;
	if (get_y() < x.get_y())
		return -1;
	if (x.get_y() < get_y())
		return 1;
	return 0;
}

MeteoProfile::StratusCloudBlueNoise::StratusCloudBlueNoise(const std::string& filename, const Cairo::RefPtr<Cairo::Surface> sfc)
	: m_width(0), m_height(0), m_x(0), m_y(0)
{
	// load cloud file
	{
		std::string fname(Glib::build_filename(PACKAGE_DATA_DIR, "gramet", filename));
		GError *err(0);
		RsvgHandle *h(rsvg_handle_new_from_file(fname.c_str(), &err));
		if (!h) {
			std::cerr << "Cannot open " << fname;
			if (err)
				std::cerr << ": " << err->message;
			std::cerr << std::endl;
			g_error_free(err);
			err = 0;
			return;
		}
		RsvgDimensionData dim;
		rsvg_handle_get_dimensions(h, &dim);
		if (false)
			std::cerr << "Stratus Icon " << fname << " WxH " << dim.width << ' ' << dim.height << std::endl;
		RsvgDimensionData dimtile;
		RsvgPositionData postile;
		if (!rsvg_handle_get_dimensions_sub(h, &dimtile, "#tilearea") ||
		    !rsvg_handle_get_position_sub(h, &postile, "#tilearea")) {
			g_object_unref(G_OBJECT(h));
			return;
		}
		if (true)
			std::cerr << "Stratus Icon " << fname << " WxH " << dim.width << ' ' << dim.height
				  << " Tile WxH " << dimtile.width << ' ' << dimtile.height
				  << " XxY " << postile.x << ' ' << postile.y << std::endl;
		for (unsigned int gnr = 0; gnr < 10000; ++gnr) {
			std::ostringstream gnrstr;
			gnrstr << "#g" << std::setw(3) << std::setfill('0') << gnr;
			if (!rsvg_handle_has_sub(h, gnrstr.str().c_str()))
				break;
			Cairo::RefPtr<Cairo::Surface> sfcg(Cairo::Surface::create(sfc, Cairo::CONTENT_COLOR_ALPHA,
										  dim.width, dim.height));
			Cairo::RefPtr<Cairo::Context> ctx(Cairo::Context::create(sfcg));
			if (!rsvg_handle_render_cairo_sub(h, ctx->cobj(), gnrstr.str().c_str()))
				break;
			Cairo::RefPtr<Cairo::ImageSurface> alpha(Cairo::ImageSurface::create(Cairo::FORMAT_A8, dim.width, dim.height));
			ctx = Cairo::Context::create(alpha);
			Cairo::RefPtr<Cairo::SurfacePattern> p(Cairo::SurfacePattern::create(sfcg));
			p->set_extend(Cairo::EXTEND_NONE);
			p->set_filter(Cairo::FILTER_GOOD);
			ctx->set_source(p);
			ctx->paint();
			double mx, my;
			MeteoProfile::center_of_mass(mx, my, alpha);
			if (std::isnan(mx) || std::isnan(my))
				break;
			if (false)
				std::cerr << "Group " << gnrstr.str() << " center of gravity x " << mx << " y " << my << std::endl;
			m_groups.push_back(Group(sfcg, gnr, mx, my));
		}
		g_object_unref(G_OBJECT(h));
		if (m_groups.empty())
			return;
		m_width = dimtile.width;
		m_height = dimtile.height;
		m_x = postile.x;
		m_y = postile.y;
	}
	// calculate blue noise masks
	for (unsigned int idx = 0; idx < 16; ++idx) {
		Cairo::RefPtr<Cairo::ImageSurface> sfc(Cairo::ImageSurface::create(Cairo::FORMAT_A8, 2 * get_width(), 2 * get_height()));
		Cairo::RefPtr<Cairo::Context> ctx(Cairo::Context::create(sfc));
		std::random_shuffle(m_groups.begin(), m_groups.end());
		int stride(sfc->get_stride());
		const uint8_t *data(sfc->get_data());
		double invmaxdensity(1.0 / (255.0 * get_width() * get_height()));
		for (groups_t::iterator gi(m_groups.begin()), ge(m_groups.end()); gi != ge; ++gi) {
			for (int x = 0; x < 2; ++x)
				for (int y = 0; y < 2; ++y) {
					ctx->save();
					ctx->translate(-x * get_width(), -y * get_height());
					ctx->set_source(gi->get_pattern());
					ctx->paint();
					ctx->restore();
				}
			sfc->flush();
			// compute density
			uint64_t dsum(0);
			for (int y = 0; y < get_height(); ++y) {
				const uint8_t *d(data + y * stride);
				for (int x = 0; x < get_width(); ++x)
					dsum += d[x];
			}
			double d(dsum * invmaxdensity);
			gi->set_density(idx, d);
			if (false)
				std::cerr << "Group " << gi->get_nr() << " x " << gi->get_x() << " y " << gi->get_y()
					  << " d " << gi->get_density(idx) << std::endl;
		}
	}
	std::sort(m_groups.begin(), m_groups.end());
}

MeteoProfile::MeteoProfile(const FPlanRoute& route, const TopoDb30::RouteProfile& routeprofile,
			   const GRIB2::WeatherProfile& wxprofile)
	: m_route(route), m_routeprofile(routeprofile), m_wxprofile(wxprofile)
{
}

void MeteoProfile::extract_remainder(std::vector<LinePoint>& ln, std::multiset<LinePoint>& lp, uint32_t maxydiff)
{
	bool dir(ln.back().is_dir());
	for (;;) {
		uint32_t x(ln.back().get_x() + 1);
		int32_t y(ln.back().get_y());
		std::multiset<LinePoint>::iterator li(lp.lower_bound(LinePoint(x, std::numeric_limits<int32_t>::min(), false)));
		std::multiset<LinePoint>::iterator le(lp.end());
		std::multiset<LinePoint>::iterator li2;
		bool li2set(false);
		uint32_t diff(std::numeric_limits<uint32_t>::max());
		for (; li != le && li->get_x() == x; ++li) {
			if (li->is_dir() != dir)
				continue;
			uint32_t d(abs(li->get_y() - y));
			if (d >= diff && li2set)
				continue;
			diff = d;
			li2set = true;
			li2 = li;
		}
		if (!li2set || diff > maxydiff)
			return;
		ln.push_back(*li2);
		lp.erase(li2);
	}
}

std::vector<MeteoProfile::LinePoint> MeteoProfile::extract_line(std::multiset<LinePoint>& lp, uint32_t maxydiff)
{
	for (;;) {
		if (lp.empty())
			return std::vector<LinePoint>();
		std::vector<LinePoint> ln;
		ln.push_back(*lp.begin());
		lp.erase(lp.begin());
		extract_remainder(ln, lp, maxydiff);
		return ln;
	}
}

std::vector<MeteoProfile::LinePoint> MeteoProfile::extract_area(std::multiset<LinePoint>& lp, uint32_t maxydiff)
{
	for (;;) {
		if (lp.empty())
			return std::vector<LinePoint>();
		std::vector<LinePoint> ln1;
		std::vector<LinePoint> ln2;
		{
			std::multiset<LinePoint>::iterator p2(lp.begin()), pe(lp.end());
			uint32_t x(p2->get_x());
			while (p2 != pe) {
				std::multiset<LinePoint>::iterator p1(p2);
				++p2;
				if (p2 == pe)
					break;
				if (p2->get_x() != x)
					break;
				if (p1->is_dir() == p2->is_dir())
					continue;
				ln1.push_back(*p1);
				ln2.push_back(*p2);
				lp.erase(p1);
				lp.erase(p2);
				break;
			}
			if (ln1.empty() || ln2.empty()) {
				lp.erase(lp.begin(), p2);
				continue;
			}
		}
		if (ln1.empty() || ln2.empty())
			continue;
		extract_remainder(ln1, lp, maxydiff);
		extract_remainder(ln2, lp, maxydiff);
		ln1.insert(ln1.end(), ln2.rbegin(), ln2.rend());
		return ln1;
	}
}

void MeteoProfile::draw_clouds_old(const Cairo::RefPtr<Cairo::Context>& cr, const std::vector<CloudPoint>& cldpt,
				   yaxis_t yaxis)
{
	std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end());
	if (ci == ce)
		return;
	for (;;) {
		std::vector<CloudPoint>::const_iterator cip(ci);
		++ci;
		if (ci == ce)
			break;
		cr->move_to(cip->get_dist(), yconvert(cip->get_base(), yaxis));
		cr->line_to(cip->get_dist(), yconvert(cip->get_top(), yaxis));
		cr->line_to(ci->get_dist(), yconvert(ci->get_top(), yaxis));
		cr->line_to(ci->get_dist(), yconvert(ci->get_base(), yaxis));
		cr->close_path();
		set_color_cloud(cr, cip->get_cover());
		cr->fill();
	}
}

void MeteoProfile::draw_clouds(const Cairo::RefPtr<Cairo::Context>& cr, const std::vector<CloudPoint>& cldpt,
			       yaxis_t yaxis, bool preservepath)
{
	if (cldpt.size() < 2)
		return;
	double invdist(cldpt.back().get_dist() - cldpt.front().get_dist());
	if (invdist <= 0)
		return;
	invdist = 1.0 / invdist;
	{
		std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end());
		cr->move_to(ci->get_dist(), yconvert(ci->get_base(), yaxis));
		for (++ci; ci != ce; ++ci)
			cr->line_to(ci->get_dist(), yconvert(ci->get_base(), yaxis));
	}
	for (std::vector<CloudPoint>::const_reverse_iterator ci(cldpt.rbegin()), ce(cldpt.rend()); ci != ce; ++ci)
		cr->line_to(ci->get_dist(), yconvert(ci->get_top(), yaxis));
	cr->close_path();
	Cairo::RefPtr<Cairo::LinearGradient> g(Cairo::LinearGradient::create(0, cldpt.front().get_dist(), 0, cldpt.back().get_dist()));
	for (std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end()); ci != ce; ++ci)
		set_color_cloud(g, (ci->get_dist() - cldpt.front().get_dist()) * invdist, ci->get_cover());
	cr->set_source(g);
	if (preservepath)
		cr->fill_preserve();
	else
		cr->fill();
}

void MeteoProfile::draw_clouds(const Cairo::RefPtr<Cairo::Context>& cr, const std::vector<CloudPoint>& cldpt,
			       yaxis_t yaxis, const CloudIcons& icons, double scale)
{
	static const double blursigma2 = 30;

	if (cldpt.size() < 2)
		return;
	double invdist(cldpt.back().get_dist() - cldpt.front().get_dist());
	if (invdist <= 0)
		return;
	invdist = 1.0 / invdist;
	int width(0), height(0);
	{
		double x0, y0, x1, y1;
		cr->get_clip_extents(x0, y0, x1, y1);
		cr->user_to_device(x0, y0);
		cr->user_to_device(x1, y1);
		if (false)
			std::cerr << "clip extents: " << x0 << ' ' << y0 << ' ' << x1 << ' ' << y1 << std::endl;
		width = std::ceil(std::max(x0, x1));
		height = std::ceil(std::max(y0, y1));
	}
	if (std::isnan(blursigma2) || blursigma2 <= 0) {
		for (unsigned int i = 0; i < 4; ++i) {
			double meandensity((i + 1) * (1.0 / 4));
			// create surface to serve as alpha channel for the mask operation
			Cairo::RefPtr<Cairo::Surface> sfcm(Cairo::Surface::create(cr->get_target(), Cairo::CONTENT_ALPHA, width, height));
			Cairo::RefPtr<Cairo::Context> crm(Cairo::Context::create(sfcm));
			crm->set_matrix(cr->get_matrix());
			// clip the mask channel surface to the cloud dimensions
			{
				std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end());
				crm->move_to(ci->get_dist(), yconvert(ci->get_base(), yaxis));
				for (++ci; ci != ce; ++ci)
					crm->line_to(ci->get_dist(), yconvert(ci->get_base(), yaxis));
			}
			for (std::vector<CloudPoint>::const_reverse_iterator ci(cldpt.rbegin()), ce(cldpt.rend()); ci != ce; ++ci)
				crm->line_to(ci->get_dist(), yconvert(ci->get_top(), yaxis));
			crm->close_path();
			crm->clip();
			// make linear gradient that records cloud density blending
			Cairo::RefPtr<Cairo::LinearGradient> g(Cairo::LinearGradient::create(0, cldpt.front().get_dist(), 0, cldpt.back().get_dist()));
			for (std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end()); ci != ce; ++ci)
				g->add_color_stop_rgba((ci->get_dist() - cldpt.front().get_dist()) * invdist,
						       1., 1., 1., std::max(0., 1. - fabs(ci->get_cover() - meandensity) * 4.));
			crm->set_source(g);
			crm->paint();
			crm.clear();
			// blur alpha channel
			if (true) {
				// convert mask surface to an image surface (A8)
				Cairo::RefPtr<Cairo::ImageSurface> imgsfc(Cairo::ImageSurface::create(Cairo::FORMAT_A8, width, height));
				{
					Cairo::RefPtr<Cairo::Context> ctx(Cairo::Context::create(imgsfc));
					Cairo::RefPtr<Cairo::SurfacePattern> p(Cairo::SurfacePattern::create(sfcm));
					p->set_extend(Cairo::EXTEND_NONE);
					p->set_filter(Cairo::FILTER_GOOD);
					ctx->set_source(p);
					ctx->paint();
				}
				// gaussian blur over the image surface
				gaussian_blur(imgsfc, 15.0);
				sfcm = imgsfc;			
			}
			if (false) {
				std::cerr << sfcm->get_type() << " surface" << std::endl;
				Cairo::RefPtr<Cairo::ImageSurface> imgsfc(Cairo::RefPtr<Cairo::ImageSurface>::cast_dynamic(sfcm));
				if (imgsfc) {
					std::cerr << "Image surface format " << imgsfc->get_format() << std::endl;
				}
			}
			{
				cr->save();
				Cairo::Matrix tm;
				cr->get_matrix(tm);
				cr->set_identity_matrix();
				cr->scale(scale, scale);
				cr->set_source(icons[i]);
				cr->set_identity_matrix();
				cr->mask(sfcm, 0., 0.);
				cr->set_matrix(tm);
				cr->restore();
			}
		}
		return;
	}
	// create master alpha channel
	Cairo::RefPtr<Cairo::ImageSurface> masteralpha(Cairo::ImageSurface::create(Cairo::FORMAT_A8, width, height));
	{
		Cairo::RefPtr<Cairo::Context> crm(Cairo::Context::create(masteralpha));
		crm->set_matrix(cr->get_matrix());
		// clip the mask channel surface to the cloud dimensions
		{
			std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end());
			crm->move_to(ci->get_dist(), yconvert(ci->get_base(), yaxis));
			for (++ci; ci != ce; ++ci)
				crm->line_to(ci->get_dist(), yconvert(ci->get_base(), yaxis));
		}
		for (std::vector<CloudPoint>::const_reverse_iterator ci(cldpt.rbegin()), ce(cldpt.rend()); ci != ce; ++ci)
			crm->line_to(ci->get_dist(), yconvert(ci->get_top(), yaxis));
		crm->close_path();
		crm->clip();
		// make linear gradient that records cloud density blending
		Cairo::RefPtr<Cairo::LinearGradient> g(Cairo::LinearGradient::create(0, cldpt.front().get_dist(), 0, cldpt.back().get_dist()));
		for (std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end()); ci != ce; ++ci)
			g->add_color_stop_rgba((ci->get_dist() - cldpt.front().get_dist()) * invdist, 1., 1., 1., ci->get_cover());
		crm->set_source(g);
		crm->paint();
	}
	// gaussian blur over the image surface
	gaussian_blur(masteralpha, blursigma2);
	// check whether we have any significant cloud
	{
		masteralpha->flush();
		int mastride(masteralpha->get_stride());
		const uint8_t *madata(masteralpha->get_data());
		bool alltr(true);
		for (int y = 0; y < height && alltr; ++y) {
				const uint8_t *d(madata + y * mastride);
				for (int x = 0; x < width && alltr; ++x)
					alltr = alltr && (d[x] < 8);
		}
		if (alltr)
			return;
	}
	// paint textures
	for (unsigned int i = 0; i < 4; ++i) {
		masteralpha->flush();
		Cairo::RefPtr<Cairo::ImageSurface> texturealpha(Cairo::ImageSurface::create(Cairo::FORMAT_A8, width, height));
		texturealpha->flush();
		bool nontr(false);
        	{
			int meanalpha((i << 6) | 0x3f);
			int src_stride(masteralpha->get_stride());
			int dst_stride(texturealpha->get_stride());
			const uint8_t *src(masteralpha->get_data());
			uint8_t *dst(texturealpha->get_data());
			for (int y = 0; y < height; ++y) {
				const uint8_t *s(src + y * src_stride);
				uint8_t *d(dst + y * dst_stride);
				for (int x = 0; x < width; ++x) {
					uint8_t w(abs(s[x] - meanalpha));
					if (w > 0x3f)
						w = 0x3f;
					w <<= 2;
					w = 255 - w;
					if (w < 4)
						w = 0;
					else
						nontr = true;
					d[x] = w;
				}
			}
		}
		texturealpha->mark_dirty();
		if (!nontr)
			continue;
		cr->save();
		Cairo::Matrix tm;
		cr->get_matrix(tm);
		cr->set_identity_matrix();
		cr->scale(scale, scale);
		cr->set_source(icons[i]);
		cr->set_identity_matrix();
		cr->mask(texturealpha, 0., 0.);
		cr->set_matrix(tm);
		cr->restore();
	}
}

void MeteoProfile::draw_clouds(const Cairo::RefPtr<Cairo::Context>& cr, const std::vector<CloudPoint>& cldpt,
			       yaxis_t yaxis, const StratusCloudBlueNoise& icon, double scale)
{
	if (cldpt.size() < 2)
		return;
	double invdist(cldpt.back().get_dist() - cldpt.front().get_dist());
	if (invdist <= 0)
		return;
	invdist = 1.0 / invdist;
	int width(0), height(0);
	{
		double x0, y0, x1, y1;
		cr->get_clip_extents(x0, y0, x1, y1);
		cr->user_to_device(x0, y0);
		cr->user_to_device(x1, y1);
		if (false)
			std::cerr << "clip extents: " << x0 << ' ' << y0 << ' ' << x1 << ' ' << y1 << std::endl;
		width = std::ceil(std::max(x0, x1));
		height = std::ceil(std::max(y0, y1));
	}
	// create master alpha channel
	Cairo::RefPtr<Cairo::ImageSurface> masteralpha(Cairo::ImageSurface::create(Cairo::FORMAT_A8, width, height));
	{
		Cairo::RefPtr<Cairo::Context> crm(Cairo::Context::create(masteralpha));
		crm->set_matrix(cr->get_matrix());
		// clip the mask channel surface to the cloud dimensions
		{
			std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end());
			crm->move_to(ci->get_dist(), yconvert(ci->get_base(), yaxis));
			for (++ci; ci != ce; ++ci)
				crm->line_to(ci->get_dist(), yconvert(ci->get_base(), yaxis));
		}
		for (std::vector<CloudPoint>::const_reverse_iterator ci(cldpt.rbegin()), ce(cldpt.rend()); ci != ce; ++ci)
			crm->line_to(ci->get_dist(), yconvert(ci->get_top(), yaxis));
		crm->close_path();
		crm->clip();
		// make linear gradient that records cloud density blending
		Cairo::RefPtr<Cairo::LinearGradient> g(Cairo::LinearGradient::create(0, cldpt.front().get_dist(), 0, cldpt.back().get_dist()));
		for (std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end()); ci != ce; ++ci)
			g->add_color_stop_rgba((ci->get_dist() - cldpt.front().get_dist()) * invdist, 1., 1., 1., ci->get_cover());
		crm->set_source(g);
		crm->paint();
	}
	masteralpha->flush();
	cr->save();
	cr->set_identity_matrix();
	cr->scale(scale, scale);
	{
		int mastride(masteralpha->get_stride());
		const uint8_t *madata(masteralpha->get_data());
		if (false) {
			for (int y = 0; y < height; ++y) {
				const uint8_t *d(madata + y * mastride);
				std::cerr << "master alpha y=" << y << ':';
				for (int x = 0; x < width; ++x)
					std::cerr << ' ' << (unsigned int)d[x];
				std::cerr << std::endl;
			}
		}
		int xx(-icon.get_width()), yy(-icon.get_height());
		for (;;) {
			{
				double xxd(xx), yyd(yy);
				cr->user_to_device(xxd, yyd);
				if (yyd >= height)
					break;
				if (xxd >= width) {
					xx = -icon.get_width();
					yy += icon.get_height();
					continue;
				}
			}
			unsigned int didx(rand() & 15);
			cr->save();
			cr->translate(xx, yy);
			for (unsigned int i(0), n(icon.size()); i < n; ++i) {
				const StratusCloudBlueNoise::Group& g(icon[i]);
				{
					double xg(g.get_x()), yg(g.get_y());
					cr->user_to_device(xg, yg);
					int gx(Point::round<int,double>(xg)), gy(Point::round<int,double>(yg));
					if (gx < 0 || gx >= width || gy < 0 || gy >= height)
						continue;
					if (false)
						std::cerr << "Group " << g.get_nr() << " x " << (xx + g.get_x())
							  << " y " << (yy + g.get_y()) << " xd " << gx << " yd " << gy
							  << " d " << (255.0 * g.get_density(didx))
							  << " actual " << (unsigned int)madata[gy * mastride + gx] << std::endl;
					if (madata[gy * mastride + gx] < 255.0 * g.get_density(didx))
						continue;
				}
				cr->set_source(g.get_pattern());
				cr->paint();
			}
			cr->restore();
			xx += icon.get_width();
		}
	}
	cr->restore();
}

void MeteoProfile::draw_clouds(const Cairo::RefPtr<Cairo::Context>& cr, const std::vector<CloudPoint>& cldpt,
			       yaxis_t yaxis, const ConvCloudIcons& icons, double scale, double aspectratio, int32_t tropopause)
{
	double areacount(0);
	for (std::vector<CloudPoint>::const_iterator ci(cldpt.begin()), ce(cldpt.end()); ci != ce; ++ci) {
		double ytop(yconvert(ci->get_top(), yaxis));
		double ybase(yconvert(ci->get_base(), yaxis));
		double x0(ci->get_dist()), y0(ytop), x1, y1(ybase);
		cr->user_to_device(x0, y0);
		{
			std::vector<CloudPoint>::const_iterator ci1(ci);
			++ci1;
			if (ci1 == ce)
				break;
			x1 = ci1->get_dist();
		}
		cr->user_to_device(x1, y1);
		double heightr(y1 - y0);
		double cldarea((x1 - x0) * heightr * ci->get_cover());
		if (false)
			std::cerr << "heightr " << heightr << " distr " << (x1 - x0) << " cover " << ci->get_cover()
				  << " cldarea " << cldarea << " areacount " << areacount << std::endl;
		areacount += cldarea;
		if (areacount < 0 || heightr <= 1)
			continue;
		double height(heightr / (scale * aspectratio)), width(0), area(0);
		Cairo::RefPtr<Cairo::Pattern> pat(icons.get_cloud(width, height, area, ci->get_top(), tropopause));
		height *= scale * aspectratio;
		width *= scale;
		area *= scale * scale * aspectratio;
		if (false)
			std::cerr << "cloud size requested: " << heightr << " returned: " << height << ' ' << width << ' ' << area
				  << " pos " << x0 << ' ' << y0 << ' ' << x1 << ' ' << y1 << " cover " << ci->get_cover() << std::endl;
		if (height <= 1 || !pat)
			continue;
		areacount -= area;
		cr->save();
		cr->set_identity_matrix();
		cr->translate(x0, y0);
		cr->scale(scale, scale * aspectratio * heightr / height);
		cr->set_source(pat);
		cr->paint();
		cr->restore();
	}
}

void MeteoProfile::gaussian_blur(const Cairo::RefPtr<Cairo::ImageSurface>& sfc, double sigma2)
{
	if (!sfc || sfc->get_format() != Cairo::FORMAT_A8)
		return;
	sfc->flush();
	int width(sfc->get_width());
	int height(sfc->get_height());
	int src_stride(sfc->get_stride());
	uint8_t *src(sfc->get_data());
	Cairo::RefPtr<Cairo::ImageSurface> sfc2(Cairo::ImageSurface::create(sfc->get_format(), width, height));
	int dst_stride(sfc2->get_stride());
	uint8_t *dst(sfc2->get_data());
	// compute kernel
	uint8_t kernel[31];
	const int ksize = sizeof(kernel) / sizeof(kernel[0]);
	const int khalf = (ksize - 1) / 2;
	uint32_t karea = 0;
	{
		double invsigma2(1.0 / sigma2);
		for (int i = 0; i < ksize; ++i) {
			double f(i - khalf);
			karea += kernel[i] = exp(-f * f * invsigma2) * 255.0;
		}
	}
	// horizontally blur sfc -> sfc2
	for (int i = 0; i < height; ++i) {
		const uint8_t *s = src + i * src_stride;
		uint8_t *d = dst + i * dst_stride;
		for (int j = 0; j < width; ++j) {
			uint32_t w = 0;
			for (int k = 0; k < ksize; ++k) {
				if (j - khalf + k < 0 || j - khalf + k >= width)
					continue;
				w += s[j - khalf + k] * kernel[k];
			}
			w += karea / 2;
			w /= karea;
			w = std::min(w, (uint32_t)255);
			d[j] = w;
		}
	}

	// vertically blur sfc2 -> sfc
	for (int i = 0; i < height; ++i) {
		uint8_t *d = src + i * src_stride;
		for (int j = 0; j < width; ++j) {
			uint32_t w = 0;
			for (int k = 0; k < ksize; ++k) {
				if (i - khalf + k < 0 || i - khalf + k >= height)
					continue;
				w += dst[(i - khalf + k) * dst_stride + j] * kernel[k];
			}
			w += karea / 2;
			w /= karea;
			w = std::min(w, (uint32_t)255);
			d[j] = w;
		}
	}
	sfc->mark_dirty();
}

void MeteoProfile::center_of_mass(double& x, double& y, const Cairo::RefPtr<Cairo::ImageSurface>& sfc)
{
	x = y = std::numeric_limits<double>::quiet_NaN();
	if (!sfc || sfc->get_format() != Cairo::FORMAT_A8)
		return;
	sfc->flush();
	int width(sfc->get_width());
	int height(sfc->get_height());
	int stride(sfc->get_stride());
	const uint8_t *d(sfc->get_data());
	uint64_t mx(0), my(0), m(0);
	for (int i = 0; i < height; ++i) {
		const uint8_t *dl = d + i * stride;
		for (int j = 0; j < width; ++j) {
			m += dl[j];
			mx += j * dl[j];
			my += i * dl[j];
		}
	}
	double im(m);
	im = 1.0 / im;
	x = im * mx;
	y = im * my;
}

double MeteoProfile::get_scaledist(const Cairo::RefPtr<Cairo::Context>& cr, int width, double dist) const
{
	if (dist <= 0 || std::isnan(dist))
		return 0;
	int alt_width;
	{
		cr->save();
		cr->select_font_face("sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
		cr->set_font_size(12);
		Cairo::TextExtents ext;
		cr->get_text_extents("999", ext);
		alt_width = ext.width + 4;
		cr->restore();
	}
	return (width - alt_width) / dist;
}

double MeteoProfile::get_scaleelev(const Cairo::RefPtr<Cairo::Context>& cr, int height, double elev,
				   yaxis_t yaxis) const
{
	if (elev <= 0 || std::isnan(elev))
		return 0;
	elev = yconvert(elev, yaxis) - yconvert(0, yaxis);
	int dist_height;
	int dist_line_height;
	{
		cr->save();
		cr->select_font_face("sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
		cr->set_font_size(12);
		Cairo::TextExtents ext;
		cr->get_text_extents("999", ext);
		dist_line_height = ext.height + 4;
		dist_height = 3 * dist_line_height + 5;
		cr->restore();
	}
	return (height - dist_height) / elev;
}

void MeteoProfile::draw(const Cairo::RefPtr<Cairo::Context>& cr, int width, int height,
			double origindist, double scaledist, double originelev, double scaleelev,
			yaxis_t yaxis) const
{
	cr->save();
	// background
	cr->set_source_rgb(1.0, 1.0, 1.0);
	cr->paint();
	// scaling
	cr->select_font_face("sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
	cr->set_font_size(12);
	if (m_routeprofile.size() < 1) {
		cr->set_source_rgb(0.0, 0.0, 0.0);
		Cairo::TextExtents ext;
		cr->get_text_extents("No Route Profile", ext);
		cr->move_to((width - ext.width) * 0.5 - ext.x_bearing, (height - ext.height) * 0.5 - ext.y_bearing);
		cr->show_text("No Route Profile");
		cr->restore();
		return;
	}
	bool xaxistime(m_routeprofile.size() < 2 || m_routeprofile.front().get_routedist() >= m_routeprofile.back().get_routedist());
	if (m_route.get_nrwpt() < 2 - xaxistime) {
		cr->set_source_rgb(0.0, 0.0, 0.0);
		Cairo::TextExtents ext;
		cr->get_text_extents("No Flight Plan", ext);
		cr->move_to((width - ext.width) * 0.5 - ext.x_bearing, (height - ext.height) * 0.5 - ext.y_bearing);
		cr->show_text("No Flight Plan");
		cr->restore();
		return;
	}
	if (m_wxprofile.size() < 2) {
		cr->set_source_rgb(0.0, 0.0, 0.0);
		Cairo::TextExtents ext;
		cr->get_text_extents("No Weather Profile", ext);
		cr->move_to((width - ext.width) * 0.5 - ext.x_bearing, (height - ext.height) * 0.5 - ext.y_bearing);
		cr->show_text("No Weather Profile");
		cr->restore();
		return;
	}
	{
		bool ok(true);
		for (unsigned int i(1), n(m_route.get_nrwpt()); i < n && ok; ++i)
			ok = m_route[i].get_flighttime() >= m_route[i - 1].get_flighttime();
		if (!ok) {
			cr->set_source_rgb(0.0, 0.0, 0.0);
			Cairo::TextExtents ext;
			cr->get_text_extents("Nonmonotonous Route Time", ext);
			cr->move_to((width - ext.width) * 0.5 - ext.x_bearing, (height - ext.height) * 0.5 - ext.y_bearing);
			cr->show_text("Nonmonotonous Route Time");
			cr->restore();
			return;
		}
	}
	{
		gint64 tm0(m_route[0].get_time_unix());
		gint64 tm1(tm0 + m_route[m_route.get_nrwpt() - 1].get_flighttime());
		if (tm0 < m_wxprofile.get_minefftime() ||
		    tm1 > m_wxprofile.get_maxefftime()) {
			std::ostringstream oss;
			oss << "No Weather Data available for " << Glib::TimeVal(tm0, 0).as_iso8601()
			    << " to " << Glib::TimeVal(tm1, 0).as_iso8601();
			cr->set_source_rgb(0.0, 0.0, 0.0);
			Cairo::TextExtents ext;
			cr->get_text_extents(oss.str(), ext);
			cr->move_to((width - ext.width) * 0.5 - ext.x_bearing, (height - ext.height) * 0.5 - ext.y_bearing);
			cr->show_text(oss.str());
			cr->restore();
			return;
		}
	}
	int alt_width;
	int dist_height;
	int dist_line_height;
	double elevinc, elevstart;
	double distinc, diststart;
	{
		Cairo::TextExtents ext;
		cr->get_text_extents("999", ext);
		alt_width = ext.width + 4;
		dist_line_height = ext.height + 4;
		dist_height = 3 * dist_line_height + 5;
		elevinc = (scaleelev < 0) ? -1 : 1;
		for (unsigned int i = 0; i < 18; ++i) {
			if (elevinc * scaleelev >= 100)
				break;
			elevinc *= ((i % 3) == 1) ? 2.5 : 2;
		}
		elevstart = elevinc * ceil(originelev / elevinc);
		cr->get_text_extents("999", ext);
		distinc = 1;
		if (xaxistime) {
			static const double mul[] = {
				2,
				2.5,
				3,
				2,
				2,
				2,
				2.5,
				3,
				2,
				2,
				3,
				2,
				2,
				2,
				2,
				2,
				2
			};
			for (unsigned int i = 0; i < sizeof(mul)/sizeof(mul[0]); ++i) {
				if (distinc * scaledist >= 100)
					break;
				distinc *= mul[i];
			}
		} else {
			for (unsigned int i = 0; i < 12; ++i) {
				if (distinc * scaledist >= 100)
					break;
				distinc *= ((i % 3) == 1) ? 2.5 : 2;
			}
		}
		diststart = distinc * ceil(origindist / distinc);
	}
	// draw location labels
	if (!xaxistime) {
		// prioritize labels
		unsigned int prio[m_route.get_nrwpt()];
		for (unsigned int i = 0, n = m_route.get_nrwpt(); i < n; ++i) {
			if (!i || i + 1 >= n) {
				prio[i] = 0;
				continue;
			}
			bool routechg(i > 0 && (m_route[i - 1].get_pathcode() != m_route[i].get_pathcode() ||
						m_route[i - 1].get_pathname() != m_route[i].get_pathname()));
			bool rulechg(i > 0 && m_route[i - 1].is_ifr() != m_route[i].is_ifr());
			bool altchg(i + 1 < n && (m_route[i].get_altitude() != m_route[i + 1].get_altitude() ||
						  m_route[i].is_standard() != m_route[i + 1].is_standard()));
			bool turnpt(m_route[i].is_turnpoint());
			prio[i] = 1 + (!turnpt) + (!(routechg || rulechg || altchg));
		}
		set_color_fplan(cr);
		typedef std::pair<double, double> lblext_t;
		typedef std::vector<lblext_t> lblexts_t;
		lblexts_t lblexts;
		for (unsigned int p = 0; p < 4; ++p) {
			for (unsigned int i = 0, n = m_route.get_nrwpt(); i < n; ++i) {
				if (prio[i] != p)
					continue;
				int x;
				{
					double dist(std::numeric_limits<double>::quiet_NaN());
					for (TopoDb30::RouteProfile::const_iterator pi(m_routeprofile.begin()), pe(m_routeprofile.end()); pi != pe; ++pi) {
						if (pi->get_routeindex() != i)
							continue;
						dist = pi->get_routedist();
						break;
					}
					if (std::isnan(dist))
						continue;
					x = alt_width + Point::round<int,double>((dist - origindist) * scaledist);
				}
				if (x < alt_width || x > width)
					continue;
				Glib::ustring txt(m_route[i].get_icao());
				if (txt.empty() || (m_route[i].get_type() == FPlanWaypoint::type_airport && txt == "ZZZZ"))
					txt = m_route[i].get_name();
				if (txt.empty())
					continue;
				Cairo::TextExtents ext;
				cr->get_text_extents(txt, ext);
				double lblx(x - ext.x_bearing);
				if (i + 1 >= n)
					lblx -= ext.width;
				else if (i)
					lblx -= ext.width / 2U;
				{
					lblext_t lble(lblx - 2, lblx + ext.width + 2);
					lblexts_t::const_iterator li(lblexts.begin()), le(lblexts.end());
					for (; li != le; ++li) {
						if (lble.second <= li->first ||
						    lble.first >= li->second)
							continue;
						break;
					}
					if (li != le)
						continue;
					lblexts.push_back(lble);
				}
				cr->move_to(lblx, height - 3U * dist_line_height + 2U - ext.y_bearing);
				cr->show_text(txt);
				cr->begin_new_path();
				cr->move_to(x - 2, height - 3 * dist_line_height);
				cr->rel_line_to(4, 0);
				cr->rel_line_to(-2, -4);
				cr->close_path();
				cr->fill();
			}
		}
	}
	// draw chart legend
	{
		cr->set_source_rgb(0.0, 0.0, 0.0);
		std::ostringstream oss;
		if (m_route.get_nrwpt() > 0) {
			const FPlanWaypoint& wpt(m_route[0]);
			std::string dep;
			if (wpt.get_icao().empty() || (wpt.get_type() == FPlanWaypoint::type_airport && wpt.get_icao() == "ZZZZ"))
				dep = wpt.get_name();
			else
				dep = wpt.get_icao();
			std::string tm0(Glib::TimeVal(wpt.get_time_unix(), 0).as_iso8601());
			oss << dep << ' ' << tm0.substr(0, 10) << ' ' << tm0.substr(11, 5) << 'Z';
			if (m_route.get_nrwpt() > 1) {
				const FPlanWaypoint& wpt(m_route[m_route.get_nrwpt() - 1]);
				oss << " \342\206\222 ";
				std::string dest;
				if (wpt.get_icao().empty() || (wpt.get_type() == FPlanWaypoint::type_airport && wpt.get_icao() == "ZZZZ"))
					dest = wpt.get_name();
				else
					dest = wpt.get_icao();
				std::string tm1(Glib::TimeVal(wpt.get_time_unix(), 0).as_iso8601());
				if (dep != dest)
					oss << dest << ' ';
				if (tm0.substr(0, 10) != tm1.substr(0, 10))
					oss << tm1.substr(0, 10) << ' ';
				oss << tm1.substr(11, 5) << 'Z';
			}
		}
		{
			std::string tm0(Glib::TimeVal(m_wxprofile.get_minreftime(), 0).as_iso8601());
			oss << " \342\200\242 GFS RefTime " << tm0.substr(0, 10) << ' ' << tm0.substr(11, 5) << 'Z';
			if (m_wxprofile.get_maxreftime() != m_wxprofile.get_minreftime()) {
				std::string tm1(Glib::TimeVal(m_wxprofile.get_maxreftime(), 0).as_iso8601());
				oss << '-';
				if (tm0.substr(0, 10) != tm1.substr(0, 10))
					oss << tm1.substr(0, 10) << ' ';
				oss << tm1.substr(11, 5) << 'Z';
			}
		}
		std::ostringstream oss2;
		if (true) {
			std::string tm0(Glib::TimeVal(m_wxprofile.get_minefftime(), 0).as_iso8601());
			oss2 << " \342\200\242 EffTime " << tm0.substr(0, 10) << ' ' << tm0.substr(11, 5) << 'Z';
			if (m_wxprofile.get_maxefftime() != m_wxprofile.get_minefftime()) {
				std::string tm1(Glib::TimeVal(m_wxprofile.get_maxefftime(), 0).as_iso8601());
				oss2 << '-';
				if (tm0.substr(0, 10) != tm1.substr(0, 10))
					oss2 << tm1.substr(0, 10) << ' ';
				oss2 << tm1.substr(11, 5) << 'Z';
			}
		}
		static const char arident[] = " \342\200\242 autorouter.eu";
		Cairo::TextExtents ext;
		cr->get_text_extents(oss.str() + oss2.str() + arident, ext);
		if (ext.width <= width - alt_width)
			oss << oss2.str();
		oss << arident;
		cr->get_text_extents(oss.str(), ext);
		cr->move_to(alt_width - ext.x_bearing, height - dist_line_height + 2U - ext.y_bearing);
		cr->show_text(oss.str());
	}
	// normalize and draw terrain
	cr->save();
	Cairo::Matrix mnorm;
	cr->get_matrix(mnorm);
	cr->rectangle(alt_width, 0, width - alt_width, height - dist_height);
	cr->clip();
	{
		double y0(height - dist_height - (yconvert(0, yaxis) - originelev) * scaleelev);
		double y1(height - dist_height - (yconvert(60000, yaxis) - originelev) * scaleelev);
		Cairo::RefPtr<Cairo::LinearGradient> g(Cairo::LinearGradient::create(0, y0, 0, y1));
		set_color_sky(g);
		cr->set_source(g);
		//set_color_sky(cr);
		cr->paint();
	}
	cr->set_line_width(2);
	cr->translate(alt_width, height - dist_height);
	cr->scale(scaledist, -scaleelev);
	cr->translate(-origindist, -originelev);
	Cairo::Matrix mscaled;
	cr->get_matrix(mscaled);
	// draw dusk/night/dawn
	{
		if (true) {
			int x(-1);
			for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ++pi) {
				int xx(pi->get_flags() & GRIB2::WeatherProfilePoint::flags_daymask);
				if (xx == x)
					continue;
				x = xx;
				switch (x) {
				case GRIB2::WeatherProfilePoint::flags_day:
					std::cerr << "Day";
					break;

				case GRIB2::WeatherProfilePoint::flags_dusk:
					std::cerr << "Dusk";
					break;

				case GRIB2::WeatherProfilePoint::flags_night:
					std::cerr << "Night";
					break;

				case GRIB2::WeatherProfilePoint::flags_dawn:
					std::cerr << "Dawn";
					break;

				default:
					std::cerr << "??";
					break;
				}
				std::cerr << " routedist " << pi->get_routedist() << " time "
					  << Glib::TimeVal(pi->get_efftime(), 0).as_iso8601() << std::endl;
			}
		}
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
			if ((pi->get_flags() & GRIB2::WeatherProfilePoint::flags_daymask) == GRIB2::WeatherProfilePoint::flags_day) {
				++pi;
				continue;
			}
			GRIB2::WeatherProfile::const_iterator pi2(pi);
			for (++pi2; pi2 != pe; ++pi2)
				if ((pi->get_flags() ^ pi2->get_flags()) & GRIB2::WeatherProfilePoint::flags_daymask)
					break;
			double dist1(xaxistime ? (pi->get_efftime() - m_wxprofile.front().get_efftime()) : pi->get_routedist());
			double dist2;
			if (pi2 != pe)
				dist2 = xaxistime ? (pi2->get_efftime() - m_wxprofile.front().get_efftime()) : pi2->get_routedist();
			else
				dist2 = xaxistime ? (m_wxprofile.back().get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile.get_dist();
			if ((pi->get_flags() & GRIB2::WeatherProfilePoint::flags_daymask) == GRIB2::WeatherProfilePoint::flags_night) {
				set_color_sky_night(cr);
			} else {
				int before(-1), after(-1);
				if (pi != m_wxprofile.begin()) {
					GRIB2::WeatherProfile::const_iterator pi2(pi);
					--pi2;
					switch (pi2->get_flags() & GRIB2::WeatherProfilePoint::flags_daymask) {
					case GRIB2::WeatherProfilePoint::flags_day:
						before = 1;
						break;

					case GRIB2::WeatherProfilePoint::flags_night:
						before = 0;
						break;

					default:
						break;
					}
				}
				if (pi2 != pe) {
					switch (pi2->get_flags() & GRIB2::WeatherProfilePoint::flags_daymask) {
					case GRIB2::WeatherProfilePoint::flags_day:
						after = 1;
						break;

					case GRIB2::WeatherProfilePoint::flags_night:
						after = 0;
						break;

					default:
						break;
					}
				}
				if (true)
					std::cerr << "Dusk/Dawn: dist: " << dist1 << "..." << dist2 << " before " << before
						  << " after " << after << std::endl;
				if ((before != 1 && after == 1) || (before == 0 && after != 0)) {
					Cairo::RefPtr<Cairo::LinearGradient> g(Cairo::LinearGradient::create(dist2, 0, dist1, 0));
					set_color_sky_duskdawn(g);
					cr->set_source(g);
				} else if ((before == 1 && after != 1) || (before != 0 && after == 0)) {
					Cairo::RefPtr<Cairo::LinearGradient> g(Cairo::LinearGradient::create(dist1, 0, dist2, 0));
					set_color_sky_duskdawn(g);
					cr->set_source(g);
				} else {
					set_color_sky_duskdawn(cr);
				}
			}
			double yg(yconvert(0, yaxis)), yt(yconvert(100000, yaxis));
			cr->rectangle(dist1, yg, dist2 - dist1, yt - yg);
			cr->fill();
			if (false)
				std::cerr << "dark sky: D" << dist1 << "-D" << dist2 << " state "
					  << (pi->get_flags() & GRIB2::WeatherProfilePoint::flags_daymask) << std::endl;
			pi = pi2;
		}
	}
	// draw boundary clouds
	for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
		if (std::isnan(pi->get_cldbdrycover()) || pi->get_cldbdrycover() <= 0.05 ||
		    pi->get_bdrylayerheight() == GRIB2::WeatherProfilePoint::invalidalt) {
			++pi;
			continue;
		}
		std::vector<CloudPoint> cldpt;
		GRIB2::WeatherProfile::const_iterator pi2(pi);
		for (;;) {
			TopoDb30::RouteProfilePoint rpp(m_routeprofile.interpolate(pi2->get_routedist()));
			if (rpp.get_elev() == TopoDb30::nodata)
				break;
			double elev(rpp.get_elev() * Point::m_to_ft);
			cldpt.push_back(CloudPoint(xaxistime ? (pi2->get_efftime() - m_wxprofile.front().get_efftime()) : pi2->get_routedist(),
						   elev, elev + pi2->get_bdrylayerheight(), pi2->get_cldbdrycover()));
			++pi2;
			if (pi2 == pe || std::isnan(pi2->get_cldbdrycover()) || pi2->get_cldbdrycover() <= 0.05 ||
			    pi2->get_bdrylayerheight() == GRIB2::WeatherProfilePoint::invalidalt)
				break;
		}
		if (pi == pi2) {
			++pi;
			continue;
		}
		pi = pi2;
		draw_clouds(cr, cldpt, yaxis, false);
	}
	// draw low clouds
	{
		CloudIcons icons("stratus", cr->get_target());
		StratusCloudBlueNoise bicon("stratus.svg", cr->get_target());
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
			if (std::isnan(pi->get_cldlowcover()) || pi->get_cldlowcover() <= 0 ||
			    pi->get_cldlowbase() == GRIB2::WeatherProfilePoint::invalidalt ||
			    pi->get_cldlowtop() == GRIB2::WeatherProfilePoint::invalidalt) {
				++pi;
				continue;
			}
			std::vector<CloudPoint> cldpt;
			GRIB2::WeatherProfile::const_iterator pi2(pi);
			for (;;) {
				cldpt.push_back(CloudPoint(xaxistime ? (pi2->get_efftime() - m_wxprofile.front().get_efftime()) : pi2->get_routedist(),
							   pi2->get_cldlowbase(), pi2->get_cldlowtop(), pi2->get_cldlowcover()));
				++pi2;
				if (pi2 == pe || std::isnan(pi2->get_cldlowcover()) || pi2->get_cldlowcover() <= 0 ||
				    pi2->get_cldlowbase() == GRIB2::WeatherProfilePoint::invalidalt ||
				    pi2->get_cldlowtop() == GRIB2::WeatherProfilePoint::invalidalt)
					break;
			}
			pi = pi2;
			if (bicon)
				draw_clouds(cr, cldpt, yaxis, bicon, 0.8);
			else if (icons)
				draw_clouds(cr, cldpt, yaxis, icons, 0.5);
			else
				draw_clouds(cr, cldpt, yaxis, false);
		}
	}
	// draw mid clouds
	{
		CloudIcons icons("altostratus", cr->get_target());
		StratusCloudBlueNoise bicon("stratus.svg", cr->get_target()); // FIXME
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
			if (std::isnan(pi->get_cldmidcover()) || pi->get_cldmidcover() <= 0 ||
			    pi->get_cldmidbase() == GRIB2::WeatherProfilePoint::invalidalt ||
			    pi->get_cldmidtop() == GRIB2::WeatherProfilePoint::invalidalt) {
				++pi;
				continue;
			}
			std::vector<CloudPoint> cldpt;
			GRIB2::WeatherProfile::const_iterator pi2(pi);
			for (;;) {
				cldpt.push_back(CloudPoint(xaxistime ? (pi2->get_efftime() - m_wxprofile.front().get_efftime()) : pi2->get_routedist(),
							   pi2->get_cldmidbase(), pi2->get_cldmidtop(), pi2->get_cldmidcover()));
				++pi2;
				if (pi2 == pe || std::isnan(pi2->get_cldmidcover()) || pi2->get_cldmidcover() <= 0 ||
				    pi2->get_cldmidbase() == GRIB2::WeatherProfilePoint::invalidalt ||
				    pi2->get_cldmidtop() == GRIB2::WeatherProfilePoint::invalidalt)
					break;
			}
			pi = pi2;
			if (bicon)
				draw_clouds(cr, cldpt, yaxis, bicon, 0.8);
			else if (icons)
				draw_clouds(cr, cldpt, yaxis, icons, 0.5);
			else
				draw_clouds(cr, cldpt, yaxis, false);
		}
	}
	// draw high clouds
	{
		CloudIcons icons("cirrus", cr->get_target());
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
			if (std::isnan(pi->get_cldhighcover()) || pi->get_cldhighcover() <= 0 ||
			    pi->get_cldhighbase() == GRIB2::WeatherProfilePoint::invalidalt ||
			    pi->get_cldhightop() == GRIB2::WeatherProfilePoint::invalidalt) {
				++pi;
				continue;
			}
			std::vector<CloudPoint> cldpt;
			GRIB2::WeatherProfile::const_iterator pi2(pi);
			for (;;) {
				cldpt.push_back(CloudPoint(xaxistime ? (pi2->get_efftime() - m_wxprofile.front().get_efftime()) : pi2->get_routedist(),
							   pi2->get_cldhighbase(), pi2->get_cldhightop(), pi2->get_cldhighcover()));
				++pi2;
				if (pi2 == pe || std::isnan(pi2->get_cldhighcover()) || pi2->get_cldhighcover() <= 0 ||
				    pi2->get_cldhighbase() == GRIB2::WeatherProfilePoint::invalidalt ||
				    pi2->get_cldhightop() == GRIB2::WeatherProfilePoint::invalidalt)
					break;
			}
			pi = pi2;
			if (icons)
				draw_clouds(cr, cldpt, yaxis, icons, 0.5);
			else
				draw_clouds(cr, cldpt, yaxis, false);
		}
	}
	// draw convective
	{
		ConvCloudIcons icons("cumulus_", cr->get_target());
		Cairo::RefPtr<Cairo::SurfacePattern> cbpattern;
		if (!icons) {
			Cairo::RefPtr<Cairo::Surface> patsurf(Cairo::Surface::create(cr->get_target(), Cairo::CONTENT_COLOR_ALPHA, 200, 200));
			Cairo::RefPtr<Cairo::Context> cr(Cairo::Context::create(patsurf));
			cr->translate(100, 100);
			for (unsigned int i = 0; i < 2; ++i) {
				cr->set_line_width(i ? 1.5 : 5);
				cr->set_source_rgba(!i, !i, !i, 1);
				cr->save();
				cr->arc(0, 40, 50, M_PI, 0);
				cr->close_path();
				cr->stroke();
				cr->arc(0, 40, 50, M_PI, 0);
				cr->line_to(100, 40);
				cr->line_to(100, -100);
				cr->line_to(-100, -100);
				cr->line_to(-100, 40);
				cr->close_path();
				cr->clip();
				if (true) {
					// TCU symbol
					cr->arc(0, 0, 20, M_PI, 0);
				} else {
					// CB symbol
					cr->arc(0, -30, 40, 0, M_PI);
				}
				cr->close_path();
				cr->stroke();
				cr->restore();
			}
			cbpattern = Cairo::SurfacePattern::create(patsurf);
			cbpattern->set_extend(Cairo::EXTEND_REPEAT);
			cbpattern->set_filter(Cairo::FILTER_GOOD);
		}
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
			if (std::isnan(pi->get_cldconvcover()) || pi->get_cldconvcover() <= 0 ||
			    pi->get_cldconvbase() == GRIB2::WeatherProfilePoint::invalidalt ||
			    pi->get_cldconvtop() == GRIB2::WeatherProfilePoint::invalidalt) {
				++pi;
				continue;
			}
			std::vector<CloudPoint> cldpt;
			GRIB2::WeatherProfile::const_iterator pi2(pi);
			int32_t tropopause(std::numeric_limits<int32_t>::max());
			for (;;) {
				cldpt.push_back(CloudPoint(xaxistime ? (pi2->get_efftime() - m_wxprofile.front().get_efftime()) : pi2->get_routedist(),
							   pi2->get_cldconvbase(), pi2->get_cldconvtop(), pi2->get_cldconvcover()));
				tropopause = std::min(tropopause, pi2->get_tropopause());
				++pi2;
				if (pi2 == pe || std::isnan(pi2->get_cldconvcover()) || pi2->get_cldconvcover() <= 0 ||
				    pi2->get_cldconvbase() == GRIB2::WeatherProfilePoint::invalidalt ||
				    pi2->get_cldconvtop() == GRIB2::WeatherProfilePoint::invalidalt)
					break;
			}
			pi = pi2;
			if (icons) {
				draw_clouds(cr, cldpt, yaxis, icons, 0.5, 4, tropopause);
			} else {
				draw_clouds(cr, cldpt, yaxis, true);
				cr->save();
				cr->clip();
				cr->set_matrix(mnorm);
				cr->scale(0.1, 0.1);
				cr->set_source(cbpattern);
				cr->paint_with_alpha(0.5);
				cr->restore();
			}
		}
	}
	// rain, snow
	{
		const double frequency(0.001 * scaledist * scaleelev);
		Glib::Rand rnd;
		if (false) {
			Glib::TimeVal tv;
			tv.assign_current_time();
			rnd.set_seed(tv.tv_usec ^ tv.tv_sec);
		}
		double yg(yconvert(0, yaxis));
		static const unsigned int nrsfc(sizeof(GRIB2::WeatherProfilePoint::isobaric_levels)/
						sizeof(GRIB2::WeatherProfilePoint::isobaric_levels[0]));
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe;) {
			GRIB2::WeatherProfile::const_iterator pi0(pi);
			++pi;
			if (pi == pe)
				break;
			if (!(pi0->get_flags() & (GRIB2::WeatherProfilePoint::flags_rain |
						  GRIB2::WeatherProfilePoint::flags_freezingrain |
						  GRIB2::WeatherProfilePoint::flags_icepellets |
						  GRIB2::WeatherProfilePoint::flags_snow)))
				continue;
			int32_t alt(-1);
			if (!std::isnan(pi->get_cldmidcover()) && pi->get_cldmidcover() >= 0.25 &&
			    pi->get_cldmidbase() != GRIB2::WeatherProfilePoint::invalidalt && 
			    pi->get_cldmidtop() != GRIB2::WeatherProfilePoint::invalidalt)
				alt = pi->get_cldmidbase();
			else if (!std::isnan(pi->get_cldlowcover()) && pi->get_cldlowcover() >= 0.25 &&
			    pi->get_cldlowbase() != GRIB2::WeatherProfilePoint::invalidalt && 
			    pi->get_cldlowtop() != GRIB2::WeatherProfilePoint::invalidalt)
				alt = pi->get_cldlowbase();
			else
				continue;
			double yc(yconvert(alt, yaxis));
			double dist0(xaxistime ? (pi0->get_efftime() - m_wxprofile.front().get_efftime()) : pi0->get_dist());
			double dist1(xaxistime ? (pi->get_efftime() - m_wxprofile.front().get_efftime()) : pi->get_dist());
			int nr(Point::round<int,double>(rnd.get_double_range(0, fabs(yc - yg) * (dist1 - dist0) * frequency * 2)));
			for (; nr > 0; --nr) {
				int32_t alt1(Point::round<int,double>(rnd.get_double_range(0, alt)));
				double temp(0);
				for (unsigned int i = 1; i < nrsfc; ++i) {
					if (GRIB2::WeatherProfilePoint::isobaric_levels[i - 1] < 0 ||
					    GRIB2::WeatherProfilePoint::isobaric_levels[i] < 0)
						continue;
					if (alt1 < GRIB2::WeatherProfilePoint::altitudes[i - 1] ||
					    alt1 >= GRIB2::WeatherProfilePoint::altitudes[i]);
					float t0(pi->operator[](i-1).get_temp());
					float t1(pi->operator[](i).get_temp());
					temp = t0 + (t1 - t0) * (alt1 - GRIB2::WeatherProfilePoint::altitudes[i - 1])
						/ (GRIB2::WeatherProfilePoint::altitudes[i] - GRIB2::WeatherProfilePoint::altitudes[i - 1]);
					break;
				}
				bool snowcryst(temp < IcaoAtmosphere::degc_to_kelvin + 3);
				cr->move_to(rnd.get_double_range(dist0, dist1), yconvert(alt1, yaxis));
				cr->set_matrix(mnorm);
				if (snowcryst) {
					static const double snowflake[6][2] = {
						{  0.00000,   2.5000 },
						{  2.16506,   1.2500 },
						{  2.16506,  -1.2500 },
						{  0.00000,  -2.5000 },
						{ -2.16506,  -1.2500 },
						{ -2.16506,   1.2500 }
					};
					for (unsigned int i(0); i < 6; ++i) {
						cr->rel_line_to(snowflake[i][0], snowflake[i][1]);
						cr->rel_move_to(-snowflake[i][0], -snowflake[i][1]);
					}
					cr->set_line_width(1.5);
					cr->set_source_rgb(0, 0, 0);
					cr->stroke_preserve();
					cr->set_line_width(0.5);
					cr->set_source_rgb(1, 1, 1);
					cr->stroke();
				} else {
					cr->rel_move_to(-4, 4);
					cr->rel_line_to(4, -8);
					cr->rel_move_to(-2, 8);
					cr->rel_line_to(4, -8);
					cr->rel_move_to(-2, 8);
					cr->set_line_width(3);
					cr->set_source_rgb(1, 1, 1);
					cr->stroke_preserve();
					cr->set_line_width(1);
					cr->set_source_rgb(0, 0, 0);
					cr->stroke();
				}
				cr->set_matrix(mscaled);
			}
		}
	}
	// draw isotherms
	for (int therm = -60; therm < 50; therm += 10) {
		static const unsigned int nrsfc(sizeof(GRIB2::WeatherProfilePoint::isobaric_levels)/
						sizeof(GRIB2::WeatherProfilePoint::isobaric_levels[0]));
		double temp(IcaoAtmosphere::degc_to_kelvin + therm);
		std::multiset<LinePoint> isothermpoints;
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ++pi) {
			for (unsigned int i = 1; i < nrsfc; ++i) {
				if (GRIB2::WeatherProfilePoint::isobaric_levels[i - 1] < 0 ||
				    GRIB2::WeatherProfilePoint::isobaric_levels[i] < 0)
					continue;
				double t0(pi->operator[](i-1).get_temp());
				double t1(pi->operator[](i).get_temp());
				if (std::isnan(t0) || std::isnan(t1))
					continue;
				if ((t0 <= temp && t1 > temp) || (t0 > temp && t1 <= temp)) {
					int32_t alt0(GRIB2::WeatherProfilePoint::altitudes[i - 1]);
					int32_t alt1(GRIB2::WeatherProfilePoint::altitudes[i]);
					int32_t alt(alt0 + (alt1 - alt0) * (temp - t0) / (t1 - t0));
					isothermpoints.insert(LinePoint(pi - m_wxprofile.begin(), alt, t1 > t0));
				}
			}
		}
		{
			std::vector<double> dashes;
			dashes.push_back(2);
			dashes.push_back(2);
			cr->set_dash(dashes, 0);
		}
		cr->set_line_width(1.0);
		double labeldist(std::numeric_limits<double>::quiet_NaN()), labelalt(0);
		for (;;) {
			std::vector<LinePoint> isotherm(extract_line(isothermpoints, 1000));
			if (isotherm.empty())
				break;
			if (isotherm.size() < 2)
				continue;
			cr->begin_new_path();
			std::vector<MeteoProfile::LinePoint>::const_iterator li(isotherm.begin()), le(isotherm.end());
			{
				double dist(xaxistime ? (m_wxprofile[li->get_x()].get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile[li->get_x()].get_routedist());
				cr->move_to(dist, yconvert(li->get_y(), yaxis));
			}
			for (++li; li != le; ++li) {
				double dist(xaxistime ? (m_wxprofile[li->get_x()].get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile[li->get_x()].get_routedist());
				cr->line_to(dist, yconvert(li->get_y(), yaxis));
			}
			cr->set_matrix(mnorm);
			if (therm)
				set_color_otherisotherm(cr);
			else
				set_color_0degisotherm(cr);
			cr->stroke();
			cr->set_matrix(mscaled);
			for (std::vector<MeteoProfile::LinePoint>::const_iterator li(isotherm.begin()), le(isotherm.end()); li != le; ++li) {
				double dist(xaxistime ? (m_wxprofile[li->get_x()].get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile[li->get_x()].get_routedist());
				if ((dist - origindist) * scaledist < 10)
					continue;
				if (std::isnan(labeldist) || dist < labeldist) {
					labeldist = dist;
					labelalt = li->get_y();
				}
			}
		}
		cr->unset_dash();
		if (!std::isnan(labeldist)) {
			std::ostringstream oss;
			oss << therm << "\302\260C";
			cr->move_to(labeldist, yconvert(labelalt, yaxis));
			cr->set_matrix(mnorm);
			cr->set_font_size(8);
			Cairo::TextExtents ext;
			cr->get_text_extents(oss.str(), ext);
			cr->rel_move_to(-ext.x_bearing, -ext.y_bearing - 0.5*ext.height);
			cr->set_source_rgb(1., 1., 1.);
			double x, y;
			cr->get_current_point(x, y);
			cr->set_line_width(2.0);
			cr->text_path(oss.str());
			cr->stroke();
			cr->move_to(x, y);
			if (therm)
				set_color_otherisotherm(cr);
			else
				set_color_0degisotherm(cr);
			cr->show_text(oss.str());
			cr->set_font_size(12);
			cr->set_matrix(mscaled);
		}
	}
	// draw 0degC isotherm
	cr->set_line_width(1.0);
	if (false) {
		double labeldist(std::numeric_limits<double>::quiet_NaN()), labelalt(0);
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
			if (pi->get_zerodegisotherm() == GRIB2::WeatherProfilePoint::invalidalt) {
				++pi;
				continue;
			}
			cr->begin_new_path();
			{
				double dist(xaxistime ? (pi->get_efftime() - m_wxprofile.front().get_efftime()) : pi->get_routedist());
				cr->move_to(dist, yconvert(pi->get_zerodegisotherm(), yaxis));
				if ((dist - origindist) * scaledist >= 10 &&
				    (std::isnan(labeldist) || dist < labeldist)) {
					labeldist = dist;
					labelalt = pi->get_zerodegisotherm();
				}
			}
			GRIB2::WeatherProfile::const_iterator pi2(pi);
			++pi2;
			while (pi2 != pe && pi2->get_zerodegisotherm() != GRIB2::WeatherProfilePoint::invalidalt) {
				double dist(xaxistime ? (pi2->get_efftime() - m_wxprofile.front().get_efftime()) : pi2->get_routedist());
				cr->line_to(dist, yconvert(pi2->get_zerodegisotherm(), yaxis));
				if ((dist - origindist) * scaledist >= 10 &&
				    (std::isnan(labeldist) || dist < labeldist)) {
					labeldist = dist;
					labelalt = pi2->get_zerodegisotherm();
				}
				++pi2;
			}
			if (false && (pi2 - pi) <= 1) {
				pi = pi2;
				continue;
			}
			cr->set_matrix(mnorm);
			set_color_0degisotherm(cr);
			cr->stroke();
			cr->set_matrix(mscaled);
			pi = pi2;
		}
		if (!std::isnan(labeldist)) {
			static const char label[] = "0\302\260C";
			cr->move_to(labeldist, yconvert(labelalt, yaxis));
			cr->set_matrix(mnorm);
			cr->set_font_size(8);
			Cairo::TextExtents ext;
			cr->get_text_extents(label, ext);
			cr->rel_move_to(-ext.x_bearing, -ext.y_bearing - 0.5*ext.height);
			cr->set_source_rgb(1., 1., 1.);
			double x, y;
			cr->get_current_point(x, y);
			cr->set_line_width(2.0);
			cr->text_path(label);
			cr->stroke();
			cr->move_to(x, y);
			set_color_otherisotherm(cr);
			cr->show_text(label);
			cr->set_font_size(12);
			cr->set_matrix(mscaled);
		}
	}
	cr->unset_dash();
	// draw tropopause
	{
		cr->set_line_width(1.0);
		double labeldist(std::numeric_limits<double>::quiet_NaN()), labelalt(0);
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ) {
			if (pi->get_tropopause() == GRIB2::WeatherProfilePoint::invalidalt) {
				++pi;
				continue;
			}
			cr->begin_new_path();
			{
				double dist(xaxistime ? (pi->get_efftime() - m_wxprofile.front().get_efftime()) : pi->get_routedist());
				cr->move_to(dist, yconvert(pi->get_tropopause(), yaxis));
				if ((dist - origindist) * scaledist >= 10 &&
				    (std::isnan(labeldist) || pi->get_routedist() < labeldist)) {
					labeldist = dist;
					labelalt = pi->get_tropopause();
				}
			}
			GRIB2::WeatherProfile::const_iterator pi2(pi);
			++pi2;
			while (pi2 != pe && pi2->get_tropopause() != GRIB2::WeatherProfilePoint::invalidalt) {
				double dist(xaxistime ? (pi2->get_efftime() - m_wxprofile.front().get_efftime()) : pi2->get_routedist());
				cr->line_to(dist, yconvert(pi2->get_tropopause(), yaxis));
				if ((dist - origindist) * scaledist >= 10 &&
				    (std::isnan(labeldist) || dist < labeldist)) {
					labeldist = dist;
					labelalt = pi2->get_tropopause();
				}
				++pi2;
			}
			if (false && (pi2 - pi) <= 1) {
				pi = pi2;
				continue;
			}
			cr->set_matrix(mnorm);
			cr->set_source_rgb(1.0, 0.0, 0.0);
			{
				std::vector<double> dashes;
				dashes.push_back(4);
				dashes.push_back(4);
				cr->set_dash(dashes, 0);
			}
			cr->stroke_preserve();
			cr->unset_dash();
			cr->set_source_rgb(1.0, 0.9, 0.0);
			{
				std::vector<double> dashes;
				dashes.push_back(4);
				dashes.push_back(4);
				cr->set_dash(dashes, 4);
			}
			cr->stroke();
			cr->unset_dash();
			cr->set_matrix(mscaled);
			pi = pi2;
		}
		if (!std::isnan(labeldist)) {
			static const char label[] = "Tropopause";
			cr->move_to(labeldist, yconvert(labelalt, yaxis));
			cr->set_matrix(mnorm);
			cr->set_font_size(8);
			Cairo::TextExtents ext;
			cr->get_text_extents(label, ext);
			cr->rel_move_to(-ext.x_bearing, -ext.y_bearing - 0.5*ext.height);
			cr->set_source_rgb(1., 1., 1.);
			double x, y;
			cr->get_current_point(x, y);
			cr->set_line_width(2.0);
			cr->text_path(label);
			cr->stroke();
			cr->move_to(x, y);
			cr->set_source_rgb(1.0, 0.0, 0.0);
			cr->show_text(label);
			cr->set_font_size(12);
			cr->set_matrix(mscaled);
		}
	}
	// draw isotachs
	cr->set_line_width(1.0);
	for (unsigned int tach = 25; tach <= 200; tach += 25) {
		static const unsigned int nrsfc(sizeof(GRIB2::WeatherProfilePoint::isobaric_levels)/
						sizeof(GRIB2::WeatherProfilePoint::isobaric_levels[0]));
		static const float knots_to_mps(1.0 / 3.6 / Point::km_to_nmi_dbl);
		float mps(tach * knots_to_mps);
		std::multiset<LinePoint> isotachpoints;
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ++pi) {
			int xcnt(0);
			for (unsigned int i = 1; i < nrsfc; ++i) {
				if (GRIB2::WeatherProfilePoint::isobaric_levels[i - 1] < 0 ||
				    GRIB2::WeatherProfilePoint::isobaric_levels[i] < 0)
					continue;
				float t0(pi->operator[](i-1).get_wind());
				float t1(pi->operator[](i).get_wind());
				if (std::isnan(t0) || std::isnan(t1))
					continue;
				if ((t0 <= mps && t1 > mps) || (t0 > mps && t1 <= mps)) {
					int32_t alt0(GRIB2::WeatherProfilePoint::altitudes[i - 1]);
					int32_t alt1(GRIB2::WeatherProfilePoint::altitudes[i]);
					int32_t alt(alt0 + (alt1 - alt0) * (mps - t0) / (t1 - t0));
					isotachpoints.insert(LinePoint(pi - m_wxprofile.begin(), alt, t1 > t0));
					xcnt += (t1 > t0) ? 1 : -1;
				}
			}
			while (xcnt) {
				isotachpoints.insert(LinePoint(pi - m_wxprofile.begin(),
							       (xcnt < 0) ? 0 : GRIB2::WeatherProfilePoint::altitudes[nrsfc - 1],
							       xcnt < 0));
				xcnt += (xcnt < 0) ? 1 : -1;
			}
		}
		for (;;) {
			std::vector<LinePoint> isotach(extract_area(isotachpoints, 1000));
			if (isotach.empty())
				break;
			if (isotach.size() < 3)
				continue;
			cr->begin_new_path();
			std::vector<LinePoint>::const_iterator li(isotach.begin()), le(isotach.end());
			{
				double dist(xaxistime ? (m_wxprofile[li->get_x()].get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile[li->get_x()].get_routedist());
				cr->move_to(dist, yconvert(li->get_y(), yaxis));
			}
			for (++li; li != le; ++li) {
				double dist(xaxistime ? (m_wxprofile[li->get_x()].get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile[li->get_x()].get_routedist());
				cr->line_to(dist, yconvert(li->get_y(), yaxis));
			}
			cr->close_path();
			cr->set_matrix(mnorm);
			set_color_isotach(cr);
			std::vector<double> dashes;
			dashes.push_back(10);
			dashes.push_back(2);
			for (unsigned int x(0); x < tach; x += 25) {
				dashes.push_back(2);
				dashes.push_back(2);
			}
			cr->set_dash(dashes, 0);
			cr->stroke();
			cr->set_matrix(mscaled);
		}
	}
	// draw turbulence
	{
		std::multiset<LinePoint> turb[2]; // moderate, severe
		static const unsigned int nrsfc(sizeof(GRIB2::WeatherProfilePoint::isobaric_levels)/
						sizeof(GRIB2::WeatherProfilePoint::isobaric_levels[0]));
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ++pi) {
			int xcnt[2];
			xcnt[0] = xcnt[1] = 0;
			for (unsigned int i = 1; i < nrsfc; ++i) {
				if (GRIB2::WeatherProfilePoint::isobaric_levels[i - 1] < 0 ||
				    GRIB2::WeatherProfilePoint::isobaric_levels[i] < 0)
					continue;
				float t0(pi->operator[](i-1).get_turbulenceindex());
				float t1(pi->operator[](i).get_turbulenceindex());
				if (std::isnan(t0) || std::isnan(t1))
					continue;
				for (unsigned int k = 0; k < 2; ++k) {
					static const float turblim[2] = { 80, 160 };
					float tlim(turblim[k]);
					if ((t0 <= tlim && t1 > tlim) || (t0 > tlim && t1 <= tlim)) {
						int32_t alt0(GRIB2::WeatherProfilePoint::altitudes[i - 1]);
						int32_t alt1(GRIB2::WeatherProfilePoint::altitudes[i]);
						int32_t alt(alt0 + (alt1 - alt0) * (tlim - t0) / (t1 - t0));
						turb[k].insert(LinePoint(pi - m_wxprofile.begin(), alt, t1 > t0));
						xcnt[k] += (t1 > t0) ? 1 : -1;
					}
				}
			}
			for (unsigned int k = 0; k < 2; ++k) {
				while (xcnt[k]) {
					turb[k].insert(LinePoint(pi - m_wxprofile.begin(),
								 (xcnt[k] < 0) ? 0 : GRIB2::WeatherProfilePoint::altitudes[nrsfc - 1],
								 xcnt[k] < 0));
					xcnt[k] += (xcnt[k] < 0) ? 1 : -1;
				}
			}
		}
		for (unsigned int k(0); k < 2;) {
			std::vector<LinePoint> turbarea(extract_area(turb[k], 1000));
			if (turbarea.empty()) {
				++k;
				continue;
			}
			if (turbarea.size() < 3)
				continue;
			cr->begin_new_path();
			std::vector<LinePoint>::const_iterator li(turbarea.begin()), le(turbarea.end());
			{
				double dist(xaxistime ? (m_wxprofile[li->get_x()].get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile[li->get_x()].get_routedist());
				cr->move_to(dist, yconvert(li->get_y(), yaxis));
			}
			for (++li; li != le; ++li) {
				double dist(xaxistime ? (m_wxprofile[li->get_x()].get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile[li->get_x()].get_routedist());
				cr->line_to(dist, yconvert(li->get_y(), yaxis));
			}
			cr->close_path();
			cr->set_matrix(mnorm);
			cr->set_source_rgb(156 / 256.0, 85 / 256.0, 27 / 256.0);
			std::vector<double> dashes;
			dashes.push_back(5);
			dashes.push_back(2);
			dashes.push_back(5);
			dashes.push_back(2);
			if (k) {
				dashes.push_back(2);
				dashes.push_back(2);
			}
			cr->set_dash(dashes, 0);
			cr->stroke_preserve();
			cr->set_source_rgba(156 / 256.0, 85 / 256.0, 27 / 256.0, 0.2);
			cr->fill();
			cr->set_matrix(mscaled);
		}
	}
	// draw icing
	{
		// IENG_CNV = 200 * (Lwc(bottom_of_cloud) - lwc) / lwc(293K)) * sqrt((T - 253.15)/20.0);  si 253.15<=T<= 273.15
		// IENG_LYR = 100 * t *(t + 14)/ 49; where t = (T - 273.15); when  -14 <= t <= 0.
		// where:
		// lwc(293K) = 0.017281 Kg / m3  (water vapor density at 20ºC in a satured air)
		// lwc = es(T) / (Rv * T) (water vapor density at air cell conditions)
		// es: saturation vapor pressure (SVP)
		// lwc: liquid water content
		std::multiset<LinePoint> icing[2]; // moderate, severe
		static const unsigned int nrsfc(sizeof(GRIB2::WeatherProfilePoint::isobaric_levels)/
						sizeof(GRIB2::WeatherProfilePoint::isobaric_levels[0]));
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ++pi) {
			int xcnt[2];
			xcnt[0] = xcnt[1] = 0;
			float convbasetemp(std::numeric_limits<float>::quiet_NaN());
			double convbaselwc(std::numeric_limits<double>::quiet_NaN());
			for (unsigned int i = 1; i < nrsfc; ++i) {
				if (GRIB2::WeatherProfilePoint::isobaric_levels[i - 1] < 0 ||
				    GRIB2::WeatherProfilePoint::isobaric_levels[i] < 0)
					continue;
				const GRIB2::WeatherProfilePoint::Surface& surf0(pi->operator[](i-1));
				const GRIB2::WeatherProfilePoint::Surface& surf1(pi->operator[](i));
				float ieng0(0), ieng1(0);
				if (!std::isnan(pi->get_cldlowcover())) {
					if (GRIB2::WeatherProfilePoint::altitudes[i-1] >= pi->get_cldlowbase() &&
					    GRIB2::WeatherProfilePoint::altitudes[i-1] <= pi->get_cldlowtop() &&
					    surf0.get_temp() >= IcaoAtmosphere::degc_to_kelvin - 14 &&
					    surf0.get_temp() <= IcaoAtmosphere::degc_to_kelvin) {
						float t(surf0.get_temp() - IcaoAtmosphere::degc_to_kelvin);
						ieng0 += pi->get_cldlowcover() * (100. / 49) * -t * (t + 14);
					}
					if (GRIB2::WeatherProfilePoint::altitudes[i] >= pi->get_cldlowbase() &&
					    GRIB2::WeatherProfilePoint::altitudes[i] <= pi->get_cldlowtop() &&
					    surf1.get_temp() >= IcaoAtmosphere::degc_to_kelvin - 14 &&
					    surf1.get_temp() <= IcaoAtmosphere::degc_to_kelvin) {
						float t(surf1.get_temp() - IcaoAtmosphere::degc_to_kelvin);
						ieng1 += pi->get_cldlowcover() * (100. / 49) * -t * (t + 14);
					}
				}
				if (!std::isnan(pi->get_cldmidcover())) {
					if (GRIB2::WeatherProfilePoint::altitudes[i-1] >= pi->get_cldmidbase() &&
					    GRIB2::WeatherProfilePoint::altitudes[i-1] <= pi->get_cldmidtop() &&
					    surf0.get_temp() >= IcaoAtmosphere::degc_to_kelvin - 14 &&
					    surf0.get_temp() <= IcaoAtmosphere::degc_to_kelvin) {
						float t(surf0.get_temp() - IcaoAtmosphere::degc_to_kelvin);
						ieng0 += pi->get_cldmidcover() * (100. / 49) * -t * (t + 14);
					}
					if (GRIB2::WeatherProfilePoint::altitudes[i] >= pi->get_cldmidbase() &&
					    GRIB2::WeatherProfilePoint::altitudes[i] <= pi->get_cldmidtop() &&
					    surf1.get_temp() >= IcaoAtmosphere::degc_to_kelvin - 14 &&
					    surf1.get_temp() <= IcaoAtmosphere::degc_to_kelvin) {
						float t(surf1.get_temp() - IcaoAtmosphere::degc_to_kelvin);
						ieng1 += pi->get_cldmidcover() * (100. / 49) * -t * (t + 14);
					}
				}
				if (!std::isnan(pi->get_cldhighcover())) {
					if (GRIB2::WeatherProfilePoint::altitudes[i-1] >= pi->get_cldhighbase() &&
					    GRIB2::WeatherProfilePoint::altitudes[i-1] <= pi->get_cldhightop() &&
					    surf0.get_temp() >= IcaoAtmosphere::degc_to_kelvin - 14 &&
					    surf0.get_temp() <= IcaoAtmosphere::degc_to_kelvin) {
						float t(surf0.get_temp() - IcaoAtmosphere::degc_to_kelvin);
						ieng0 += pi->get_cldhighcover() * (100. / 49) * -t * (t + 14);
					}
					if (GRIB2::WeatherProfilePoint::altitudes[i] >= pi->get_cldhighbase() &&
					    GRIB2::WeatherProfilePoint::altitudes[i] <= pi->get_cldhightop() &&
					    surf1.get_temp() >= IcaoAtmosphere::degc_to_kelvin - 14 &&
					    surf1.get_temp() <= IcaoAtmosphere::degc_to_kelvin) {
						float t(surf1.get_temp() - IcaoAtmosphere::degc_to_kelvin);
						ieng1 += pi->get_cldhighcover() * (100. / 49) * -t * (t + 14);
					}
				}
				if (!std::isnan(pi->get_cldconvcover())) {
					if (std::isnan(convbasetemp)) {
						if (pi->get_cldconvbase() <= GRIB2::WeatherProfilePoint::altitudes[i-1])
							convbasetemp = surf0.get_temp();
						else if (pi->get_cldconvbase() <= GRIB2::WeatherProfilePoint::altitudes[i])
							convbasetemp = surf0.get_temp() + (surf1.get_temp() - surf0.get_temp())
								* (pi->get_cldconvbase() - GRIB2::WeatherProfilePoint::altitudes[i-1])
								/ (GRIB2::WeatherProfilePoint::altitudes[i] - GRIB2::WeatherProfilePoint::altitudes[i-1]);
						static const double lwccldbasepoly[] = {
							1.4637e+00, 4.8000e-02, -7.5000e-04
						};
						if (!std::isnan(convbasetemp)) {
							double tt(convbasetemp - IcaoAtmosphere::degc_to_kelvin), t(1);
							convbaselwc = 0;
							for (unsigned int i(0), n(sizeof(lwccldbasepoly)/sizeof(lwccldbasepoly[0])); i < n; ++i) {
								convbaselwc += lwccldbasepoly[i] * t;
								t *= tt;
							}
						}
					}
					static const double lwc293k(17.281);
					// mmHg->Pa: *133.322368 ; Pa = kg·m^−1·s^−2
					// R=8.3144621 m^2·kg·s^-2·K^-1·mol^-1
					// m=18.01528 g/mol
					static const double lwcc1(20.386+log(18.01528*133.322368/8.3144621));
					static const double lwcc2(-5132);
					if (GRIB2::WeatherProfilePoint::altitudes[i-1] >= pi->get_cldconvbase() &&
					    GRIB2::WeatherProfilePoint::altitudes[i-1] <= pi->get_cldconvtop() &&
					    surf0.get_temp() >= IcaoAtmosphere::degc_to_kelvin - 20 &&
					    surf0.get_temp() <= IcaoAtmosphere::degc_to_kelvin) {
						double x(1.0 / surf0.get_temp());
						x *= exp(lwcc1+lwcc2*x);
						x = (convbaselwc - x) * (1.0 / lwc293k);
						x *= 200 * sqrt((surf0.get_temp() + (20 - IcaoAtmosphere::degc_to_kelvin))/20.0);
						if (x >= 0)
							ieng0 += pi->get_cldconvcover() * x;
					}
					if (GRIB2::WeatherProfilePoint::altitudes[i] >= pi->get_cldconvbase() &&
					    GRIB2::WeatherProfilePoint::altitudes[i] <= pi->get_cldconvtop() &&
					    surf1.get_temp() >= IcaoAtmosphere::degc_to_kelvin - 20 &&
					    surf1.get_temp() <= IcaoAtmosphere::degc_to_kelvin) {
						double x(1.0 / surf1.get_temp());
						x *= exp(lwcc1+lwcc2*x);
						x = (convbaselwc - x) * (1.0 / lwc293k);
						x *= 200 * sqrt((surf1.get_temp() + (20 - IcaoAtmosphere::degc_to_kelvin))/20.0);
						if (x >= 0)
							ieng1 += pi->get_cldconvcover() * x;
					}
				}
				ieng0 *= 0.5;
				ieng1 *= 0.5;
				if (false)
					std::cerr << "icing: D" << pi->get_routedist() << ' ' << GRIB2::WeatherProfilePoint::isobaric_levels[i]
						  << "hPa ieng0 " << ieng0 << " ieng1 " << ieng1 << std::endl;
				for (unsigned int k = 0; k < 2; ++k) {
					static const float icelim[2] = { 30, 80 };
					float ilim(icelim[k]);
					if ((ieng0 <= ilim && ieng1 > ilim) || (ieng0 > ilim && ieng1 <= ilim)) {
						int32_t alt0(GRIB2::WeatherProfilePoint::altitudes[i - 1]);
						int32_t alt1(GRIB2::WeatherProfilePoint::altitudes[i]);
						int32_t alt(alt0 + (alt1 - alt0) * (ilim - ieng0) / (ieng1 - ieng0));
						icing[k].insert(LinePoint(pi - m_wxprofile.begin(), alt, ieng1 > ieng0));
						xcnt[k] += (ieng1 > ieng0) ? 1 : -1;
					}
				}
			}
			for (unsigned int k = 0; k < 2; ++k) {
				while (xcnt[k]) {
					icing[k].insert(LinePoint(pi - m_wxprofile.begin(),
								  (xcnt[k] < 0) ? 0 : GRIB2::WeatherProfilePoint::altitudes[nrsfc - 1],
								  xcnt[k] < 0));
					xcnt[k] += (xcnt[k] < 0) ? 1 : -1;
				}
			}
		}
		for (unsigned int k(0); k < 2;) {
			std::vector<LinePoint> icingarea(extract_area(icing[k], 1000));
			if (icingarea.empty()) {
				++k;
				continue;
			}
			if (icingarea.size() < 3)
				continue;
			cr->begin_new_path();
			std::vector<LinePoint>::const_iterator li(icingarea.begin()), le(icingarea.end());
			{
				double dist(xaxistime ? (m_wxprofile[li->get_x()].get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile[li->get_x()].get_routedist());
				cr->move_to(dist, yconvert(li->get_y(), yaxis));
			}
			for (++li; li != le; ++li) {
				double dist(xaxistime ? (m_wxprofile[li->get_x()].get_efftime() - m_wxprofile.front().get_efftime()) : m_wxprofile[li->get_x()].get_routedist());
				cr->line_to(dist, yconvert(li->get_y(), yaxis));
			}
			cr->close_path();
			cr->set_matrix(mnorm);
			cr->set_source_rgb(0 / 256.0, 256 / 256.0, 160 / 256.0);
			std::vector<double> dashes;
			dashes.push_back(5);
			dashes.push_back(2);
			if (k) {
				dashes.push_back(2);
				dashes.push_back(2);
			}
			cr->set_dash(dashes, 0);
			cr->stroke_preserve();
			cr->set_source_rgba(0 / 256.0, 256 / 256.0, 160 / 256.0, 0.2);
			cr->fill();
			cr->set_matrix(mscaled);
		}
	}
	cr->unset_dash();
	// wind barbs
	if (true) {
		Cairo::LineCap lcap(cr->get_line_cap());
		Cairo::LineJoin ljoin(cr->get_line_join());
		cr->set_line_cap(Cairo::LINE_CAP_ROUND);
		cr->set_line_join(Cairo::LINE_JOIN_ROUND);
		double lastx(100);
		const unsigned int endlevels(sizeof(GRIB2::WeatherProfilePoint::isobaric_levels)/
					     sizeof(GRIB2::WeatherProfilePoint::isobaric_levels[0]));
		for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ++pi) {
			double dist(xaxistime ? (pi->get_efftime() - m_wxprofile.front().get_efftime()) : pi->get_routedist());
			{
				double x((dist - origindist) * scaledist);
				if (x < lastx)
					continue;
				lastx = x + 100;
			}
			double lasty(100);
			bool first(true);
			for (unsigned int idx(0); idx < endlevels; ++idx) {
				if (GRIB2::WeatherProfilePoint::isobaric_levels[idx] < 0)
					continue;
				const GRIB2::WeatherProfilePoint::Surface& surf(pi->operator[](idx));
				int32_t alt(GRIB2::WeatherProfilePoint::altitudes[idx]);
				{
					double y((yconvert(alt, yaxis) - originelev) * scaleelev);
					if (y < lasty)
						continue;
					lasty = y + 100;
				}
				// compute cloud cover
				double alpha(1);
				if (!std::isnan(pi->get_cldlowcover()) &&
				    pi->get_cldlowbase() != GRIB2::WeatherProfilePoint::invalidalt &&
				    pi->get_cldlowtop() != GRIB2::WeatherProfilePoint::invalidalt &&
				    pi->get_cldlowbase() <= alt && pi->get_cldlowtop() >= alt)
					alpha *= 1 - pi->get_cldlowcover();
				if (!std::isnan(pi->get_cldmidcover()) &&
				    pi->get_cldmidbase() != GRIB2::WeatherProfilePoint::invalidalt &&
				    pi->get_cldmidtop() != GRIB2::WeatherProfilePoint::invalidalt &&
				    pi->get_cldmidbase() <= alt && pi->get_cldmidtop() >= alt)
					alpha *= 1 - pi->get_cldmidcover();
				if (!std::isnan(pi->get_cldhighcover()) &&
				    pi->get_cldhighbase() != GRIB2::WeatherProfilePoint::invalidalt &&
				    pi->get_cldhightop() != GRIB2::WeatherProfilePoint::invalidalt &&
				    pi->get_cldhighbase() <= alt && pi->get_cldhightop() >= alt)
					alpha *= 1 - pi->get_cldhighcover();
				if (!std::isnan(pi->get_cldconvcover()) &&
				    pi->get_cldconvbase() != GRIB2::WeatherProfilePoint::invalidalt &&
				    pi->get_cldconvtop() != GRIB2::WeatherProfilePoint::invalidalt &&
				    pi->get_cldconvbase() <= alt && pi->get_cldconvtop() >= alt)
					alpha *= 1 - pi->get_cldconvcover();
				cr->save();
				cr->translate(dist, yconvert(alt, yaxis));
				cr->scale(10.0 / scaledist, -10.0 / scaleelev);
				double winddirrad(0);
				int windspeed5(0);
				{
					float wu(surf.get_uwind()), wv(surf.get_vwind());
					wu *= (-1e-3f * Point::km_to_nmi * 3600);
					wv *= (-1e-3f * Point::km_to_nmi * 3600);
					windspeed5 = Point::round<int,double>(sqrtf(wu * wu + wv * wv) * 0.2);
					winddirrad = M_PI - atan2f(wu, wv);
				}
				// draw aircraft track
				static const double windradius = 1.5;
				if (!xaxistime && pi->get_routeindex() < m_route.get_nrwpt()) {
					cr->save();
					cr->rotate(m_route[pi->get_routeindex()].get_truetrack_rad());
					for (unsigned int i(0); i < 2; ++i) {
						cr->set_line_width(i ? 0.10 : 0.30);
						if (i)
							set_color_fplan(cr);
						else
							cr->set_source_rgb(1.0, 1.0, 1.0);
						cr->move_to(0, windradius);
						cr->line_to(0, -windradius);
						cr->move_to(-0.2, 0.4-windradius);
						cr->line_to(0, -windradius);
						cr->line_to(0.2, 0.4-windradius);
						cr->stroke();
					}
					cr->restore();
				}
				if (windspeed5) {
					cr->save();
					cr->rotate(M_PI - winddirrad);
					// draw twice, first white with bigger line width
					for (unsigned int i(0); i < 2; ++i) {
						cr->set_line_width(i ? 0.10 : 0.30);
						cr->set_source_rgb(i ? 0.0 : 1.0, i ? 0.0 : 1.0, i ? 0.0 : 1.0);
						double y(-windradius);
						cr->move_to(0, 0);
						cr->line_to(0, y);
						cr->stroke();
						int speed(windspeed5);
						if (speed < 10)
							y -= 0.2;
						while (speed >= 10) {
							speed -= 10;
							cr->move_to(0, y);
							cr->rel_line_to(-0.4, 0.2);
							cr->rel_line_to(0.4, 0.2);
							if (i)
								cr->fill();
							else
								cr->stroke();
							y += 0.4;
						}
						while (speed >= 2) {
							speed -= 2;
							y += 0.2;
							cr->move_to(0, y);
							cr->rel_line_to(-0.4, -0.2);
							cr->stroke();
						}
						if (speed) {
							y += 0.2;
							cr->move_to(0, y);
							cr->rel_line_to(-0.2, -0.1);
							cr->stroke();
						}
					}
					cr->restore();
				}
				// draw cloud circle
				static const double cldradius = 0.5;
				cr->begin_new_path();
				cr->arc(0.0, 0.0, cldradius, 0.0, 2.0 * M_PI);
				cr->close_path();
				cr->set_line_width(0.10);
				if (alpha >= 0.95) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
				} else if (alpha >= 0.85) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->move_to(0, cldradius);
					cr->line_to(0, -cldradius);
					cr->stroke();
				} else if (alpha >= 0.65) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 2.0 * M_PI);
					cr->line_to(0, 0);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
				} else if (alpha >= 0.55) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 2.0 * M_PI);
					cr->line_to(0, 0);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
					cr->move_to(0, cldradius);
					cr->line_to(0, -cldradius);
					cr->stroke();
				} else if (alpha >= 0.45) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 0.5 * M_PI);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
				} else if (alpha >= 0.35) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 0.5 * M_PI);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
					cr->move_to(-cldradius, 0);
					cr->line_to(cldradius, 0);
					cr->stroke();
				} else if (alpha >= 0.15) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 1.0 * M_PI);
					cr->line_to(0, 0);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
				} else if (alpha >= 0.05) {
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->fill_preserve();
					cr->stroke();
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->move_to(0, -0.9 * cldradius);
					cr->line_to(0, 0.9 * cldradius);
					cr->stroke();
				} else {
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->fill_preserve();
					cr->stroke();
				}
				// show temperature
				static const double textdist = windradius * 1.01;
				std::ostringstream oss;
				oss << Point::round<int,double>(surf.get_temp() - IcaoAtmosphere::degc_to_kelvin) << "\302\260";
				cr->set_source_rgb(1.0, 1.0, 1.0);
				cr->set_font_size(1.0);
				Cairo::TextExtents ext;
				cr->get_text_extents(oss.str(), ext);
				cr->move_to(textdist, ext.height * 0.5);
				cr->set_line_width(0.15);
				cr->text_path(oss.str());
				cr->stroke();
				cr->set_source_rgb(1.0, 0.0, 0.0);
				cr->move_to(textdist, ext.height * 0.5);
				cr->show_text(oss.str());
				if (first && !std::isnan(pi->get_liftedindex()) && pi->get_liftedindex() <= -0.5) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					std::ostringstream oss;
			        	oss << "\342\230\210 " << std::fixed << std::setprecision(1) << pi->get_liftedindex();
					cr->move_to(textdist, ext.height * 1.7);
					cr->set_line_width(0.15);
					cr->text_path(oss.str());
					cr->stroke();
					if (pi->get_liftedindex() <= -6.5)
						// violent
						cr->set_source_rgb(1.0, 0.0, 1.0);
					else if (pi->get_liftedindex() <= -5.5)
						// severe
						cr->set_source_rgb(1.0, 0.0, 0.0);
					else if (pi->get_liftedindex() <= -3.5)
						// probable
						cr->set_source_rgb(1.0, 1.0, 0.0);
        				else if (pi->get_liftedindex() <= -0.5)
						// possible
						cr->set_source_rgb(0.0, 1.0, 0.0);
					else
						// unlikely
						cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->move_to(textdist, ext.height * 1.7);
					cr->show_text(oss.str());
				}
				cr->restore();
				first = false;
			}
		}
    		cr->set_line_cap(lcap);
		cr->set_line_join(ljoin);
	}
	cr->set_matrix(mscaled);
	// draw max elev
	cr->set_line_width(1.0);
	if (xaxistime) {
		double yg(yconvert(0, yaxis));
		double ye(yconvert(m_routeprofile.front().get_maxelev() * Point::m_to_ft, yaxis));
		double dist(m_wxprofile.back().get_efftime() - m_wxprofile.front().get_efftime());
		cr->begin_new_path();
		cr->move_to(0, yg);
		cr->line_to(dist, yg);
		cr->line_to(dist, ye);
		cr->line_to(0, ye);
		cr->close_path();
	} else {
		double yg(yconvert(0, yaxis));
		cr->begin_new_path();
		cr->move_to(0, yg);
		double dist(0);
		for (TopoDb30::RouteProfile::const_iterator pi(m_routeprofile.begin()), pe(m_routeprofile.end()); pi != pe; ++pi)
			cr->line_to(dist = pi->get_routedist(), yconvert(pi->get_maxelev() * Point::m_to_ft, yaxis));
		cr->line_to(dist, yg);
		cr->close_path();
	}
	cr->set_matrix(mnorm);
	cr->set_source_rgb(1.0, 1.0, 1.0);
	cr->fill_preserve();
	set_color_terrain_fill(cr);
	cr->fill_preserve();
	set_color_terrain(cr);
	cr->stroke();
	cr->set_matrix(mscaled);
	// draw elev
	if (xaxistime) {
		double yg(yconvert(0, yaxis));
		double ye(yconvert(m_routeprofile.front().get_elev() * Point::m_to_ft, yaxis));
		double dist(m_wxprofile.back().get_efftime() - m_wxprofile.front().get_efftime());
		cr->begin_new_path();
		cr->move_to(0, yg);
		cr->line_to(dist, yg);
		cr->line_to(dist, ye);
		cr->line_to(0, ye);
		cr->close_path();
	} else {
		double yg(yconvert(0, yaxis));
		cr->begin_new_path();
		cr->move_to(0, yg);
		double dist(0);
		for (TopoDb30::RouteProfile::const_iterator pi(m_routeprofile.begin()), pe(m_routeprofile.end()); pi != pe; ++pi)
			cr->line_to(dist = pi->get_routedist(), yconvert(pi->get_elev() * Point::m_to_ft, yaxis));
		cr->line_to(dist, yg);
		cr->close_path();
	}
	cr->set_matrix(mnorm);
	set_color_terrain_fill(cr);
	cr->fill_preserve();
	set_color_terrain(cr);
	cr->stroke();
	cr->restore();
	// draw scales
	cr->set_line_width(0.5);
	{
		std::vector<double> dashes;
		dashes.push_back(2);
		dashes.push_back(2);
		cr->set_dash(dashes, 0);
 	}
	if (true && yaxis == yaxis_pressure) {
		typedef std::map<double,int32_t> grids_t;
		grids_t grids;
		{
			static const double ymindiff(60);
			double yheight(height - dist_height);
			int32_t altinc(10000), altmin(0), altmax(90000);
			for (unsigned int i = 0; altinc >= 10 && i < 20; ++i) {
				bool work(false);
				for (int32_t alt(altmin); alt <= altmax; alt += altinc) {
					double y(yheight - (yconvert(alt, yaxis) - originelev) * scaleelev);
					if (false)
						std::cerr << "alt " << alt << " y " << y << " yh " << yheight << std::endl;
					if (y < 0) {
						altmax = std::min(altmax, alt);
						continue;
					}
					if (y > yheight) {
						altmin = std::max(altmin, alt);
						continue;
					}
					grids_t::iterator gi(grids.lower_bound(y));
					if (gi != grids.end() && (gi->first - y) < ymindiff)
						continue;
					if (gi != grids.begin()) {
						grids_t::iterator gi2(gi);
						--gi2;
						if ((y - gi2->first) < ymindiff)
							continue;
					}
					work = true;
					grids.insert(gi, grids_t::value_type(y, alt));
				}
				if (!work && !grids.empty())
					break;
				if ((i % 3) == 1)
					altinc = (altinc << 1) / 5;
				else
					altinc >>= 1;
			}
		}
		for (grids_t::const_iterator gi(grids.begin()), ge(grids.end()); gi != ge; ++gi) {
			cr->move_to(alt_width, gi->first);
			cr->line_to(width, gi->first);
			set_color_grid(cr);
			cr->stroke();
			set_color_marking(cr);
			std::ostringstream oss;
			oss << std::fixed << std::setprecision(0) << (gi->second * 0.01);
			Cairo::TextExtents ext;
			cr->get_text_extents(oss.str(), ext);
			cr->move_to(alt_width - 2 - ext.width - ext.x_bearing, gi->first - ext.height - 2 - ext.y_bearing);
			cr->show_text(oss.str());
			std::ostringstream ossp;
			{
				float p;
				IcaoAtmosphere::std_altitude_to_pressure(&p, 0, gi->second * Point::ft_to_m);
				ossp << std::fixed << std::setprecision(0) << p;
			}
			cr->set_font_size(8);
			cr->get_text_extents(ossp.str(), ext);
			cr->move_to(alt_width - 2 - ext.width - ext.x_bearing, gi->first + 2 - ext.y_bearing);
			cr->show_text(ossp.str());
			cr->set_font_size(12);
		}
	} else {
		for (;; elevstart += elevinc) {
			if (yaxis == yaxis_pressure && elevstart <= 10)
				break;
			double y(height - dist_height - (elevstart - originelev) * scaleelev);
			if (y < 0)
				break;
			cr->move_to(alt_width, y);
			cr->line_to(width, y);
			set_color_grid(cr);
			cr->stroke();
			set_color_marking(cr);
			std::ostringstream oss;
			switch (yaxis) {
			default:
				oss << std::fixed << std::setprecision(0) << (elevstart * 0.01);
				break;

			case yaxis_pressure:
			{
				float a;
				IcaoAtmosphere::std_pressure_to_altitude(&a, 0, elevstart);
				oss << std::fixed << std::setprecision(0) << (a * (Point::m_to_ft_dbl * 0.01));
				break;
			}
			}
			Cairo::TextExtents ext;
			cr->get_text_extents(oss.str(), ext);
			cr->move_to(alt_width - 2 - ext.width - ext.x_bearing, y - ext.height - 2 - ext.y_bearing);
			cr->show_text(oss.str());
			std::ostringstream ossp;
			switch (yaxis) {
			default:
			{
				float p;
				IcaoAtmosphere::std_altitude_to_pressure(&p, 0, elevstart * Point::ft_to_m);
				ossp << std::fixed << std::setprecision(0) << p;
				break;
			}

			case yaxis_pressure:
				ossp << std::fixed << std::setprecision(0) << elevstart;
				break;
			}
			cr->set_font_size(8);
			cr->get_text_extents(ossp.str(), ext);
			cr->move_to(alt_width - 2 - ext.width - ext.x_bearing, y + 2 - ext.y_bearing);
			cr->show_text(ossp.str());
			cr->set_font_size(12);
		}
	}
	for (;; diststart += distinc) {
		double x(alt_width + (diststart - origindist) * scaledist);
		if (x >= width)
			break;
		cr->move_to(x, 0);
		cr->line_to(x, height - dist_height);
		set_color_grid(cr);
		cr->stroke();
		set_color_marking(cr);
		if (xaxistime) {
			std::ostringstream oss;
			oss << Glib::TimeVal(m_wxprofile.front().get_efftime() + diststart, 0).as_iso8601().substr(11, 5);
			Cairo::TextExtents ext;
			cr->get_text_extents(oss.str(), ext);
			double lblx(x - ext.width / 2U - ext.x_bearing);
			cr->move_to(lblx, height - 2U * dist_line_height + 2U - ext.y_bearing);
			cr->show_text(oss.str());
		} else {
			std::ostringstream oss;
			oss << std::fixed << std::setprecision(0) << diststart;
			Cairo::TextExtents ext;
			cr->get_text_extents(oss.str(), ext);
			double lblx(x - ext.width / 2U - ext.x_bearing);
			cr->move_to(lblx, height - 2U * dist_line_height + 2U - ext.y_bearing);
			cr->show_text(oss.str());
			lblx += ext.width;
			gint64 efftime(-1);
			{
				double bd(std::numeric_limits<double>::max());
				for (GRIB2::WeatherProfile::const_iterator pi(m_wxprofile.begin()), pe(m_wxprofile.end()); pi != pe; ++pi) {
					double d(fabs(pi->get_routedist() - diststart));
					if (d > bd)
						continue;
					bd = d;
					efftime = pi->get_efftime();
				}
			}
			if (efftime < 0)
				continue;
			double h12(ext.height);
			std::string efftstr("  " + Glib::TimeVal(efftime, 0).as_iso8601().substr(11, 5));
			cr->set_font_size(8);
			cr->get_text_extents(efftstr, ext);
			cr->move_to(lblx, height - 2U * dist_line_height + 2U - ext.y_bearing + h12 - ext.height);
			cr->show_text(efftstr);
			cr->set_font_size(12);
		}
	}
        cr->unset_dash();
	cr->restore();
	// normalize and draw route
	if (!xaxistime) {
		cr->save();
		cr->rectangle(alt_width, 0, width - alt_width, height - dist_height);
		set_color_marking(cr);
		cr->set_line_width(1);
		cr->stroke_preserve();
		cr->clip();
		set_color_fplan(cr);
		cr->set_line_width(2);
		cr->set_matrix(mscaled);
		bool first(true);
		cr->begin_new_path();
		for (unsigned int i = 0, j = m_route.get_nrwpt(); i < j; ++i) {
			double dist(std::numeric_limits<double>::quiet_NaN());
			for (TopoDb30::RouteProfile::const_iterator pi(m_routeprofile.begin()), pe(m_routeprofile.end()); pi != pe; ++pi) {
				if (pi->get_routeindex() != i)
					continue;
				dist = pi->get_routedist();
				break;
			}
			if (std::isnan(dist))
				continue;
			double y(yconvert(m_route[i].get_true_altitude(), yaxis));
			if (first)
				cr->move_to(dist, y);
			else
				cr->line_to(dist, y);
			first = false;
		}
		cr->set_matrix(mnorm);
		cr->stroke();
		cr->restore();
	}
	// end
}

MeteoChart::AreaExtract::ScalarGridPt::ScalarGridPt()
	: m_val(std::numeric_limits<float>::min()), m_inside(false)
{
	m_coord.set_invalid();
}

MeteoChart::AreaExtract::ScalarGridPt::ScalarGridPt(const Glib::RefPtr<GRIB2::LayerInterpolateResult>& lay, unsigned int x, unsigned int y,
						    float lim, gint64 efftime, double sfc1value)
	: m_val(std::numeric_limits<float>::min()), m_inside(false)
{
	m_coord.set_invalid();
	if (!lay)
		return;
	m_coord = lay->get_center(x, y);
	float v(lay->operator()(x, y, efftime, sfc1value));
	if (std::isnan(v))
		return;
	m_val = v;
	m_inside = m_val >= lim;
}

MeteoChart::AreaExtract::ScalarGridPt::ScalarGridPt(const ScalarGridPt& x, const ScalarGridPt& y, float lim)
	: m_val((x.get_val() + y.get_val()) * 0.5), m_inside(false)
{
	m_coord =x.get_coord().halfway(y.get_coord());
	m_inside = m_val >= lim;
}

Point MeteoChart::AreaExtract::ScalarGridPt::contour_point(const ScalarGridPt& x, float lim) const
{
	if (is_inside() == x.is_inside())
		throw std::runtime_error("ScalarGridPt::contour_point");
	float t((lim - get_val()) / (x.get_val() - get_val()));
	t = std::max(0.f, std::min(1.f, t));
	Point pdiff(x.get_coord() - get_coord());
	return get_coord() + Point(pdiff.get_lon() * t, pdiff.get_lat() * t);
}

MeteoChart::AreaExtract::WindGridPt::WindGridPt()
{
	m_val = 0;
}

MeteoChart::AreaExtract::WindGridPt::WindGridPt(const Glib::RefPtr<GRIB2::LayerInterpolateResult>& layu,
						const Glib::RefPtr<GRIB2::LayerInterpolateResult>& layv,
						unsigned int x, unsigned int y, float lim, gint64 efftime, double sfc1value)
{
	m_val = 0;
	if (!layu || !layv)
		return;
	m_coord = layu->get_center(x, y);
	float u(layu->operator()(x, y, efftime, sfc1value));
	float v(layv->operator()(x, y, efftime, sfc1value));
	if (std::isnan(u) || std::isnan(v))
		return;
	m_val = sqrtf(u * u + v * v);
	m_inside = m_val >= lim;
}

MeteoChart::AreaExtract::WindGridPt::WindGridPt(const WindGridPt& x, const WindGridPt& y, float lim)
	: ScalarGridPt(x, y, lim)
{
}

const float MeteoChart::AreaExtract::CloudGridPt::mincover;

MeteoChart::AreaExtract::CloudGridPt::CloudGridPt()
	: m_cover(0), m_base(0), m_top(200000), m_inside(false)
{
	m_coord.set_invalid();	
}

MeteoChart::AreaExtract::CloudGridPt::CloudGridPt(const Glib::RefPtr<GRIB2::LayerInterpolateResult>& laycover,
						  const Glib::RefPtr<GRIB2::LayerInterpolateResult>& laybase,
						  const Glib::RefPtr<GRIB2::LayerInterpolateResult>& laytop,
						  unsigned int x, unsigned int y, gint64 efftime, float pressure)
	: m_cover(0), m_base(0), m_top(200000), m_inside(false)
{
	m_coord.set_invalid();
	if (!laycover || !laybase || !laytop)
		return;
	m_coord = laycover->get_center(x, y);
	float c(laycover->operator()(x, y, efftime, 0));
	float b(laybase->operator()(x, y, efftime, 0));
	float t(laytop->operator()(x, y, efftime, 0));
	if (std::isnan(c) || std::isnan(b) || std::isnan(t))
		return;
	m_cover = std::max(0.f, std::min(1.f, c * 0.01f));
	m_base = b;
	m_top = t;
	m_inside = m_cover >= mincover && pressure <= m_base && pressure >= m_top;
}

MeteoChart::AreaExtract::CloudGridPt::CloudGridPt(const CloudGridPt& x, const CloudGridPt& y, float pressure)
	: m_cover((x.get_cover() + y.get_cover()) * 0.5), m_base((x.get_base() + y.get_base()) * 0.5),
	  m_top((x.get_top() + y.get_top()) * 0.5), m_inside(false)
{
	m_coord =x.get_coord().halfway(y.get_coord());
	m_inside = m_cover >= mincover && pressure <= m_base && pressure >= m_top;
}

Point MeteoChart::AreaExtract::CloudGridPt::contour_point(const CloudGridPt& x, float pressure) const
{
	if (is_inside() == x.is_inside())
		throw std::runtime_error("CloudGridPt::contour_point");
	float tc((mincover - get_cover()) / (x.get_cover() - get_cover()));
	float tb((pressure - get_base()) / (x.get_base() - get_base()));
	float tt((pressure - get_top()) / (x.get_top() - get_top()));
	tc = std::max(0.f, std::min(1.f, tc));
	tb = std::max(0.f, std::min(1.f, tb));
	tt = std::max(0.f, std::min(1.f, tt));
	float t;
	if (is_inside())
		t = std::min(tc, std::min(tb, tt));
	else
		t = std::max(tc, std::max(tb, tt));
	Point pdiff(x.get_coord() - get_coord());
	return get_coord() + Point(pdiff.get_lon() * t, pdiff.get_lat() * t);
}

template <typename PT> MeteoChart::AreaExtract::Grid<PT>::Grid(unsigned int w, unsigned int h)
	: m_width(w), m_height(h)
{
	if (m_width < 4 || m_height < 4) {
		m_width = m_height = 0;
		return;
	}
	m_v.resize(m_width * m_height);
}

template <typename PT> PT& MeteoChart::AreaExtract::Grid<PT>::operator()(unsigned int x, unsigned int y)
{
	if (x >= get_width() || y >= get_height())
		return *(PT *)0;
	return m_v[y * get_width() + x];
}

template <typename PT> const PT& MeteoChart::AreaExtract::Grid<PT>::operator()(unsigned int x, unsigned int y) const
{
	if (x >= get_width() || y >= get_height())
		return *(PT *)0;
	return m_v[y * get_width() + x];
}

template <typename PT> void MeteoChart::AreaExtract::Grid<PT>::set_border(void)
{
	if (m_width < 4 || m_height < 4)
		return;
	Point pdiff0(operator()(2, 1).get_coord() - operator()(1, 1).get_coord());
	Point pdiff1(operator()(1, 2).get_coord() - operator()(1, 1).get_coord());
	for (unsigned int x = 2; x < get_width(); ++x) {
		operator()(x - 1, 0).set_coord(operator()(x - 1, 1).get_coord() - pdiff1);
		operator()(x - 1, get_height() - 1).set_coord(operator()(x - 1, get_height() - 2).get_coord() + pdiff1);
	}
	for (unsigned int y = 2; y < get_height(); ++y) {
		operator()(0, y - 1).set_coord(operator()(1, y - 1).get_coord() - pdiff0);
		operator()(get_width() - 1, y - 1).set_coord(operator()(get_width() - 2, y - 1).get_coord() + pdiff0);
	}
	operator()(0, 0).set_coord(operator()(1, 0).get_coord() - pdiff0);
	operator()(get_width() - 1, 0).set_coord(operator()(get_width() - 2, 0).get_coord() + pdiff0);
	operator()(0, get_height() - 1).set_coord(operator()(1, get_height() - 1).get_coord() - pdiff0);
	operator()(get_width() - 1, get_height() - 1).set_coord(operator()(get_width() - 2, get_height() - 1).get_coord() + pdiff0);	
}

template <typename PT> unsigned int MeteoChart::AreaExtract::Grid<PT>::get_index(unsigned int x, unsigned int y) const
{
	if (x >= get_width() || y >= get_height())
		return ~0U;
	return y * get_width() + x;
}

template <typename PT> unsigned int MeteoChart::AreaExtract::Grid<PT>::get_edgenr(unsigned int x0, unsigned int y0, unsigned int x1, unsigned int y1) const
{
	if ((x0 != x1 && y0 != y1) ||
	    (x0 == x1 && y0 == y1))
		throw std::runtime_error("Grid::get_edgenr");
	if (x0 > x1)
		std::swap(x0, x1);
	if (y0 > y1)
		std::swap(y0, y1);
	if (x1 - x0 > 1 || y1 - y0 > 1)
		throw std::runtime_error("Grid::get_edgenr");
	return (get_index(x0, y0) << 1) | (y1 > y0);
}

int MeteoChart::AreaExtract::Edge::compare(const Edge& x) const
{
	if (get_startedge() < x.get_startedge())
		return -1;
	if (x.get_startedge() < get_startedge())
		return 1;
	if (get_endedge() < x.get_endedge())
		return -1;
	if (x.get_endedge() < get_endedge())
		return 1;
	return 0;
}

MeteoChart::AreaExtract::AreaExtract(const Glib::RefPtr<GRIB2::LayerInterpolateResult>& lay, float lim, gint64 efftime, double sfc1value)
{
	if (!lay)
		return;
	unsigned int width(lay->get_width());
	unsigned int height(lay->get_height());
	if (width < 2 || height < 2)
		return;

	Grid<ScalarGridPt> grid(width + 2, height + 2);
	for (unsigned int x(0); x < width; ++x)
		for (unsigned int y(0); y < height; ++y)
			grid(x + 1, y + 1) = ScalarGridPt(lay, x, y, lim, efftime, sfc1value);
	grid.set_border();
	compute(grid, lim);
}

MeteoChart::AreaExtract::AreaExtract(const Glib::RefPtr<GRIB2::LayerInterpolateResult>& layu,
				     const Glib::RefPtr<GRIB2::LayerInterpolateResult>& layv, float lim, gint64 efftime, double sfc1value)
{
	if (!layu || !layv)
		return;
	if (layu->get_bbox() != layv->get_bbox())
		return;
	unsigned int width(layu->get_width());
	unsigned int height(layu->get_height());
	if (layv->get_width() != width || layv->get_height() != height)
		return;
	if (width < 2 || height < 2)
		return;
	Grid<WindGridPt> grid(width + 2, height + 2);
	for (unsigned int x(0); x < width; ++x)
		for (unsigned int y(0); y < height; ++y)
			grid(x + 1, y + 1) = WindGridPt(layu, layv, x, y, lim, efftime, sfc1value);
	grid.set_border();
	compute(grid, lim);
}

MeteoChart::AreaExtract::AreaExtract(const Glib::RefPtr<GRIB2::LayerInterpolateResult>& laycover,
				     const Glib::RefPtr<GRIB2::LayerInterpolateResult>& laybase,
				     const Glib::RefPtr<GRIB2::LayerInterpolateResult>& laytop, gint64 efftime, double pressure)
{
	if (!laycover || !laybase || !laytop)
		return;
	if (laycover->get_bbox() != laybase->get_bbox() || laycover->get_bbox() != laytop->get_bbox())
		return;
	unsigned int width(laycover->get_width());
	unsigned int height(laycover->get_height());
	if (laybase->get_width() != width || laybase->get_height() != height ||
	    laytop->get_width() != width || laytop->get_height() != height)
		return;
	if (width < 2 || height < 2)
		return;
	Grid<CloudGridPt> grid(width + 2, height + 2);
	for (unsigned int x(0); x < width; ++x)
		for (unsigned int y(0); y < height; ++y)
			grid(x + 1, y + 1) = CloudGridPt(laycover, laybase, laytop, x, y, efftime, pressure);
	grid.set_border();
	compute(grid, pressure);
	if (false) {
		std::cerr << "Area Extract: pressure " << pressure << " efftime " << Glib::TimeVal(efftime, 0).as_iso8601() << std::endl
			  << "  Cover: " << laycover->get_layer()->get_parameter()->get_str_nonnull()
			  << " sfc1 " << GRIB2::find_surfacetype_str(laycover->get_layer()->get_surface1type(), "?") << ' '
			  << laycover->get_minsurface1value() << "..." << laycover->get_maxsurface1value() << std::endl
			  << "  Base: " << laybase->get_layer()->get_parameter()->get_str_nonnull()
			  << " sfc1 " << GRIB2::find_surfacetype_str(laybase->get_layer()->get_surface1type(), "?") << ' '
			  << laybase->get_minsurface1value() << "..." << laybase->get_maxsurface1value() << std::endl
			  << "  Top: " << GRIB2::find_surfacetype_str(laytop->get_layer()->get_surface1type(), "?") << ' '
			  << laytop->get_layer()->get_parameter()->get_str_nonnull()
			  << " sfc1 " << laytop->get_minsurface1value() << "..." << laytop->get_maxsurface1value() << std::endl;
		for (unsigned int x(0); x < width + 2; ++x)
			for (unsigned int y(0); y < height + 2; ++y) {
				const CloudGridPt& p(grid(x, y));
				float cc(laycover->operator()(p.get_coord(), efftime, 0) * 0.01f);
				float pb(laybase->operator()(p.get_coord(), efftime, 0));
				float pt(laytop->operator()(p.get_coord(), efftime, 0));
				std::cerr << "  x " << x << " y " << y << " c " << p.get_cover() << " p " << p.get_base() << "..." << p.get_top()
					  << ' ' << (p.is_inside() ? 'T' : 'F') << ' ' << " [c " << cc << " p " << pb << "..." << pt << "] "
					  << p.get_coord().get_lat_str2() << ' ' << p.get_coord().get_lon_str2() << std::endl;
				if (std::isnan(cc)) {
					GRIB2::LayerInterpolateResult::LinInterp intp(laycover->operator()(p.get_coord()));
					std::cerr << "  cover: " << intp[0] << ' ' << intp[1] << ' ' << intp[2] << ' ' << intp[3] << std::endl;
				}
				if (std::isnan(pb)) {
					GRIB2::LayerInterpolateResult::LinInterp intp(laybase->operator()(p.get_coord()));
					std::cerr << "  base: " << intp[0] << ' ' << intp[1] << ' ' << intp[2] << ' ' << intp[3] << std::endl;
				}
				if (std::isnan(pt)) {
					GRIB2::LayerInterpolateResult::LinInterp intp(laytop->operator()(p.get_coord()));
					std::cerr << "  top: " << intp[0] << ' ' << intp[1] << ' ' << intp[2] << ' ' << intp[3] << std::endl;
				}
			}
		std::cerr << std::endl;
	}
}

template <typename PT> void MeteoChart::AreaExtract::compute(Grid<PT>& grid, float lim)
{
	unsigned int width(grid.get_width()), height(grid.get_height());
	if (width < 2 || height < 2)
		return;
	width -= 2;
	height -= 2;
	for (unsigned int x(0); x <= width; ++x)
		for (unsigned int y(0); y <= height; ++y) {
			// Ordering of Points:
			// 0 1
			// 3 2
			unsigned int xx[4], yy[4];
			for (unsigned int i = 0; i < 4; ++i) {
				xx[i] = x + ((i ^ (i >> 1)) & 1);
				yy[i] = y + ((i >> 1) & 1);
			}
			for (unsigned int i = 0; i < 4; ++i) {
				PT& p0(grid(xx[i & 3], yy[i & 3]));
				PT& p1(grid(xx[(i + 1) & 3], yy[(i + 1) & 3]));
				PT& p2(grid(xx[(i + 2) & 3], yy[(i + 2) & 3]));
				PT& p3(grid(xx[(i + 3) & 3], yy[(i + 3) & 3]));
				if (i < 2 && p0.is_inside() && !p1.is_inside() && p2.is_inside() && !p3.is_inside()) {
					PT mid(p0, p2, lim);
					if (!mid.is_inside()) {
						m_edges.insert(Edge(p3.contour_point(p0, lim), grid.get_edgenr(xx[(i + 3) & 3], yy[(i + 3) & 3], xx[i & 3], yy[i & 3]),
								    p1.contour_point(p0, lim), grid.get_edgenr(xx[(i + 1) & 3], yy[(i + 1) & 3], xx[i & 3], yy[i & 3]),
								    mid.contour_point(p0, lim)));
						m_edges.insert(Edge(p1.contour_point(p2, lim), grid.get_edgenr(xx[(i + 1) & 3], yy[(i + 1) & 3], xx[(i + 2) & 3], yy[(i + 2) & 3]),
								    p3.contour_point(p2, lim), grid.get_edgenr(xx[(i + 3) & 3], yy[(i + 3) & 3], xx[(i + 2) & 3], yy[(i + 2) & 3]),
								    mid.contour_point(p2, lim)));
						break;
					}
					m_edges.insert(Edge(p1.contour_point(p2, lim), grid.get_edgenr(xx[(i + 1) & 3], yy[(i + 1) & 3], xx[(i + 2) & 3], yy[(i + 2) & 3]),
							    p1.contour_point(p0, lim), grid.get_edgenr(xx[(i + 1) & 3], yy[(i + 1) & 3], xx[i & 3], yy[i & 3]),
							    p1.contour_point(mid, lim)));
					m_edges.insert(Edge(p3.contour_point(p0, lim), grid.get_edgenr(xx[(i + 3) & 3], yy[(i + 3) & 3], xx[i & 3], yy[i & 3]),
							    p3.contour_point(p2, lim), grid.get_edgenr(xx[(i + 3) & 3], yy[(i + 3) & 3], xx[(i + 2) & 3], yy[(i + 2) & 3]),
							    p3.contour_point(mid, lim)));
					break;
				}
				if (p0.is_inside() && !p1.is_inside() && !p2.is_inside() && !p3.is_inside()) {
					m_edges.insert(Edge(p3.contour_point(p0, lim), grid.get_edgenr(xx[(i + 3) & 3], yy[(i + 3) & 3], xx[i & 3], yy[i & 3]),
							    p1.contour_point(p0, lim), grid.get_edgenr(xx[(i + 1) & 3], yy[(i + 1) & 3], xx[i & 3], yy[i & 3]),
							    p2.contour_point(p0, lim)));
					break;
				}
				if (p0.is_inside() && p1.is_inside() && !p2.is_inside() && !p3.is_inside()) {
					m_edges.insert(Edge(p3.contour_point(p0, lim), grid.get_edgenr(xx[(i + 3) & 3], yy[(i + 3) & 3], xx[i & 3], yy[i & 3]),
							    p2.contour_point(p1, lim), grid.get_edgenr(xx[(i + 2) & 3], yy[(i + 2) & 3], xx[(i + 1) & 3], yy[(i + 1) & 3])));
					break;
				}
				if (p0.is_inside() && p1.is_inside() && p2.is_inside() && !p3.is_inside()) {
					m_edges.insert(Edge(p0.contour_point(p3, lim), grid.get_edgenr(xx[i & 3], yy[i & 3], xx[(i + 3) & 3], yy[(i + 3) & 3]),
							    p2.contour_point(p3, lim), grid.get_edgenr(xx[(i + 2) & 3], yy[(i + 2) & 3], xx[(i + 3) & 3], yy[(i + 3) & 3]),
							    p1.contour_point(p3, lim)));
					break;
				}
			}
		}
}

std::pair<float,float> MeteoChart::AreaExtract::minmax(const Glib::RefPtr<GRIB2::LayerInterpolateResult>& lay, gint64 efftime, double sfc1value)
{
	std::pair<float,float> r(std::numeric_limits<float>::max(), std::numeric_limits<float>::min());
	if (!lay)
		return r;
	unsigned int width(lay->get_width());
	unsigned int height(lay->get_height());
	for (unsigned int x(0); x < width; ++x)
		for (unsigned int y(0); y < height; ++y) {
			float v(lay->operator()(x, y, efftime, sfc1value));
			if (std::isnan(v))
				continue;
			r.first = std::min(r.first, v);
			r.second = std::max(r.second, v);
		}
	return r;
}

std::pair<float,float> MeteoChart::AreaExtract::minmax(const Glib::RefPtr<GRIB2::LayerInterpolateResult>& layu,
						       const Glib::RefPtr<GRIB2::LayerInterpolateResult>& layv, gint64 efftime, double sfc1value)
{
	std::pair<float,float> r(std::numeric_limits<float>::max(), std::numeric_limits<float>::min());
	if (!layu || !layv)
		return r;
	if (layu->get_bbox() != layv->get_bbox())
		return r;
	unsigned int width(layu->get_width());
	unsigned int height(layu->get_height());
	if (layv->get_width() != width || layv->get_height() != height)
		return r;
	for (unsigned int x(0); x < width; ++x)
		for (unsigned int y(0); y < height; ++y) {
			float u(layu->operator()(x, y, efftime, sfc1value));
			float v(layv->operator()(x, y, efftime, sfc1value));
			if (std::isnan(u) || std::isnan(v))
				continue;
			float w(sqrtf(u * u + v * v));
			r.first = std::min(r.first, w);
			r.second = std::max(r.second, w);
		}
	return r;
}

std::vector<Point> MeteoChart::AreaExtract::extract_contour(void)
{
	if (m_edges.empty())
		return std::vector<Point>();
	std::vector<Edge> edges;
	edges.push_back(*m_edges.begin());
	m_edges.erase(m_edges.begin());
	if (false)
		std::cerr << "Edge: " << edges.back().get_startedge() << " -> " << edges.back().get_endedge()
			  << ' ' << edges.back().get_start().get_lat_str2() << ' ' << edges.back().get_start().get_lon_str2()
			  << " -> " << edges.back().get_end().get_lat_str2() << ' ' << edges.back().get_end().get_lon_str2()
			  << " (*)" << std::endl;
	for (;;) {
		edges_t::iterator i(m_edges.lower_bound(Edge(Point::invalid, edges.back().get_endedge(), Point::invalid, 0)));
		if (i == m_edges.end() || i->get_startedge() != edges.back().get_endedge())
			break;
		edges.push_back(*i);
		m_edges.erase(i);
		if (false)
			std::cerr << "Edge: " << edges.back().get_startedge() << " -> " << edges.back().get_endedge()
				  << ' ' << edges.back().get_start().get_lat_str2() << ' ' << edges.back().get_start().get_lon_str2()
				  << " -> " << edges.back().get_end().get_lat_str2() << ' ' << edges.back().get_end().get_lon_str2() << std::endl;
	}
	std::vector<Point> p;
	for (std::vector<Edge>::const_iterator i(edges.begin()), e(edges.end()); i != e; ++i) {
		p.push_back(i->get_start());
		if (!i->get_intermediate().is_invalid())
			p.push_back(i->get_intermediate());
	}
	if (edges.front().get_startedge() != edges.back().get_endedge()) {
		std::cerr << "AreaExtract::extract_contour: warning: non-closed contour" << std::endl;
		p.push_back(edges.back().get_end());
	}
	return p;
}

void MeteoChart::AreaExtract::extract_contours(const Cairo::RefPtr<Cairo::Context>& cr, const Point& center, double scalelon, double scalelat)
{
	if (!cr)
		return;
	for (;;) {
		std::vector<Point> pts(extract_contour());
		if (false) {
			std::cerr << "Contour: " << pts.size() << std::endl;
			if (false)
				for (std::vector<Point>::const_iterator pi(pts.begin()), pe(pts.end()); pi != pe; ++pi)
					std::cerr << "  " << pi->get_lat_str2() << ' ' << pi->get_lon_str2() << std::endl;
		}
		if (pts.size() < 3)
			break;
		{
			float d(pts.front().simple_distance_nmi(pts.back()));
			if (d > 50)
				std::cerr << "Contour distance: " << d << std::endl;
		}
		bool first(true);
		for (std::vector<Point>::const_iterator pi(pts.begin()), pe(pts.end()); pi != pe; ++pi) {
			Point pdiff(*pi - center);
			double x(pdiff.get_lon() * scalelon);
			double y(pdiff.get_lat() * scalelat);
			if (first) {
				cr->move_to(x, y);
				first = false;
			} else {
				cr->line_to(x, y);
			}
		}
		cr->close_path();
	}
}

MeteoChart::MeteoChart(const FPlanRoute& route, const GRIB2& wxdb)
	: m_route(route), m_wxdb(&wxdb),
	  m_minefftime(std::numeric_limits<gint64>::max()),
	  m_maxefftime(std::numeric_limits<gint64>::min()),
	  m_minreftime(std::numeric_limits<gint64>::max()),
	  m_maxreftime(std::numeric_limits<gint64>::min()),
	  m_gndchart(false)
{
}

void MeteoChart::clear(void)
{
	m_wxwindu = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxwindv = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxprmsl = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxtemperature = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxtropopause = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldbdrycover = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldlowcover = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldlowbase = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldlowtop = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldmidcover = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldmidbase = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldmidtop = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldhighcover = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldhighbase = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldhightop = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldconvcover = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldconvbase = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_wxcldconvtop = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	m_minefftime = std::numeric_limits<gint64>::max();
	m_maxefftime = std::numeric_limits<gint64>::min();
	m_minreftime = std::numeric_limits<gint64>::max();
	m_maxreftime = std::numeric_limits<gint64>::min();
}

void MeteoChart::loaddb(const Rect& bbox, gint64 efftime, double pressure)
{
	if (!m_wxdb) {
		clear();
		return;
	}
	m_minefftime = std::numeric_limits<gint64>::max();
	m_maxefftime = std::numeric_limits<gint64>::min();
	m_minreftime = std::numeric_limits<gint64>::max();
	m_maxreftime = std::numeric_limits<gint64>::min();
	Rect bbox1(bbox.oversize_nmi(200));
	bool gndchart(pressure <= 0);
	if (gndchart != m_gndchart) {
		m_gndchart = gndchart;
		m_wxwindu = m_wxwindv = m_wxtemperature = m_wxprmsl = Glib::RefPtr<GRIB2::LayerInterpolateResult>();
	}
	if (m_gndchart) {
		if (!m_wxwindu || efftime < m_wxwindu->get_minefftime() || efftime > m_wxwindu->get_maxefftime() ||
		    !m_wxwindu->get_bbox().is_inside(bbox)) {
			GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_momentum_ugrd),
										       efftime, GRIB2::surface_specific_height_gnd, 100));
			m_wxwindu = GRIB2::interpolate_results(bbox1, ll, efftime, 100);
		}
		if (!m_wxwindv || efftime < m_wxwindv->get_minefftime() || efftime > m_wxwindv->get_maxefftime() ||
		    !m_wxwindv->get_bbox().is_inside(bbox)) {
			GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_momentum_vgrd),
										       efftime, GRIB2::surface_specific_height_gnd, 100));
			m_wxwindv = GRIB2::interpolate_results(bbox1, ll, efftime, 100);
		}
		if (!m_wxtemperature || efftime < m_wxtemperature->get_minefftime() || efftime > m_wxtemperature->get_maxefftime() ||
		    !m_wxtemperature->get_bbox().is_inside(bbox)) {
			GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_temperature_tmp),
										       efftime, GRIB2::surface_ground_or_water, 0));
			m_wxtemperature = GRIB2::interpolate_results(bbox1, ll, efftime);
                }
		if (!m_wxprmsl || efftime < m_wxprmsl->get_minefftime() || efftime > m_wxprmsl->get_maxefftime() ||
		    !m_wxprmsl->get_bbox().is_inside(bbox)) {
			GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_prmsl),
										       efftime));
			m_wxprmsl = GRIB2::interpolate_results(bbox1, ll, efftime);
                }
	} else {
		if (!m_wxwindu || efftime < m_wxwindu->get_minefftime() || efftime > m_wxwindu->get_maxefftime() ||
		    pressure < m_wxwindu->get_minsurface1value() || pressure > m_wxwindu->get_maxsurface1value() ||
		    !m_wxwindu->get_bbox().is_inside(bbox)) {
			GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_momentum_ugrd),
										       efftime, GRIB2::surface_isobaric_surface, pressure));
			m_wxwindu = GRIB2::interpolate_results(bbox1, ll, efftime, pressure);
		}
		if (!m_wxwindv || efftime < m_wxwindv->get_minefftime() || efftime > m_wxwindv->get_maxefftime() ||
		    pressure < m_wxwindv->get_minsurface1value() || pressure > m_wxwindv->get_maxsurface1value() ||
		    !m_wxwindv->get_bbox().is_inside(bbox)) {
			GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_momentum_vgrd),
										       efftime, GRIB2::surface_isobaric_surface, pressure));
			m_wxwindv = GRIB2::interpolate_results(bbox1, ll, efftime, pressure);
		}
		if (!m_wxtemperature || efftime < m_wxtemperature->get_minefftime() || efftime > m_wxtemperature->get_maxefftime() ||
		    pressure < m_wxtemperature->get_minsurface1value() || pressure > m_wxtemperature->get_maxsurface1value() ||
		    !m_wxtemperature->get_bbox().is_inside(bbox)) {
			GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_temperature_tmp),
										       efftime, GRIB2::surface_isobaric_surface, pressure));
			m_wxtemperature = GRIB2::interpolate_results(bbox1, ll, efftime, pressure);
                }
		if (!m_wxprmsl || efftime < m_wxprmsl->get_minefftime() || efftime > m_wxprmsl->get_maxefftime() ||
		    pressure < m_wxprmsl->get_minsurface1value() || pressure > m_wxprmsl->get_maxsurface1value() ||
		    !m_wxprmsl->get_bbox().is_inside(bbox)) {
			GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_hgt),
										       efftime, GRIB2::surface_isobaric_surface, pressure));
			m_wxprmsl = GRIB2::interpolate_results(bbox1, ll, efftime, pressure);
                }
	}
	if (m_wxwindu) {
		m_minefftime = std::min(m_minefftime, m_wxwindu->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxwindu->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxwindu->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxwindu->get_maxreftime());
	}
	if (m_wxwindv) {
		m_minefftime = std::min(m_minefftime, m_wxwindv->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxwindv->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxwindv->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxwindv->get_maxreftime());
	}
	if (m_wxtemperature) {
		m_minefftime = std::min(m_minefftime, m_wxtemperature->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxtemperature->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxtemperature->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxtemperature->get_maxreftime());
	}
	if (m_wxprmsl) {
		m_minefftime = std::min(m_minefftime, m_wxprmsl->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxprmsl->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxprmsl->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxprmsl->get_maxreftime());
	}
	if (!m_wxtropopause || efftime < m_wxtropopause->get_minefftime() || efftime > m_wxtropopause->get_maxefftime() ||
	    !m_wxtropopause->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_hgt),
									       efftime, GRIB2::surface_tropopause, 0));
		m_wxtropopause = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxtropopause) {
		m_minefftime = std::min(m_minefftime, m_wxtropopause->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxtropopause->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxtropopause->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxtropopause->get_maxreftime());
	}
	if (!m_wxcldbdrycover || efftime < m_wxcldbdrycover->get_minefftime() || efftime > m_wxcldbdrycover->get_maxefftime() ||
	    !m_wxcldbdrycover->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_cloud_tcdc),
									       efftime, GRIB2::surface_boundary_layer_cloud, 0));
		m_wxcldbdrycover = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldbdrycover) {
		m_minefftime = std::min(m_minefftime, m_wxcldbdrycover->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldbdrycover->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldbdrycover->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldbdrycover->get_maxreftime());
	}
	if (!m_wxcldlowcover || efftime < m_wxcldlowcover->get_minefftime() || efftime > m_wxcldlowcover->get_maxefftime() ||
	    !m_wxcldlowcover->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_cloud_tcdc),
									       efftime, GRIB2::surface_low_cloud, 0));
		m_wxcldlowcover = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldlowcover) {
		m_minefftime = std::min(m_minefftime, m_wxcldlowcover->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldlowcover->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldlowcover->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldlowcover->get_maxreftime());
	}
	if (!m_wxcldlowbase || efftime < m_wxcldlowbase->get_minefftime() || efftime > m_wxcldlowbase->get_maxefftime() ||
	    !m_wxcldlowbase->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
									       efftime, GRIB2::surface_low_cloud_bottom, 0));
		m_wxcldlowbase = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldlowbase) {
		m_minefftime = std::min(m_minefftime, m_wxcldlowbase->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldlowbase->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldlowbase->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldlowbase->get_maxreftime());
	}
	if (!m_wxcldlowtop || efftime < m_wxcldlowtop->get_minefftime() || efftime > m_wxcldlowtop->get_maxefftime() ||
	    !m_wxcldlowtop->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
									       efftime, GRIB2::surface_low_cloud_top, 0));
		m_wxcldlowtop = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldlowtop) {
		m_minefftime = std::min(m_minefftime, m_wxcldlowtop->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldlowtop->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldlowtop->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldlowtop->get_maxreftime());
	}
	if (!m_wxcldmidcover || efftime < m_wxcldmidcover->get_minefftime() || efftime > m_wxcldmidcover->get_maxefftime() ||
	    !m_wxcldmidcover->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_cloud_tcdc),
									       efftime, GRIB2::surface_middle_cloud, 0));
		m_wxcldmidcover = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldmidcover) {
		m_minefftime = std::min(m_minefftime, m_wxcldmidcover->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldmidcover->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldmidcover->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldmidcover->get_maxreftime());
	}
	if (!m_wxcldmidbase || efftime < m_wxcldmidbase->get_minefftime() || efftime > m_wxcldmidbase->get_maxefftime() ||
	    !m_wxcldmidbase->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
									       efftime, GRIB2::surface_middle_cloud_bottom, 0));
		m_wxcldmidbase = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldmidbase) {
		m_minefftime = std::min(m_minefftime, m_wxcldmidbase->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldmidbase->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldmidbase->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldmidbase->get_maxreftime());
	}
	if (!m_wxcldmidtop || efftime < m_wxcldmidtop->get_minefftime() || efftime > m_wxcldmidtop->get_maxefftime() ||
	    !m_wxcldmidtop->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
									       efftime, GRIB2::surface_middle_cloud_top, 0));
		m_wxcldmidtop = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldmidtop) {
		m_minefftime = std::min(m_minefftime, m_wxcldmidtop->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldmidtop->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldmidtop->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldmidtop->get_maxreftime());
	}
	if (!m_wxcldhighcover || efftime < m_wxcldhighcover->get_minefftime() || efftime > m_wxcldhighcover->get_maxefftime() ||
	    !m_wxcldhighcover->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_cloud_tcdc),
									       efftime, GRIB2::surface_top_cloud, 0));
		m_wxcldhighcover = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldhighcover) {
		m_minefftime = std::min(m_minefftime, m_wxcldhighcover->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldhighcover->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldhighcover->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldhighcover->get_maxreftime());
	}
	if (!m_wxcldhighbase || efftime < m_wxcldhighbase->get_minefftime() || efftime > m_wxcldhighbase->get_maxefftime() ||
	    !m_wxcldhighbase->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
									       efftime, GRIB2::surface_top_cloud_bottom, 0));
		m_wxcldhighbase = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldhighbase) {
		m_minefftime = std::min(m_minefftime, m_wxcldhighbase->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldhighbase->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldhighbase->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldhighbase->get_maxreftime());
	}
	if (!m_wxcldhightop || efftime < m_wxcldhightop->get_minefftime() || efftime > m_wxcldhightop->get_maxefftime() ||
	    !m_wxcldhightop->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
									       efftime, GRIB2::surface_top_cloud_top, 0));
		m_wxcldhightop = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldhightop) {
		m_minefftime = std::min(m_minefftime, m_wxcldhightop->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldhightop->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldhightop->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldhightop->get_maxreftime());
	}
	if (!m_wxcldconvcover || efftime < m_wxcldconvcover->get_minefftime() || efftime > m_wxcldconvcover->get_maxefftime() ||
	    !m_wxcldconvcover->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_cloud_tcdc),
									       efftime, GRIB2::surface_convective_cloud, 0));
		m_wxcldconvcover = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldconvcover) {
		m_minefftime = std::min(m_minefftime, m_wxcldconvcover->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldconvcover->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldconvcover->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldconvcover->get_maxreftime());
	}
	if (!m_wxcldconvbase || efftime < m_wxcldconvbase->get_minefftime() || efftime > m_wxcldconvbase->get_maxefftime() ||
	    !m_wxcldconvbase->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
									       efftime, GRIB2::surface_convective_cloud_bottom, 0));
		m_wxcldconvbase = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldconvbase) {
		m_minefftime = std::min(m_minefftime, m_wxcldconvbase->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldconvbase->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldconvbase->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldconvbase->get_maxreftime());
	}
	if (!m_wxcldconvtop || efftime < m_wxcldconvtop->get_minefftime() || efftime > m_wxcldconvtop->get_maxefftime() ||
	    !m_wxcldconvtop->get_bbox().is_inside(bbox)) {
		GRIB2::layerlist_t ll(const_cast<GRIB2 *>(m_wxdb)->find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_pres),
									       efftime, GRIB2::surface_convective_cloud_top, 0));
		m_wxcldconvtop = GRIB2::interpolate_results(bbox1, ll, efftime);
	}
	if (m_wxcldconvtop) {
		m_minefftime = std::min(m_minefftime, m_wxcldconvtop->get_minefftime());
		m_maxefftime = std::max(m_maxefftime, m_wxcldconvtop->get_maxefftime());
		m_minreftime = std::min(m_minreftime, m_wxcldconvtop->get_minreftime());
		m_maxreftime = std::max(m_maxreftime, m_wxcldconvtop->get_maxreftime());
	}
}

void MeteoChart::get_fullscale(int width, int height, Point& center, double& scalelon, double& scalelat) const
{
	center = Point::invalid;
	scalelon = scalelat = std::numeric_limits<double>::quiet_NaN();
	if (width <= 20 || height <= 20 || m_route.get_nrwpt() < 2)
		return;
	width -= 20;
	height -= 20;
	Rect bbox(m_route.get_bbox().oversize_nmi(50));
	center = bbox.boxcenter();
	Point psz(bbox.get_northeast() - bbox.get_southwest());
	double c(std::max(0.01, cos(center.get_lat_rad())));
	double sc(std::min(width / c / (double)(uint32_t)psz.get_lon(), 
			   height / (double)(uint32_t)psz.get_lat()));
	scalelon = sc * c;
	scalelat = -sc;
	if (false)
		std::cerr << "MeteoChart::get_fullscale: " << width << 'x' << height << ' ' 
			  << center.get_lat_str2() << ' ' << center.get_lon_str2() << ' '
			  << scalelat << ' ' << scalelon << std::endl;
}

gint64 MeteoChart::get_midefftime(void) const
{
	if (m_route.get_nrwpt() < 1)
		return 0;
	time_t tst(m_route[0].get_time_unix());
	time_t td(m_route[m_route.get_nrwpt()-1].get_time_unix());
	td -= tst;
	return tst + td / 2;
}

void MeteoChart::draw(const Cairo::RefPtr<Cairo::Context>& cr, int width, int height,
		      const Point& center, double scalelon, double scalelat, double pressure)
{
	if (false)
		std::cerr << "MeteoChart::draw: " << width << 'x' << height << ' '
			  << center.get_lat_str2() << ' ' << center.get_lon_str2() << ' '
			  << scalelat << ' ' << scalelon << ' ' << pressure << std::endl;
	cr->save();
	// background
	cr->set_source_rgb(1.0, 1.0, 1.0);
	cr->paint();
	// scaling
	cr->select_font_face("sans", Cairo::FONT_SLANT_NORMAL, Cairo::FONT_WEIGHT_NORMAL);
	cr->set_font_size(12);
	// check for database
	if (!m_wxdb) {
		cr->set_source_rgb(0.0, 0.0, 0.0);
		cr->move_to(width * 0.5, height * 0.5);
		cr->show_text("No Data");
		cr->restore();
		return;
	}
	if (m_route.get_nrwpt() < 2) {
		cr->set_source_rgb(0.0, 0.0, 0.0);
		cr->move_to(width * 0.5, height * 0.5);
		cr->show_text("No Flight Plan");
		cr->restore();
		return;
	}
	Rect bbox;
	{
		Point phsz(width * 0.5 / fabs(scalelat), height * 0.5 / fabs(scalelon));
		bbox = Rect(center - phsz, center + phsz).oversize_nmi(100);
	}
	gint64 efftime(get_midefftime());
	loaddb(bbox, efftime, pressure);
	// draw chart legend
	int32_t altitude(0);
	if (!m_gndchart) {
		float alt;
		IcaoAtmosphere::std_pressure_to_altitude(&alt, 0, pressure * 0.01);
		altitude = Point::round<int32_t,float>(alt * Point::m_to_ft);
	}
	{
		cr->set_source_rgb(0.0, 0.0, 0.0);
		std::ostringstream oss;
		if (m_gndchart)
			oss << "GND";
		else
			oss << std::fixed << std::setprecision(0) << (pressure * 0.01) << "hPa (F" << std::setw(3)
			    << std::setfill('0') << ((altitude + 50) / 100) << ')';
		if (m_route.get_nrwpt() > 0) {
			oss << " \342\200\242 ";
			const FPlanWaypoint& wpt(m_route[0]);
			if (wpt.get_icao().empty() || (wpt.get_type() == FPlanWaypoint::type_airport && wpt.get_icao() == "ZZZZ"))
				oss << wpt.get_name();
			else
				oss << wpt.get_icao();
			std::string tm0(Glib::TimeVal(wpt.get_time_unix(), 0).as_iso8601());
			oss << ' ' << tm0.substr(0, 10) << ' ' << tm0.substr(11, 5) << 'Z';
			if (m_route.get_nrwpt() > 1) {
				const FPlanWaypoint& wpt(m_route[m_route.get_nrwpt() - 1]);
				oss << " \342\206\222 ";
				if (wpt.get_icao().empty() || (wpt.get_type() == FPlanWaypoint::type_airport && wpt.get_icao() == "ZZZZ"))
					oss << wpt.get_name();
				else
					oss << wpt.get_icao();
				std::string tm1(Glib::TimeVal(wpt.get_time_unix(), 0).as_iso8601());
				oss << ' ';
				if (tm0.substr(0, 10) != tm1.substr(0, 10))
					oss << tm1.substr(0, 10) << ' ';
				oss << tm1.substr(11, 5) << 'Z';
			}
		}
		{
			std::string tm0(Glib::TimeVal(get_minreftime(), 0).as_iso8601());
			oss << " \342\200\242 GFS RefTime " << tm0.substr(0, 10) << ' ' << tm0.substr(11, 5) << 'Z';
			if (get_maxreftime() != get_minreftime()) {
				std::string tm1(Glib::TimeVal(get_maxreftime(), 0).as_iso8601());
				oss << '-';
				if (tm0.substr(0, 10) != tm1.substr(0, 10))
					oss << tm1.substr(0, 10) << ' ';
				oss << tm1.substr(11, 5) << 'Z';
			}
		}
		std::ostringstream oss2;
		{
			std::string tm0(Glib::TimeVal(get_minefftime(), 0).as_iso8601());
			oss2 << " * EffTime " << tm0.substr(0, 10) << ' ' << tm0.substr(11, 5) << 'Z';
			if (get_maxefftime() != get_minefftime()) {
				std::string tm1(Glib::TimeVal(get_maxefftime(), 0).as_iso8601());
				oss2 << '-';
				if (tm0.substr(0, 10) != tm1.substr(0, 10))
					oss2 << tm1.substr(0, 10) << ' ';
				oss2 << tm1.substr(11, 5) << 'Z';
			}
		}
		static const char arident[] = " \342\200\242 autorouter.eu";
		Cairo::TextExtents ext;
		cr->get_text_extents(oss.str() + oss2.str() + arident, ext);
		if (ext.width <= width)
			oss << oss2.str();
		oss << arident;
		cr->get_text_extents(oss.str(), ext);
		cr->move_to(0 - ext.x_bearing, height - ext.height - 2U - ext.y_bearing);
		cr->show_text(oss.str());
		height -= ext.height + 9U;
	}
	// border
	cr->rectangle(0, 0, width, height);
	set_color_marking(cr);
	cr->set_line_width(2);
	cr->stroke_preserve();
	cr->clip();
	cr->translate(width * 0.5, height * 0.5);
	// grid
	{
		Point pgrid;
		pgrid.set_lon_deg(0.5);
		pgrid.set_lat_deg(0.5);
		static const double minpixel(100);
		static const double coordpx(10);
		for (unsigned int i = 0; i < 7; ++i) {
			if (fabs(pgrid.get_lon() * scalelon) >= minpixel)
				break;
			if ((i - 2) % 3)
				pgrid.set_lon(pgrid.get_lon() << 1);
			else
				pgrid.set_lon((pgrid.get_lon() << 1) + (pgrid.get_lon() >> 1));
		}
		for (unsigned int i = 0; i < 7; ++i) {
			if (fabs(pgrid.get_lat() * scalelat) >= minpixel)
				break;
			if ((i - 2) % 3)
				pgrid.set_lat(pgrid.get_lat() << 1);
			else
				pgrid.set_lat((pgrid.get_lat() << 1) + (pgrid.get_lat() >> 1));
		}
		Point pst(bbox.get_southwest());
		pst.set_lon(pst.get_lon() / pgrid.get_lon() * pgrid.get_lon());
		pst.set_lat(pst.get_lat() / pgrid.get_lat() * pgrid.get_lat());
		Point plen(bbox.get_northeast() - pst);
		// Lon Grid
		for (Point::coord_t cd(0); cd < plen.get_lon(); cd += pgrid.get_lon()) {
			Point::coord_t coord(pst.get_lon() + cd);
			double xc((coord - center.get_lon()) * scalelon);
			std::ostringstream oss;
			{
				int min(Point::round<int,double>(Point(coord, coord).get_lon_deg_dbl() * 60));
				if (min < -180 * 60)
					min += 360 * 60;
				else if (min > 180 * 60)
					min -= 360 * 60;
				if (min < 0) {
					oss << 'W';
					min = -min;
				} else {
					oss << 'E';
				}
				unsigned int deg(min);
				deg /= 60;
				min -= deg * 60;
				oss << std::setw(3) << std::setfill('0') << deg;
				if (min)
					oss << ' ' << std::setw(2) << std::setfill('0') << min;
			}
			Cairo::TextExtents ext;
			cr->get_text_extents(oss.str(), ext);
			cr->set_source_rgb(0.0, 0.0, 0.0);
			cr->set_line_width(2);
			cr->move_to(xc, -0.5 * height);
			cr->rel_line_to(0, coordpx);
			cr->move_to(xc, 0.5 * height);
			cr->rel_line_to(0, -coordpx);
			cr->stroke();
			cr->move_to(xc - ext.width * 0.5 - ext.x_bearing, coordpx - 0.5 * height + 1 - ext.y_bearing);
			cr->show_text(oss.str());
			cr->move_to(xc - ext.width * 0.5 - ext.x_bearing, 0.5 * height - coordpx - 1 - ext.height - ext.y_bearing);
			cr->show_text(oss.str());
			if (false)
				std::cerr << "Lon Grid: " << Point(coord, coord).get_lon_str2() << " xc " << xc
					  << ' ' << oss.str() << std::endl;
		}
		// Lat Grid
		for (Point::coord_t cd(0); cd < plen.get_lat(); cd += pgrid.get_lat()) {
			Point::coord_t coord(pst.get_lat() + cd);
			double yc((coord - center.get_lat()) * scalelat);
			std::ostringstream oss;
			{
				int min(Point::round<int,double>(Point(coord, coord).get_lat_deg_dbl() * 60));
				if (min < 0) {
					oss << 'S';
					min = -min;
				} else {
					oss << 'N';
				}
				unsigned int deg(min);
				deg /= 60;
				min -= deg * 60;
				oss << std::setw(2) << std::setfill('0') << deg;
				if (min)
					oss << ' ' << std::setw(2) << std::setfill('0') << min;
			}
			Cairo::TextExtents ext;
			cr->get_text_extents(oss.str(), ext);
			cr->set_source_rgb(0.0, 0.0, 0.0);
			cr->set_line_width(2);
			cr->move_to(-0.5 * width, yc);
			cr->rel_line_to(coordpx, 0);
			cr->move_to(0.5 * width, yc);
			cr->rel_line_to(-coordpx, 0);
			cr->stroke();
			cr->move_to(coordpx - 0.5 * width + 1 - ext.x_bearing, yc - ext.height * 0.5 - ext.y_bearing);
			cr->show_text(oss.str());
			cr->move_to(0.5 * width - coordpx - 1 - ext.width - ext.x_bearing, yc - ext.height * 0.5 - ext.y_bearing);
			cr->show_text(oss.str());
			if (false)
				std::cerr << "Lat Grid: " << Point(coord, coord).get_lat_str2() << " yc " << yc
					  << ' ' << oss.str() << std::endl;
		}
		// wind barbs
		for (Point::coord_t cdlon(0); cdlon < plen.get_lon(); cdlon += pgrid.get_lon()) {
			for (Point::coord_t cdlat(0); cdlat < plen.get_lat(); cdlat += pgrid.get_lat()) {
				Point pt(pst.get_lon() + cdlon, pst.get_lat() + cdlat);
				double xc, yc;
				{
					Point pdiff(pt - center);
					xc = pdiff.get_lon() * scalelon;
					yc = pdiff.get_lat() * scalelat;
				}
				if (fabs(xc) > 0.5 * width - 30 || fabs(yc) > 0.5 * height - 30)
					continue;
				std::ostringstream ossflg;
				// compute cloud cover
				double alpha(1);
				if (m_gndchart) {
					if (m_wxcldbdrycover) {
						float v(m_wxcldbdrycover->operator()(pt, efftime, 0));
						if (!std::isnan(v))
							alpha = 1 - std::min(1.f, std::max(0.f, v * 0.01f));
					}
				} else {
					if (m_wxcldlowcover && m_wxcldlowbase && m_wxcldlowtop) {
						float cover(m_wxcldlowcover->operator()(pt, efftime, 0));
						float base(m_wxcldlowbase->operator()(pt, efftime, 0));
						float top(m_wxcldlowtop->operator()(pt, efftime, 0));
						if (!std::isnan(cover) && !std::isnan(base) && !std::isnan(top) &&
						    pressure <= base && pressure >= top)
						       	alpha *= 1 - std::min(1.f, std::max(0.f, cover * 0.01f));
					}
					if (m_wxcldmidcover && m_wxcldmidbase && m_wxcldmidtop) {
						float cover(m_wxcldmidcover->operator()(pt, efftime, 0));
						float base(m_wxcldmidbase->operator()(pt, efftime, 0));
						float top(m_wxcldmidtop->operator()(pt, efftime, 0));
						if (!std::isnan(cover) && !std::isnan(base) && !std::isnan(top) &&
						    pressure <= base && pressure >= top)
						       	alpha *= 1 - std::min(1.f, std::max(0.f, cover * 0.01f));
					}
					if (m_wxcldhighcover && m_wxcldhighbase && m_wxcldhightop) {
						float cover(m_wxcldhighcover->operator()(pt, efftime, 0));
						float base(m_wxcldhighbase->operator()(pt, efftime, 0));
						float top(m_wxcldhightop->operator()(pt, efftime, 0));
						if (!std::isnan(cover) && !std::isnan(base) && !std::isnan(top) &&
						    pressure <= base && pressure >= top)
						       	alpha *= 1 - std::min(1.f, std::max(0.f, cover * 0.01f));
					}
					if (m_wxcldconvcover && m_wxcldconvbase && m_wxcldconvtop) {
						float cover(m_wxcldconvcover->operator()(pt, efftime, 0));
						float base(m_wxcldconvbase->operator()(pt, efftime, 0));
						float top(m_wxcldconvtop->operator()(pt, efftime, 0));
						if (!std::isnan(cover) && !std::isnan(base) && !std::isnan(top) &&
						    pressure <= base && pressure >= top) {
							cover = std::min(1.f, std::max(0.f, cover * 0.01f));
						       	alpha *= 1 - cover;
							if (cover >= 0.25)
								ossflg << "CB";
							else if (cover >= 0.1)
								ossflg << "ISOL CB";
						}
					}
				}
				cr->save();
				cr->translate(xc, yc);
				cr->scale(10.0, -10.0);
				double winddirrad(0);
				int windspeed5(0);
				if (m_wxwindu && m_wxwindv) {
					float wu(m_wxwindu->operator()(pt, efftime, m_gndchart ? 100 : pressure));
					float wv(m_wxwindv->operator()(pt, efftime, m_gndchart ? 100 : pressure));
					if (!std::isnan(wu) && !std::isnan(wv)) {
						wu *= (-1e-3f * Point::km_to_nmi * 3600);
						wv *= (-1e-3f * Point::km_to_nmi * 3600);
						windspeed5 = Point::round<int,double>(sqrtf(wu * wu + wv * wv) * 0.2);
						winddirrad = M_PI - atan2f(wu, wv);
					}
				}
				static const double windradius = 1.5;
				if (windspeed5) {
					cr->save();
					cr->rotate(M_PI - winddirrad);
					// draw twice, first white with bigger line width
					for (unsigned int i(0); i < 2; ++i) {
						cr->set_line_width(i ? 0.10 : 0.30);
						cr->set_source_rgb(i ? 0.0 : 1.0, i ? 0.0 : 1.0, i ? 0.0 : 1.0);
						double y(-windradius);
						cr->move_to(0, 0);
						cr->line_to(0, y);
						cr->stroke();
						int speed(windspeed5);
						if (speed < 10)
							y -= 0.2;
						while (speed >= 10) {
							speed -= 10;
							cr->move_to(0, y);
							cr->rel_line_to(-0.4, 0.2);
							cr->rel_line_to(0.4, 0.2);
							cr->fill();
							y += 0.4;
						}
						while (speed >= 2) {
							speed -= 2;
							y += 0.2;
							cr->move_to(0, y);
							cr->rel_line_to(-0.4, -0.2);
							cr->stroke();
						}
						if (speed) {
							y += 0.2;
							cr->move_to(0, y);
							cr->rel_line_to(-0.2, -0.1);
							cr->stroke();
						}
					}
					cr->restore();
				}
				// draw cloud circle
				static const double cldradius = 0.5;
				cr->begin_new_path();
				cr->arc(0.0, 0.0, cldradius, 0.0, 2.0 * M_PI);
				cr->close_path();
				cr->set_line_width(0.10);
				if (alpha >= 0.95) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
				} else if (alpha >= 0.85) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->move_to(0, cldradius);
					cr->line_to(0, -cldradius);
					cr->stroke();
				} else if (alpha >= 0.65) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 2.0 * M_PI);
					cr->line_to(0, 0);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
				} else if (alpha >= 0.55) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 2.0 * M_PI);
					cr->line_to(0, 0);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
					cr->move_to(0, cldradius);
					cr->line_to(0, -cldradius);
					cr->stroke();
				} else if (alpha >= 0.45) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 0.5 * M_PI);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
				} else if (alpha >= 0.35) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 0.5 * M_PI);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
					cr->move_to(-cldradius, 0);
					cr->line_to(cldradius, 0);
					cr->stroke();
				} else if (alpha >= 0.15) {
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->fill_preserve();
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->stroke();
					cr->begin_new_path();
					cr->arc(0.0, 0.0, cldradius, 1.5 * M_PI, 1.0 * M_PI);
					cr->line_to(0, 0);
					cr->close_path();
					cr->fill_preserve();
					cr->stroke();
				} else if (alpha >= 0.05) {
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->fill_preserve();
					cr->stroke();
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->move_to(0, -0.9 * cldradius);
					cr->line_to(0, 0.9 * cldradius);
					cr->stroke();
				} else {
					cr->set_source_rgb(0.0, 0.0, 0.0);
					cr->fill_preserve();
					cr->stroke();
				}
				cr->restore();
				// show temperature
				static const double textdist = 15;
				if (m_wxtemperature) {
					float temp(m_wxtemperature->operator()(pt, efftime, m_gndchart ? 0 : pressure));
					if (!std::isnan(temp)) {
						std::ostringstream oss;
						oss << Point::round<int,double>(temp - IcaoAtmosphere::degc_to_kelvin) << "\302\260";
						Cairo::TextExtents ext;
						cr->get_text_extents(oss.str(), ext);
						double xc1(xc + textdist - ext.x_bearing);
						double yc1(yc - 1.5 * ext.height - 4 - ext.y_bearing);
						cr->set_source_rgb(1.0, 1.0, 1.0);
						cr->move_to(xc1, yc1 + 1);
						cr->show_text(oss.str());
						cr->move_to(xc1, yc1 - 1);
						cr->show_text(oss.str());
						cr->move_to(xc1 + 1, yc1);
						cr->show_text(oss.str());
						cr->move_to(xc1 - 1, yc1);
						cr->show_text(oss.str());
						set_color_temp(cr);
						cr->move_to(xc1, yc1);
						cr->show_text(oss.str());
					}
				}
				if (m_wxprmsl) {
					float pr(m_wxprmsl->operator()(pt, efftime, m_gndchart ? 0 : pressure));
					if (!std::isnan(pr)) {
						std::ostringstream oss;
						if (m_gndchart)
							oss << Point::round<int,double>(pr * 0.01);
						else
							oss << Point::round<int,double>(pr * Point::m_to_ft_dbl) << '\'';
						Cairo::TextExtents ext;
						cr->get_text_extents(oss.str(), ext);
						double xc1(xc + textdist - ext.x_bearing);
						double yc1(yc - 0.5 * ext.height - ext.y_bearing);
						cr->set_source_rgb(1.0, 1.0, 1.0);
						cr->move_to(xc1, yc1 + 1);
						cr->show_text(oss.str());
						cr->move_to(xc1, yc1 - 1);
						cr->show_text(oss.str());
						cr->move_to(xc1 + 1, yc1);
						cr->show_text(oss.str());
						cr->move_to(xc1 - 1, yc1);
						cr->show_text(oss.str());
						set_color_qnh(cr);
						cr->move_to(xc1, yc1);
						cr->show_text(oss.str());
					}
				}
				if (!ossflg.str().empty()) {
					Cairo::TextExtents ext;
					cr->get_text_extents(ossflg.str(), ext);
					double xc1(xc + textdist - ext.x_bearing);
					double yc1(yc + 0.5 * ext.height + 4 - ext.y_bearing);
					cr->set_source_rgb(1.0, 1.0, 1.0);
					cr->move_to(xc1, yc1 + 1);
					cr->show_text(ossflg.str());
					cr->move_to(xc1, yc1 - 1);
					cr->show_text(ossflg.str());
					cr->move_to(xc1 + 1, yc1);
					cr->show_text(ossflg.str());
					cr->move_to(xc1 - 1, yc1);
					cr->show_text(ossflg.str());
					set_color_wxnote(cr);
					cr->move_to(xc1, yc1);
					cr->show_text(ossflg.str());
				}
			}
		}
	}
	// QNH / altitude contours
	if (m_wxprmsl) {
		std::pair<float,float> mm(AreaExtract::minmax(m_wxprmsl, efftime, m_gndchart ? 0 : pressure));
		if (mm.first < mm.second) {
			double pinc(m_gndchart ? 100 : 25 * Point::ft_to_m_dbl);
			double p0(std::floor(mm.first / pinc) * pinc);
			double p1(std::ceil(mm.second / pinc) * pinc);
			for (; p0 <= p1; p0 += pinc) {
				AreaExtract area(m_wxprmsl, p0, efftime, m_gndchart ? 0 : pressure);
				cr->begin_new_path();
				area.extract_contours(cr, center, scalelon, scalelat);
				set_color_qnh(cr);
				cr->set_line_width(1);
				cr->stroke();
			}
		}
	}
	// wind contours
	if (m_wxwindu && m_wxwindv) {
		std::pair<float,float> mm(AreaExtract::minmax(m_wxwindu, m_wxwindv, efftime, m_gndchart ? 100 : pressure));
		if (mm.first < mm.second) {
			double pinc(1e3f / Point::km_to_nmi / 3600.0 * 25);
			double p0(std::max(1., std::floor(mm.first / pinc)) * pinc);
			double p1(std::ceil(mm.second / pinc) * pinc);
			for (; p0 <= p1; p0 += pinc) {
				AreaExtract area(m_wxwindu, m_wxwindv, p0, efftime, m_gndchart ? 0 : pressure);
				cr->begin_new_path();
				area.extract_contours(cr, center, scalelon, scalelat);
				set_color_isotach(cr);
				cr->set_line_width(1);
				std::vector<double> dashes;
				dashes.push_back(10);
				dashes.push_back(2);
				for (unsigned int x(1), n(Point::round<int,double>(p0 / pinc)); x < n; ++x) {
					dashes.push_back(2);
					dashes.push_back(2);
				}
				cr->set_dash(dashes, 0);
				cr->stroke();
			}
		}
	}
	cr->unset_dash();
	if (m_gndchart) {
		// boundary clouds
		if (m_wxcldbdrycover) {
			AreaExtract area(m_wxcldbdrycover, 0.4, efftime, 0);
			cr->begin_new_path();
			area.extract_contours(cr, center, scalelon, scalelat);
			cr->set_source_rgba(0.5, 0.5, 0.5, 0.2);
			cr->fill();
		}
	} else {
		// low clouds
		if (m_wxcldlowcover && m_wxcldlowbase && m_wxcldlowtop) {
			AreaExtract area(m_wxcldlowcover, m_wxcldlowbase, m_wxcldlowtop, efftime, pressure);
			cr->begin_new_path();
			area.extract_contours(cr, center, scalelon, scalelat);
			cr->set_source_rgba(0.5, 0.5, 0.5, 0.2);
			cr->fill();
		}
		// mid clouds
		if (m_wxcldmidcover && m_wxcldmidbase && m_wxcldmidtop) {
			AreaExtract area(m_wxcldmidcover, m_wxcldmidbase, m_wxcldmidtop, efftime, pressure);
			cr->begin_new_path();
			area.extract_contours(cr, center, scalelon, scalelat);
			cr->set_source_rgba(0.5, 0.5, 0.5, 0.2);
			cr->fill();
		}
		// high clouds
		if (m_wxcldhighcover && m_wxcldhighbase && m_wxcldhightop) {
			AreaExtract area(m_wxcldhighcover, m_wxcldhighbase, m_wxcldhightop, efftime, pressure);
			cr->begin_new_path();
			area.extract_contours(cr, center, scalelon, scalelat);
			cr->set_source_rgba(0.5, 0.5, 0.5, 0.2);
			cr->fill();
		}
		// convective clouds
		if (m_wxcldconvcover && m_wxcldconvbase && m_wxcldconvtop) {
			AreaExtract area(m_wxcldconvcover, m_wxcldconvbase, m_wxcldconvtop, efftime, pressure);
			cr->begin_new_path();
			area.extract_contours(cr, center, scalelon, scalelat);
			cr->set_source_rgba(0.5, 0.5, 0.5, 0.2);
			cr->fill();
		}
	}
	// draw route
	{
		set_color_fplan(cr);
		cr->set_line_width(2);
		cr->begin_new_path();
		for (unsigned int i = 0, j = m_route.get_nrwpt(); i < j; ++i) {
			const FPlanWaypoint& wpt(m_route[i]);
			double xc, yc;
			{
				Point pdiff(wpt.get_coord() - center);
				xc = pdiff.get_lon() * scalelon;
				yc = pdiff.get_lat() * scalelat;
			}
			if (i)
				cr->line_to(xc, yc);
			else
				cr->move_to(xc, yc);
		}
		cr->stroke();
		for (unsigned int i = 0, j = m_route.get_nrwpt(); i < j; ++i) {
			const FPlanWaypoint& wpt(m_route[i]);
			Glib::ustring txt(wpt.get_icao());
			if (txt.empty() || (wpt.get_type() == FPlanWaypoint::type_airport && txt == "ZZZZ"))
				txt = wpt.get_name();
			if (txt.empty())
				continue;
			double xc, yc;
			{
				Point pdiff(wpt.get_coord() - center);
				xc = pdiff.get_lon() * scalelon;
				yc = pdiff.get_lat() * scalelat;
			}
			unsigned int quadrant(0);
			quadrant |= 1 << ((wpt.get_truetrack() >> 14) & 3);
			if (i)
				quadrant |= 1 << (((m_route[i-1].get_truetrack() + 0x8000) >> 14) & 3);
			Cairo::TextExtents ext;
			cr->get_text_extents(txt, ext);
			xc -= ext.x_bearing;
			yc -= ext.y_bearing;
			// place label
			if (!(quadrant & 0x3)) {
				xc += 3;
				yc -= ext.height * 0.5;
			} else if (!(quadrant & 0x6)) {
				xc -= ext.width * 0.5;
				yc += 3;
			} else if (!(quadrant & 0xC)) {
				xc -= 3 + ext.width;
				yc -= ext.height * 0.5;
			} else if (!(quadrant & 0x9)) {
				xc -= ext.width * 0.5;
				yc -= 3 + ext.height;
			} else if (!(quadrant & 0x1)) {
				xc += 3;
				yc -= 3 + ext.height;
			} else if (!(quadrant & 0x2)) {
				xc += 3;
				yc += 3;
			} else if (!(quadrant & 0x4)) {
				xc -= 3 + ext.width;
				yc += 3;
			} else if (!(quadrant & 0x8)) {
				xc -= 3 + ext.width;
				yc -= 3 + ext.height;
			} else {
				xc -= ext.width * 0.5;
				yc -= ext.height * 0.5;
			}
			cr->set_source_rgb(1.0, 1.0, 1.0);
			cr->move_to(xc, yc + 1);
			cr->show_text(txt);
			cr->move_to(xc, yc - 1);
			cr->show_text(txt);
			cr->move_to(xc + 1, yc);
			cr->show_text(txt);
			cr->move_to(xc - 1, yc);
			cr->show_text(txt);
			set_color_fplan(cr);
			cr->move_to(xc, yc);
			cr->show_text(txt);
		}
	}
	// end
	cr->restore();
	cr->save();
	cr->rectangle(0, 0, width, height);
	cr->set_source_rgb(0.0, 0.0, 0.0);
	cr->set_line_width(3);
	cr->stroke();
	cr->restore();
}

const std::string& to_str(Cairo::SurfaceType st)
{
	switch (st) {
	case Cairo::SURFACE_TYPE_IMAGE:
	{
		static const std::string r("IMAGE");
		return r;
	}

	case Cairo::SURFACE_TYPE_PDF:
	{
		static const std::string r("PDF");
		return r;
	}

	case Cairo::SURFACE_TYPE_PS:
	{
		static const std::string r("PS");
		return r;
	}

	case Cairo::SURFACE_TYPE_XLIB:
	{
		static const std::string r("XLIB");
		return r;
	}

	case Cairo::SURFACE_TYPE_XCB:
	{
		static const std::string r("XCB");
		return r;
	}

	case Cairo::SURFACE_TYPE_GLITZ:
	{
		static const std::string r("GLITZ");
		return r;
	}

	case Cairo::SURFACE_TYPE_QUARTZ:
	{
		static const std::string r("QUARTZ");
		return r;
	}

	case Cairo::SURFACE_TYPE_WIN32:
	{
		static const std::string r("WIN32");
		return r;
	}

	case Cairo::SURFACE_TYPE_BEOS:
	{
		static const std::string r("BEOS");
		return r;
	}

	case Cairo::SURFACE_TYPE_DIRECTFB:
	{
		static const std::string r("DIRECTFB");
		return r;
	}

	case Cairo::SURFACE_TYPE_SVG:
	{
		static const std::string r("SVG");
		return r;
	}

	case Cairo::SURFACE_TYPE_OS2:
	{
		static const std::string r("OS2");
		return r;
	}

	case Cairo::SURFACE_TYPE_WIN32_PRINTING:
	{
		static const std::string r("PRINTING");
		return r;
	}

	case Cairo::SURFACE_TYPE_QUARTZ_IMAGE:
	{
		static const std::string r("IMAGE");
		return r;
	}

#if (CAIRO_VERSION_MAJOR > 1) || (CAIRO_VERSION_MAJOR == 1 && CAIRO_VERSION_MINOR >= 10)
	case Cairo::SURFACE_TYPE_SCRIPT:
	{
		static const std::string r("SCRIPT");
		return r;
	}

	case Cairo::SURFACE_TYPE_QT:
	{
		static const std::string r("QT");
		return r;
	}

	case Cairo::SURFACE_TYPE_RECORDING:
	{
		static const std::string r("RECORDING");
		return r;
	}

	case Cairo::SURFACE_TYPE_VG:
	{
		static const std::string r("VG");
		return r;
	}

	case Cairo::SURFACE_TYPE_GL:
	{
		static const std::string r("GL");
		return r;
	}

	case Cairo::SURFACE_TYPE_DRM:
	{
		static const std::string r("DRM");
		return r;
	}

	case Cairo::SURFACE_TYPE_TEE:
	{
		static const std::string r("TEE");
		return r;
	}

	case Cairo::SURFACE_TYPE_XML:
	{
		static const std::string r("XML");
		return r;
	}

	case Cairo::SURFACE_TYPE_SKIA:
	{
		static const std::string r("SKIA");
		return r;
	}

	case Cairo::SURFACE_TYPE_SUBSURFACE:
	{
		static const std::string r("SUBSURFACE");
		return r;
	}
#endif

	default:	
	{
		static const std::string r("?");
		return r;
	}
	}
}

const std::string& to_str(Cairo::Format f)
{
	switch (f) {
	case Cairo::FORMAT_ARGB32:
	{
		static const std::string r("ARGB32");
		return r;
	}

	case Cairo::FORMAT_RGB24:
	{
		static const std::string r("RGB32");
		return r;
	}

	case Cairo::FORMAT_A8:
	{
		static const std::string r("A8");
		return r;
	}

	case Cairo::FORMAT_A1:
	{
		static const std::string r("A1");
		return r;
	}

	case Cairo::FORMAT_RGB16_565:
	{
		static const std::string r("RGB16_565");
		return r;
	}

	default:
	{
		static const std::string r("Unknown Format");
		return r;
	}
	}
}
